/*
 * ar2c-lp-utils - ar2c-par-merge.c
 *
 * 	Merge file with AR-2c parameters
 *
 * Copyright (C) 2008  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <ar2cpar.h>
#include <ctype.h>
#include <stdarg.h>

static int verbose = 0;
static int force;
static const char *output_fname = NULL;
static const char *source_fname = NULL;

static struct ar2c_config config;

static void show_usage(char *name)
{
	fprintf(stderr, "ar2c-par-merge " PACKAGE_VERSION ".\n"
"Copyright (C) 2008  Krzysztof Mazur <krzysiek@podlesie.net>.\n");

	fprintf(stderr,	"ar2c-par-merge.\n"
			"usage: %s [options] [files...]\n"
			"Options:\n"
			"  -f, --force		ignore some errors\n"
			"  -h, --help		show this help\n"
			"  -q, --quiet		be quiet\n"
			"  -o, --output		set output to file\n"
			"  -t, --timeout\t	set timeout\n"
			"  -v, --verbose\t	be verbose\n"
			"  -V, --version\t	show version\n",
			(name == NULL) ? "ar2c-par-merge" : name);
	exit(EXIT_SUCCESS);
}

static void show_version(void)
{
	fprintf(stderr, "ar2c-par-merge " PACKAGE_VERSION ".\n"
"Copyright (C) 2008  Krzysztof Mazur <krzysiek@podlesie.net>.\n\n");

	fprintf(stderr,	"This program comes with ABSOLUTELY NO WARRANTY; "
			"see COPYING for details.\n"
			"This is free software, and you are welcome to "
			"redistribute it\n"
			"under certain conditions; see COPYING for details.\n");
	exit(EXIT_SUCCESS);
}

static void parse_command_line(int argc, char **args)
{
	int c;
	int option_index = 0;

	static struct option long_options[] = {
		{"force", 0, 0, 'f'},
		{"help", 0, 0, 'h'},
		{"quiet", 0, 0, 'q'},
		{"output", 1, 0, 'o'},
		{"verbose", 0, 0, 'v'},
		{"version", 0, 0, 'V'},
		{0, 0, 0, 0},
	};

	while (1) {
		c = getopt_long(argc, args, "fhqo:vV",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'f':
			force = 1;
			break;
		case 'h':
			show_usage(args[0]);
			break;
		case 'v':
			verbose++;
			break;
		case 'o':
			output_fname = optarg;
			break;
		case 'q':
			verbose--;
			break;
		case 'V':
			show_version();
			break;
		default:
			fprintf(stderr, "Unknown option.\n");
			exit(EXIT_FAILURE);
		}
	}

	while (optind < argc) {
		const char *fname = args[optind++];
		FILE *f = fopen(fname, "r");
		int ret;

		if (f == NULL) {
			fprintf(stderr, "fopen(%s): %s\n", fname,
					strerror(errno));
			exit(EXIT_FAILURE);
		}
		source_fname = fname;
		ret = ar2c_config_load(&config, f);
		if (ret) {
			if (verbose >= 0)
				fprintf(stderr, "error: ar2c_config_load(%s): "
						"%s\n", fname, strerror(-ret));
			if (!force)
				exit(EXIT_FAILURE);
		}
		fclose(f);
	}
}

#define log_warning(fmt...)	fprintf(stderr, fmt)

static int config_error_handler(struct ar2c_config *c, int priority,
		int line, const char *fmt, ...)
{
	va_list args;
	const char *errname = "error";
	
	if (priority == AR2C_CONFIG_WARNING)
		errname = "warning";
	else if (priority == AR2C_CONFIG_NOTICE)
		errname = "notice";

	fprintf(stderr, "%s: %s:%d ", errname, source_fname, line);
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);

	if (force)
		return 0;
	return priority <= AR2C_CONFIG_ERR;
}

int main(int argc, char **args)
{
	FILE *out_file;
	int ret;

	ar2c_config_init(&config);
	ar2c_config_set_error_handler(&config, config_error_handler);
	parse_command_line(argc, args);

	if (output_fname != NULL) {
		out_file = fopen(output_fname, "w");
		if (out_file == NULL) {
			fprintf(stderr, "fopen(%s): %s\n", output_fname,
					strerror(errno));
			exit(EXIT_FAILURE);
		}
	} else
		out_file = stdout;

	ret = ar2c_config_store(&config, out_file);
	if (ret) {
		fprintf(stderr, "error: ar2c_config_store(%s): %s\n",
				output_fname ? output_fname : "stdout",
				strerror(-ret));
		exit(EXIT_FAILURE);
	}
	return EXIT_SUCCESS;
}
