/*
 * ar2c-par-utils - ar2c-par-to-ar3c.c
 *
 * Translate AR-2c parameters to AR-3c configuration
 *
 * Copyright (C) 2012..2013  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <signal.h>
#include <tread.h>
#include <inttypes.h>
#include <ar2cpar.h>
#include <ctype.h>
#include <stdarg.h>

#ifdef __GCC__
#define likely(x)	__builtin_expect(!!(x), 1)
#define unlikely(x)	__builtin_expect(!!(x), 0)
#else
#define likely(x)	(x)
#define unlikely(x)	(x)
#endif

static struct ar2c_config config;
static int verbose;
static int force;
static const char *source_fname;
static const char *output_dir;

static void vreportf(const char *prefix, const char *err, va_list params)
{
	char msg[4096];
	vsnprintf(msg, sizeof(msg), err, params);
	fprintf(stderr, "%s%s\n", prefix, msg);
}

static void die(const char *err, ...)
{
	va_list params;

	va_start(params, err);
	vreportf("fatal: ", err, params);
	va_end(params);
	exit(EXIT_FAILURE);
}

static void error(const char *err, ...)
{
	va_list params;

	va_start(params, err);
	vreportf("error: ", err, params);
	va_end(params);
}

void fclose_or_error(FILE *f)
{
	int ret;

	ret = fclose(f);
	if (unlikely(ret))
		error("");
}

static const char copyright_string[] =
"ar2c-par-to-ar3c " PACKAGE_VERSION ".\n"
"Copyright (C) 2012..2013  Krzysztof Mazur <krzysiek@podlesie.net>.\n";

static void show_copyright(void)
{
	puts(copyright_string);
}

static const char usage[] =
"usage: %s [options] -o DIRECTORY [<parameter-file>...]\n"
"Options:\n"
"  -V, --version              show version\n"
"  -o, --output=DIRECTORY     set output directory, directory must exist\n"
"  -h, --help                 show this help\n"
"\n"
"Report bugs to krzysiek@podlesie.net.\n";

static void show_usage(char *name)
{
	show_copyright();

	printf(usage, (name == NULL) ? "ar2c-par-to-ar3c" : name);
	exit(EXIT_SUCCESS);
}

static const char license[] =
"This program comes with ABSOLUTELY NO WARRANTY; see COPYING for details.\n"
"This is free software, and you are welcome to redistribute it\n"
"under certain conditions; see COPYING for details.\n";

static void show_version(void)
{
	show_copyright();

	puts(license);
	exit(EXIT_SUCCESS);
}

static void parse_command_line(int argc, char **args)
{
	static struct option long_options[] = {
		{"help", 0, 0, 'h'},
		{"output", 1, 0, 'o'},
		{"quiet", 0, 0, 'q'},
		{"verbose", 0, 0, 'v'},
		{"version", 0, 0, 'V'},
		{0, 0, 0, 0},
	};
	int option_index = 0;
	int c;

	while (1) {
		c = getopt_long(argc, args, "ho:qvV",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'h':
			show_usage(args[0]);
			break;
		case 'v':
			verbose++;
			break;
		case 'q':
			verbose--;
			break;
		case 'o':
			output_dir = optarg;
			break;
		case 'V':
			show_version();
			break;
		default:
			die("unknown option '%c'.\n", c);
		}
	}

	while (optind < argc) {
		const char *fname = args[optind++];
		FILE *f = fopen(fname, "r");
		int ret;

		if (f == NULL)
			die("fopen(%s): %s\n", fname, strerror(errno));
		source_fname = fname;
		ret = ar2c_config_load(&config, f);
		if (ret) {
			if (verbose >= 0)
				error("ar2c_config_load(%s): %s\n",
						fname, strerror(-ret));
			if (!force)
				exit(EXIT_FAILURE);
		}
		fclose(f);
	}
}

static int config_error_handler(struct ar2c_config *c, int priority,
		int line, const char *fmt, ...)
{
	va_list args;
	const char *errname = "error";

	if (priority == AR2C_CONFIG_WARNING)
		errname = "warning";
	else if (priority == AR2C_CONFIG_NOTICE)
		errname = "notice";

	fprintf(stderr, "%s: %s:%d ", errname, source_fname, line);
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);

	if (force)
		return 0;
	return priority <= AR2C_CONFIG_ERR;
}

#define log_warning(fmt...)	fprintf(stderr, fmt)

static char *make_message(const char *fmt, ...)
{
	/* Guess we need no more than 100 bytes. */
	int n, size = 128;
	char *p, *p2;
	va_list ap;

	p = malloc(size);
	if (p == NULL)
		return NULL;

	while (1) {
		/* Try to print in the allocated space. */
		va_start(ap, fmt);
		n = vsnprintf(p, size, fmt, ap);
		va_end(ap);

		/* If that worked, return the string. */
		if (n > -1 && n < size)
			return p;

		/* Else try again with more space. */
		if (n > -1)    /* glibc 2.1 */
			size = n+1; /* precisely what is needed */
		else           /* glibc 2.0 */
			size *= 2;  /* twice the old size */

		p2 = realloc(p, size);
		if (p2 == NULL) {
			free(p);
			return NULL;
		}
		p = p2;
	}
}

static void gen_empty(const char *file)
{
	char *fname;
	FILE *f;

	fname = make_message("%s/%s", output_dir, file);
	f = fopen(fname, "w");
	if (f == NULL)
		die("cannot create '%s': %s\n", fname, strerror(errno));
	fclose_or_error(f);
	free(fname);
}

static void gen_global(void)
{
	char *fname;
	FILE *f;
	const char *mine;
	const char *shaft;

	fname = make_message("%s/global", output_dir);
	f = fopen(fname, "w");
	if (unlikely(f == NULL))
		die("cannot create '%s': %s\n", fname, strerror(errno));

	mine = __ar2c_config_get(&config, "nazwa kopalni");
	if (!mine)
		mine = "unknown";
	shaft = __ar2c_config_get(&config, "nazwa szybu");
	if (!shaft)
		shaft = "unknown";

	fprintf(f, "mine \"%s\"\n", mine);
	fprintf(f, "shaft \"%s\"\n", shaft);
	fprintf(f, "serial_number 0\n");
	fclose_or_error(f);
	free(fname);
}

static void gen_magic(void)
{
	char *fname;
	FILE *f;

	fname = make_message("%s/magic", output_dir);
	f = fopen(fname, "w");
	if (unlikely(f == NULL))
		die("cannot create '%s': %s\n", fname, strerror(errno));

	fprintf(f, "AR3C-CONFIG-1.0\n");
	fclose_or_error(f);
	free(fname);
}

static int gen_binary_input(FILE *f, unsigned int id, const char *symbol,
		const char *caption, const char *filter)
{
	fprintf(f, "binary_input\n");
	fprintf(f, "\tphysical\t%d\n", id);
	fprintf(f, "\tsymbol\t\t\"%s\"\n", symbol);
	fprintf(f, "\tcaption\t\t\"%s\"\n", caption ? caption : symbol);
	fprintf(f, "\tid\t\t\"b%d\"\n", id);
	if (filter)
		fprintf(f, "\tfilter_id\t\t\"%s\"\n", filter);
	fprintf(f, "\n");
	return 0;
}

static void gen_binary(void)
{
	unsigned int i;
	char *fname;
	FILE *f;
	const char *symbol;
	const char *caption;

	fname = make_message("%s/binary", output_dir);
	f = fopen(fname, "w");
	if (f == NULL)
		die("cannot create '%s': %s\n", fname, strerror(errno));

	for (i = 0; i < 192; i++) {
		char key[64];

		snprintf(key, sizeof(key), "dwustan %d", i);
		symbol = __ar2c_config_get(&config, key);

		if (symbol == NULL)
			continue;

		snprintf(key, sizeof(key), "caption %d", i);
		caption = __ar2c_config_get(&config, key);

		gen_binary_input(f, i, symbol, caption, NULL);
	}
	gen_binary_input(f, 514, "DW", "Dzwon Wykonawczy", "bell");
	gen_binary_input(f, 515, "AW", "Akustyczny Wykonawczy", "bell");
	gen_binary_input(f, 516, "DP", "Dzwon Porozumiewawczy", "bell");
	gen_binary_input(f, 517, "AP", "Akustyczny Porozumiewawczy", "bell");
	fclose_or_error(f);
	free(fname);
}

static int gen_analog_input(FILE *f, unsigned int id, const char *symbol,
		const char *unit, double uv, const char *fmt, ...)
{
	va_list ap;

	fprintf(f, "analog_input\n");
	fprintf(f, "\tphysical\t%d\n", id);
	fprintf(f, "\tsymbol\t\t%s\n", symbol);
	fprintf(f, "\tid\t\t\"a%d\"\n", id);
	fprintf(f, "\tunit_value\t%f\n", 1285.0 / uv);
	fprintf(f, "\tunit\t\t%s\n", unit);
	fprintf(f, "\tmin_value\t%f\n", -25.6 * uv);
	fprintf(f, "\tmax_value\t%f\n", 25.6 * uv);
	fprintf(f, "\tplot_color\tHSV(%d, 255, 255)\n", 360 * id / 16);
	va_start(ap, fmt);
	if (fmt)
		vfprintf(f, fmt, ap);
	va_end(ap);
	fprintf(f, "\n");
	return 0;
}

static const char position_config[] =
"analog_input\n"
"	physical 16\n"
"	id \"position\"\n"
"	unit_value 100.0\n"
"	unit \"m\"\n"
"	min_value -1300.0\n"
"	max_value 100.0\n"
"	plot_color HSV(0, 0, 60)\n"
"	level 0.0 label \"0 m\" color HSV(60, 255, 255)\n";

static void gen_position(FILE *f)
{
	const char *t;
	unsigned int i;

	t = __ar2c_config_get(&config, "szybowskaz");
	if (!t || !strcmp(t, "0"))
		return;

	fputs(position_config, f);
	for (i = 1; i < 13; i++) {
		fprintf(f, "\tlevel %.1f label \"%d m\" "
				"color HSV(90, 255, %d)\n",
				-100.0 * i, -100 * i, 250 - 10 * i);
	}

	fprintf(f, "\tposition\n\n");
}

static void gen_analogs(void)
{
	unsigned int i;
	char *fname;
	FILE *f;
	const char *symbol;
	const char *unit;
	const char *unitvalue;
	char *end;
	double uv;

	fname = make_message("%s/analogs", output_dir);
	f = fopen(fname, "w");
	if (f == NULL)
		die("cannot create '%s': %s\n", fname, strerror(errno));

	for (i = 0; i < 16; i++) {
		char key[64];

		snprintf(key, sizeof(key), "analog %d", i);
		symbol = __ar2c_config_get(&config, key);
		if (symbol == NULL)
			continue;

		snprintf(key, sizeof(key), "analog jedn %d", i);
		unit = __ar2c_config_get(&config, key);
		if (unit == NULL)
			unit = "";

		snprintf(key, sizeof(key), "analog skala %d", i);
		unitvalue = __ar2c_config_get(&config, key);
		uv = 1;
		if (unitvalue != NULL && *unitvalue != '\0') {
			uv = strtod(unitvalue, &end);
			if (*end != '\0') {
				error("analog %d: '%s' is not a number\n",
						unitvalue);
				uv = 1;
			}
		}

		/*
		 * Analog 0 is special - in AR-2c dataloggers it's always
		 * velocity
		 */
		if (i == 0) {
			gen_analog_input(f, i, symbol, unit, uv, "\tvelocity\n");
		} else {
			gen_analog_input(f, i, symbol, unit, uv, NULL);
		}
	}

	gen_position(f);
	fclose_or_error(f);
	free(fname);
}

int main(int argc, char **args)
{
	ar2c_config_init(&config);
	ar2c_config_set_error_handler(&config, config_error_handler);

	parse_command_line(argc, args);

	if (output_dir == NULL)
		die("output directory not defined\n");

	gen_analogs();
	gen_binary();
	gen_empty("diag");
	gen_global();
	gen_empty("hardware");
	gen_empty("logger");
	gen_magic();
	gen_empty("oki");
	gen_empty("visualizations");
	gen_empty("wydal");

	return EXIT_SUCCESS;
}
