/*
 * ar2c-lp-utils - ar2c.c
 * 	version 0.1.2
 * 
 * 	AR-2c Logger Protocol Client basic commands
 * 
 * Copyright (C) 2005  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id: ar2c.c,v 1.11 2006/09/18 18:22:05 krzysiek Exp $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <ar2c.h>
#include <inttypes.h>
#include <bcd.h>
#include <tread.h>
#include <termios.h>
#include <comm.h>
#include <ar2c_parser.h>
#include <netinet/in.h>

#define RSPD_NOT_CHECKED	(-1)
#define RSPD_NOT_SUPPORTED	(-2)

static uint8_t rspd_tab[255];
static int rspd_count = RSPD_NOT_CHECKED;
static const char *rspd_serial_names[] = {
	"undefined",			/* 0 */
	"9600",				/* 1 */
	"19200",			/* 2 */
	"38400",			/* 3 */
	"57600",			/* 4 */
	"115200",			/* 5 */
	"230400",			/* 6 */
	"460800",			/* 7 */
};

extern int cmd_timeout;

#define RSPD_SERIAL_NAMES_COUNT	8

static char dummy_buff[1024];

static int __input_resync(int ifd)
{
	int ret;
	unsigned int timeout;
	
	timeout = cmd_timeout;
	if (timeout > 500)
		timeout = 500;
	do {
		ret = t_read(ifd, dummy_buff, sizeof(dummy_buff), timeout);
		if (ret < 0)
			return -ret;
	} while (ret > 0);
	return ret;
}

static ssize_t write_all(int fd, const void *buf, size_t len)
{
	const char *t = buf;
	ssize_t ret, total = 0;

	while (total < len) {
		ret = write(fd, t + total, len - total);
		if (ret <= 0) {
			if (!total)
				return ret;
			break;
		}
		total += ret;
	}
	return total;
}

static void drain_input(int ifd)
{
	unsigned int i = 0;
	int ret;

	do {
		ret = t_read_nb(ifd, dummy_buff, sizeof(dummy_buff), 0);
		i++;
	} while ((ret > 0) && (i < 100));
}

/*
 * The Logger Protocol does not use request identifiers and in case
 * of any problems (transmission errors, timeouts, ...) we can
 * use some random data as response. To prevent this in case of any
 * problem resynchronization is performed. We need to be sure that
 * no request is pending and all data send by server is skipped.
 *
 * We cannot just don't send new requests and wait for any data from
 * server for specified amount of time. If the server is in monitor mode
 * it will never stop sending new data, so we must leave monitor mode.
 * We can do it by sending any command - currently we'are using 0xff 0xd2
 * command.
 */
int do_resync(int ifd, int ofd)
{
	static const char sync_seq[] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xd2 };
	unsigned int timeout;
	unsigned int i = 0;
	int ret;

	timeout = cmd_timeout;
	do {
		ret = write_all(ofd, sync_seq, sizeof(sync_seq));
		if (ret < sizeof(sync_seq))
			return -errno;

		ret = t_read(ifd, dummy_buff, 2, timeout);
		if (ret < 0)
			return -ret;

		/* check if server sends data */
		ret = t_read(ifd, dummy_buff, sizeof(dummy_buff), 100);
		if (ret < 0)
			return -ret;

		if (ret > 0) {
			drain_input(ifd);
			i = 0;
		} else {
			i++;
		}
	} while (i < 3);
	return 0;
}

static int ar2c_detect_rspd(int ifd, int ofd)
{
	uint8_t cmd[] = { 0xff, 0x0f, 0x13, 0x00, };
	int retry = 2;
	uint8_t len;

	do {
		if (retry != 2)
			do_resync(ifd, ofd);
		if (write(ofd, cmd, sizeof(cmd)) < sizeof(cmd))
			return -1;
		if (t_read(ifd, &len, 1, cmd_timeout) < 1)
			continue;
		if (t_read(ifd, rspd_tab, len, cmd_timeout) < len) {
			__input_resync(ifd);
			continue;
		}
		rspd_count = len;
		return 0;
	} while (--retry);
	rspd_count = RSPD_NOT_SUPPORTED;
	return -1;
}

int ar2c_spd_count(int ifd, int ofd)
{
	if (rspd_count == RSPD_NOT_CHECKED)
		ar2c_detect_rspd(ifd, ofd);
	return rspd_count;
}

const char *ar2c_spd_name(unsigned int id)
{
	int lid;

	if (id >= rspd_count)
		return NULL;
	lid = rspd_tab[id];
	if (lid >= RSPD_SERIAL_NAMES_COUNT)
		return "undefined";
	return rspd_serial_names[lid];
}

int ar2c_get_date(int ifd, int ofd, struct tm *tm)
{
	uint8_t cmd[] = { 0xff, 0xff, 0x96 };
	int retry = 4;
	uint8_t buff[5];

	do {
		if (retry != 4)
			do_resync(ifd, ofd);
		if (write(ofd, cmd, sizeof(cmd)) < sizeof(cmd))
			return -1;
		if (t_read(ifd, buff, 5, cmd_timeout) >= 5) {
			if ((tm->tm_year = bcd2bin(buff[0])) == BCD_ERR)
				continue;
			if ((tm->tm_mon = bcd2bin(buff[1])) == BCD_ERR)
				continue;
			if ((tm->tm_mday = bcd2bin(buff[2])) == BCD_ERR)
				continue;
			if ((tm->tm_hour = bcd2bin(buff[3])) == BCD_ERR)
				continue;
			if ((tm->tm_min = bcd2bin(buff[4])) == BCD_ERR)
				continue;
			tm->tm_year += 100;
			return 0;			
		}
	} while (--retry);
	return -1;
}

int ar2c_get_data_zone(int ifd, int ofd, struct tm *tm, struct tm *tm2)
{
	uint8_t cmd[] = { 0xff, 0xff, 0xb4 };
	int retry = 4;
	uint8_t buff[10];

	do {
		if (retry != 4)
			do_resync(ifd, ofd);
		if (write(ofd, cmd, sizeof(cmd)) < sizeof(cmd))
			return -1;
		if (t_read(ifd, buff, 10, cmd_timeout) >= 10) {
			if ((tm->tm_year = bcd2bin(buff[0])) == BCD_ERR)
				continue;
			if (((tm->tm_mon = bcd2bin(buff[1])) == BCD_ERR) ||
					!tm->tm_mon || (tm->tm_mon > 12))
				continue;
			if ((tm->tm_mday = bcd2bin(buff[2])) == BCD_ERR)
				continue;
			if ((tm->tm_hour = bcd2bin(buff[3])) == BCD_ERR)
				continue;
			if ((tm->tm_min = bcd2bin(buff[4])) == BCD_ERR)
				continue;
			if ((tm2->tm_year = bcd2bin(buff[5])) == BCD_ERR)
				continue;
			if (((tm2->tm_mon = bcd2bin(buff[6])) == BCD_ERR) ||
					!tm2->tm_mon || (tm2->tm_mon > 12))
				continue;
			if ((tm2->tm_mday = bcd2bin(buff[7])) == BCD_ERR)
				continue;
			if ((tm2->tm_hour = bcd2bin(buff[8])) == BCD_ERR)
				continue;
			if ((tm2->tm_min = bcd2bin(buff[9])) == BCD_ERR)
				continue;
			tm->tm_year += 100;
			tm2->tm_year += 100;
			tm->tm_sec = 0;
			tm2->tm_sec = 0;
			return 0;
		}
	} while (--retry);
	return -1;
}

int ar2c_get_id(int ifd, int ofd)
{
	uint8_t cmd[] = { 0xff, 0xd2 };
	int retry = 3, ret;
	uint8_t buff[2];

	do {
		if (write(ofd, cmd, sizeof(cmd)) < sizeof(cmd))
			return -1;
		ret = t_read(ifd, buff, 2, cmd_timeout);
		if (ret == 2)
			return (buff[0] << 8) + buff[1];
		do_resync(ifd, ofd);
	} while (--retry);
	return -1;
}

int ar2c_pstart(int ifd, int ofd, unsigned int scheme_id)
{
	struct {
		uint8_t magic;
		uint8_t cmd_id;
		uint8_t ecmd_id;
		uint8_t scheme_id;
	} cmd = { 
		.magic		= 0xff,
		.cmd_id		= 0x0f,
		.ecmd_id	= 0x21
	};
	int retry = 3, ret;
	uint8_t rret;

	if (scheme_id > 255)
		return -EINVAL;
	cmd.scheme_id = scheme_id;
	do {
		if (write(ofd, &cmd, sizeof(cmd)) < sizeof(cmd))
			return -1;
		ret = t_read(ifd, &rret, 1, cmd_timeout);
		if (ret == 1)
			return rret;
		do_resync(ifd, ofd);
	} while (--retry);
	return -1;
}

int ar2c_pstop(int ifd, int ofd)
{
	uint8_t cmd[] = { 0xff, 0x0f, 0x20 };
	int retry = 3, ret;
	uint8_t rret;

	do {
		if (write(ofd, cmd, sizeof(cmd)) < sizeof(cmd))
			return -1;
		ret = t_read(ifd, &rret, 1, cmd_timeout);
		if (ret == 1)
			return rret;
		do_resync(ifd, ofd);
	} while (--retry);
	return -1;
}

int ar2c_pmem(int ifd, int ofd, time_t beg, time_t end, int scheme_id)
{
	struct {
		uint8_t sync;
		uint8_t id;
		uint8_t eid;
		uint32_t beg;
		uint32_t end;
		uint8_t scheme_id;
	} __attribute__((packed)) cmd = {
		.sync		= 0xff,
		.id		= 0x0f,
		.eid		= 0x22,		
	};
	int retry = 3;
	uint8_t rret;
	ssize_t ret;
	
	cmd.beg = htonl(beg);
	cmd.end = htonl(end);
	cmd.scheme_id = scheme_id;
	
	do {
		if ((ret = write(ofd, &cmd, sizeof(cmd)) < sizeof(cmd)))
			return -errno;
		if (t_read(ifd, &rret, 1, cmd_timeout) >= 1)
			return -rret;
	} while (--retry);
	return -ETIMEDOUT;
}

int ar2c_set_speed(int fd, int speed)
{
	struct {
		uint8_t sync;
		uint8_t cmd;
		uint8_t subcmd;
		uint8_t speed;
	} cmd = {
		.sync		= 0xff,
		.cmd		= 0x0f,
		.subcmd		= 0x13,
	};
	int ret, i;
	uint8_t buff[16];
	speed_t new_speed;

	do_resync(fd, fd);
	if (speed > 0) {
		if (rspd_count == RSPD_NOT_CHECKED)
			ar2c_detect_rspd(fd, fd);
		if (rspd_count < 0)
			return -ENOSYS;
		ret = 1;
		for (i = 0; i < rspd_count; i++)
			if (rspd_tab[i] == speed) {
				ret = 0;
				break;
			}
		if (ret)
			return -EINVAL;
		new_speed = ar2cspeedtospeed(speed);
		cmd.speed = speed;
		if (speed > 7)
			return -EINVAL;
		write(fd, &cmd, sizeof(cmd));
		ret = t_read(fd, buff, 2, cmd_timeout);
		if (ret < 0)
			return -EIO;
		if (ret < 2)
			return -ENOSYS;
		if ((buff[0] == 0xde) && (buff[1] == 0xad))
			return -ENOSYS;
		if ((buff[0] != 0x5a) || (buff[1] != 0xa5))
			return -ENOSYS;
		serial_set_speed(fd, new_speed);
		usleep(250);
		do_resync(fd, fd);
	}
	return 0;
}

int itoar2cspeed(int speed)
{
	switch (speed) {
		case 9600: return 1;
		case 19200: return 2;
		case 38400: return 3;
		case 57600: return 4;
		case 115200: return 5;
		case 230400: return 6;
		case 460800: return 7;
	}
	return -EINVAL;
}

int speedtoar2cspeed(speed_t speed)
{
	switch (speed) {
		case B9600: return 1;
		case B19200: return 2;
		case B38400: return 3;
		case B57600: return 4;
		case B115200: return 5;
		case B230400: return 6;
		case B460800: return 7;
	}
	return -EINVAL;
}

int ar2cspeedtoi(int speed)
{
	switch (speed) {
		case 1: return 9600;
		case 2: return 19200;
		case 3: return 38400;
		case 4: return 57600;
		case 5: return 115200;
		case 6: return 230400;
		case 7: return 460800;
	}
	return -EINVAL;
}

speed_t ar2cspeedtospeed(int speed)
{
	switch (speed) {
		case 1: return B9600;
		case 2: return B19200;
		case 3: return B38400;
		case 4: return B57600;
		case 5: return B115200;
		case 6: return B230400;
		case 7: return B460800;
	}
	return -EINVAL;
}

int ar2c_lp_info(int ifd, int ofd, struct ar2c_lp_info *dst)
{
	uint8_t cmd[] = { 0xff, 0x0f, 0x15, 0x00 };
	int retry = 2, ret;

	while (retry-- > 0) {
		write(ofd, cmd, sizeof(cmd));
		ret = t_read(ifd, (char *) dst, sizeof(struct ar2c_lp_info),
				cmd_timeout);
		if (ret < sizeof(struct ar2c_lp_info)) {
			do_resync(ifd, ofd);
			continue;
		}
		return 0;
	}
	return 1;
}
