/*
 * ar2c-lp-utils - ar2cgetdata.c
 * 	version 0.1.2
 *
 * 	AR-2c LP getdata
 *
 * Copyright (C) 2005..2007  Krzysztof Mazur <krzysiek@podlesie.net>
 * Copyright (c) 2007        Pawel Bregula   <pawel@microster.pl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <comm.h>
#include <time.h>
#include <ar2c.h>
#include <signal.h>
#include <tread.h>
#include <inttypes.h>

#define TT_STDIN	0
#define TT_FILE		1
#define TT_TCP		2
#define TT_FIFO		3
#define TT_SERIAL	4

#define error(fmt...) fprintf(stderr, "error: " fmt)

static int targ_type = TT_STDIN;
static char *targ_opt = NULL;
static char *targ_opt2 = NULL;
static const char *port = "2000";
static int ifd, ofd;
static int verbose = 0;
static int rej_detect = 0;
static FILE *out;
static speed_t bspeed = B9600;
static int speed = 0, rspeed = 0;
static int data_zone = 0;
int cmd_timeout = 0;
static unsigned long get_bitmap = -1;

static void show_usage(char *name)
{
	fprintf(stderr, "ar2cgetdata " PACKAGE_VERSION ".\n"
"Copyright (C) 2005..2007  Krzysztof Mazur <krzysiek@podlesie.net>.\n"
"Copyright (C) 2007        Pawel Bregula   <pawel@microster.pl>.\n");

	fprintf(stderr,	"ar2cinfo.\n"
			"usage: %s [options]\n"
			"Options:\n"
			"      --bbitrate	base bitrate(default 9600bps)\n"
			"  -b, --bitrate	set bitrate\n"
			"  -d, --device		use selected device\n"
			"  -f, --fifo		use dual fifo(for wr)\n"
			"  -g, --getbitmap	set get bitmap\n"
			"  -h, --help		show this help\n"
			"  -O			try to detect version\n"
			"  -p, --port		remote port(2000)\n"
			"  -q, --quiet		be quiet\n"
			"      --rbitrate	return bitrate\n"
			"  -s, --host		remote host(localhost)\n"
			"  -t, --timeout	set timeout\n"
			"  -v, --verbose		be verbose\n"
			"  -V, --version		show version\n"
			"  -z, --datazone	zakres danych\n",
			(name == NULL) ? "ar2cinfo" : name);
	exit(0);
}

static void show_version(void)
{
	fprintf(stderr, "ar2cgetdata " PACKAGE_VERSION ".\n"
"Copyright (C) 2005..2007  Krzysztof Mazur <krzysiek@podlesie.net>.\n"
"Copyright (C) 2007        Pawel Bregula   <pawel@microster.pl>.\n\n");

	fprintf(stderr,	"This program comes with ABSOLUTELY NO WARRANTY; "
			"see COPYING for details.\n"
			"This is free software, and you are welcome to "
			"redistribute it\n"
			"under certain conditions; see COPYING for details.\n");
	exit(0);
}

static void parse_command_line(int argc, char **args)
{
	int c;
	int option_index = 0;

	static struct option long_options[] = {
		{"bbitrate", 1, 0, 1},
		{"bitrate", 1, 0, 'b'},
		{"datazone", 0, 0, 'z'},
		{"device", 1, 0, 'd'},
		{"fifo", 1, 0, 'f'},
		{"getbitmap", 1, 0, 'g'},
		{"help", 0, 0, 'h'},
		{"host", 1, 0, 's'},
		{"port", 1, 0, 'p'},
		{"quiet", 0, 0, 'q'},
		{"rbitrate", 1, 0, 2},
		{"timeout", 1, 0, 't'},
		{"verbose", 0, 0, 'v'},
		{"version", 0, 0, 'V'},
		{0, 0, 0, 0},
	};

	while (1) {
		c = getopt_long(argc, args, "b:d:f:g:p:hs:t:vqVOz",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
			case 1:
				bspeed = itospeed(atoi(optarg));
				if (bspeed < 0) {
					fprintf(stderr, "Unsupported initial "
							"speed setting.\n");
					exit(1);
				}
				if (rspeed == 0) {
					rspeed = speedtoar2cspeed(bspeed);
					if (rspeed < 0)
						rspeed = 0;
				}
				targ_type = TT_SERIAL;
				break;
			case 2:
				rspeed = itoar2cspeed(atoi(optarg));
				if (rspeed < 0) {
					fprintf(stderr, "Invalid return speed "
								"setting.");
					exit(1);
				}
				targ_type = TT_SERIAL;
				break;
			case 'b':
				speed = itoar2cspeed(atoi(optarg));
				if (speed < 0) {
					fprintf(stderr, "Invalid speed "
							"setting.");
					exit(1);
				}
				targ_type = TT_SERIAL;
				break;
			case 'd':
				targ_opt = optarg;
				if ((targ_type == TT_TCP) ||
						(targ_type == TT_STDIN))
					targ_type = TT_FILE;
				break;
			case 'f':
				targ_opt2 = optarg;
				targ_type = TT_FIFO;
				break;
			case 'g':
				get_bitmap = strtol(optarg, NULL, 0);
				break;
			case 'h':
				show_usage(args[0]);
				break;
			case 'p':
				port = strdup(optarg);
				targ_type = TT_TCP;
				break;
			case 'v':
				verbose++;
				break;
			case 'q':
				verbose--;
				break;
			case 's':
				targ_opt = optarg;
				targ_type = TT_TCP;
				break;
			case 't':
				cmd_timeout = atoi(optarg);
				break;
			case 'V':
				show_version();
				break;
			case 'O':
				rej_detect = 1;
				break;
			case 'z':
				data_zone = 1;
				break;
			default:
				fprintf(stderr, "Unknown option.\n");
				exit(1);
		}
	}
	if (optind < argc) {
		if (optind < argc)
			fprintf(stderr, "warning: unknown parameters "
					"detected.\n");
	}
}

static void parse_string(const char *output, uint8_t *msg, size_t len)
{
	uint8_t *tmp;

	if (len < 4)
		return;

	len -= 3;
	/*
	 * According to website spec strings are space padded, but
	 * it is $-terminated.
	 */
	for (tmp = msg + 3; (*tmp != '$') && (*tmp != '\0') && (len > 0);
			tmp++, len--);
	*tmp = '\0';

	fprintf(out, "par:2:%s:%s\n", output, msg + 3);
}

static void parse_bin_names(const char *output, uint8_t *msg, size_t len)
{
	unsigned int b_count = (msg[1] << 8) + msg[2];
	unsigned int i;

	if (len < 3)
		return;

	if (b_count + 3 > len)
		b_count = len - 3;

	for (i = 0; i < b_count - 1; i += 2)
		fprintf(out, "par:3:dwustan %d:%c%c\n", i >> 1, msg[i + 3],
				msg[i + 4]);
}

#define ARRAY_SIZE(x)	(sizeof(x) / sizeof(x[0]))

static void parse_bin_fun_table(const char *output, uint8_t *msg, size_t len)
{
	unsigned int b_count = (msg[1] << 8) + msg[2] - 1;
	unsigned int i;
	static const char *tabways[] = {
		"noop",
		"obs_alarmu",
		"obs_hamulca",
		"hamulec_bezp",
		"korekcja_k1",
		"korekcja_k2"
	};

	if (len < 4)
		return;

	if (b_count + 4 > len)
		b_count = len - 4;

	fprintf(out, "par:0:ilosc WC8:%d\n", msg[3]);
	for (i = 0; i < b_count; i++) {
		uint8_t tid = msg[i + 4];

		if (tid >= ARRAY_SIZE(tabways)) {
			fprintf(stderr, "invalid tabway id: %d\n", tid);
			exit(EXIT_FAILURE);
		}
		fprintf(out, "par:3:tabway %d:%s\n", i, tabways[msg[i + 4]]);
	}
	for (; i < 64; i++)
		fprintf(out, "par:3:tabway %d:noop\n", i);
}

static void parse_analog_names(const char *output, uint8_t *msg, size_t len)
{
	unsigned int a_count = (msg[1] << 8) + msg[2];
	unsigned int i;

	if (len < 3)
		return;

	if (a_count + 3 > len)
		a_count = len - 3;

	a_count /= 42;
	for (i = 0; i < a_count; i++)
		fprintf(out, "par:3:opis skali %d:\"%42.42s\"\n", i,
				&msg[42 * i + 3]);
}

static void parse_analog_map(const char *output, uint8_t *msg, size_t len)
{
	unsigned int b_count = (msg[1] << 8) + msg[2];
	unsigned int i;

	if (len < 3)
		return;

	if (b_count + 3 > len)
		b_count = len - 3;

	for (i = 0; i < b_count; i++)
		fprintf(out, "par:3:analog map %d:%d\n", i, msg[i + 3]);
}

static void parse_analog_divs(const char *output, uint8_t *msg, size_t len)
{
	unsigned int b_count = (msg[1] << 8) + msg[2];
	unsigned int i;

	if (len < 3)
		return;

	if (b_count + 3 > len)
		b_count = len - 3;

	for (i = 0; i < b_count; i ++)
		fprintf(out, "par:3:analog div %d:%d\n", i, msg[i + 3]);
}

static int get_int16le_unaligned(uint8_t *a)
{
	int16_t ret = a[0] + (a[1] << 8);

	return ret;
}

static void parse_analog_offset(const char *output, uint8_t *msg, size_t len)
{
	unsigned int b_count = (msg[1] << 8) + msg[2];
	unsigned int i;

	if (len < 3)
		return;

	if (b_count + 3 > len)
		b_count = len - 3;
	b_count /= 2;

	for (i = 0; i < b_count; i++) {
		fprintf(out, "par:3:analog offset %d:%hd\n", i,
				get_int16le_unaligned(msg + 3 + 2 * i));
	}
}

int get_int32le_unaligned(uint8_t *a)
{
	int32_t ret = a[0] + (a[1] << 8) + (a[2] << 16) + (a[3] << 24);

	return ret;
}

static void parse_vel_info(const char *output, uint8_t *msg, size_t len)
{
	if (len < 14)
		return;

	fprintf(out, "par:2:tacho:%d\n", msg[3]);
	fprintf(out, "par:0:szybowskaz:%d\n", msg[4]);
	fprintf(out, "par:0:typ impulsatora:%d\n", msg[5]);
	fprintf(out, "par:3:wa2:%d\n", msg[6]);
	fprintf(out, "par:3:pommax:%d\n", msg[7]);
	fprintf(out, "par:0:dzw. 22:%d\n", msg[8]);
	fprintf(out, "par:0:pkk1:%d\n", get_int32le_unaligned(msg + 9));
	fprintf(out, "par:0:pkk2:%d\n", get_int32le_unaligned(msg + 13));
}

static void parse_wc8_blocks(const char *output, uint8_t *msg, size_t len)
{
	unsigned int a_count = (msg[1] << 8) + msg[2];
	unsigned int i;

	if (len < 3)
		return;

	if (a_count + 3 > len)
		a_count = len - 3;

	a_count /= 5;
	for (i = 0; i < a_count; i++) {
		int id = msg[5 * i + 3];
		int io = msg[5 * i + 4] + (msg[5 * i + 5] << 8);
		int inputbase = msg[5 * i + 6];
		int length = msg[5 * i + 7];

		fprintf(out, "par:1:WC8_block_base %d:0%04xh\n", id, io);
		fprintf(out, "par:1:WC8_dest_offset %d:0%02xh\n", id, inputbase);
		fprintf(out, "par:1:WC8_block_lth %d:%dd\n", id, length);
	}
}

static void parse_dp_blocks(const char *output, uint8_t *msg, size_t len)
{
	unsigned int a_count = (msg[1] << 8) + msg[2];
	unsigned int i;

	if (len < 3)
		return;

	if (a_count + 3 > len)
		a_count = len - 3;

	msg += 4;
	a_count /= 26;
	for (i = 0; i < a_count; i++) {
		int id = i;
		int base = msg[26 * i + 0] + (msg[26 * i + 1] << 8);
		int offset = msg[26 * i + 2] + (msg[26 * i + 3] << 8);
		int inputbase = msg[26 * i + 4] + (msg[26 * i + 5] << 8);
		int length = msg[26 * i + 6] + (msg[26 * i + 7] << 8);
		int aoffset = msg[26 * i + 8] + (msg[26 * i + 9] << 8);
		int ainputbase = msg[26 * i + 10] + (msg[26 * i + 11] << 8);
		int alength = msg[26 * i + 12] + (msg[26 * i + 13] << 8);
		int hboffset = msg[26 * i + 14] + (msg[26 * i + 15] << 8);
		int hbinputbase = msg[26 * i + 16] + (msg[26 * i + 17] << 8);
		int hblength = msg[26 * i + 18] + (msg[26 * i + 19] << 8);
		int stat_adr = msg[26 * i + 20] + (msg[26 * i + 21] << 8);
		int ok = msg[26 * i + 22];
		int bad = msg[26 * i + 23];
		int chk_val = msg[26 * i + 24];

		fprintf(out, "par:1:card_bus_base %d:0%04xh\n", id, base);
		fprintf(out, "par:1:card_bus_sbin_offset %d:0%04xh\n", id,
				offset);
		fprintf(out, "par:1:card_bus_dest_offset %d:0%04xh\n", id,
				inputbase);
		fprintf(out, "par:1:card_bus_bin_lth %d:0%04xh\n", id,
				length);

		fprintf(out, "par:1:card_bus_sanal_offset %d:0%04xh\n", id,
				aoffset);
		fprintf(out, "par:1:card_bus_anal_offset %d:0%04xh\n", id,
				ainputbase);
		fprintf(out, "par:1:card_bus_anal_lth %d:0%04xh\n", id,
				alength);

		fprintf(out, "par:1:card_bus_sHB_offset %d:0%04xh\n", id,
				hboffset);
		fprintf(out, "par:1:card_bus_HB_offset %d:0%04xh\n", id,
				hbinputbase);
		fprintf(out, "par:1:card_bus_HB_lth %d:0%04xh\n", id,
				hblength);

		fprintf(out, "par:1:card_bus_stat_adr %d:0%04xh\n", id,
				stat_adr);
		fprintf(out, "par:1:card_bus_stat_ok %d:0%02xh\n", id, ok);
		fprintf(out, "par:1:card_bus_stat_bad %d:0%02xh\n", id, bad);
		fprintf(out, "par:1:card_bus_chk_value %d:0%02xh\n", id,
				chk_val);
	}
}

#if 0
static void parse_generic(const char *output, uint8_t *msg, size_t len)
{
	int j;

	fprintf(out, "par:%1d:%s:", 0, output);
	for (j = 0; j < len; j++)
		fprintf(out, "%02x", msg[j]);
	fprintf(out, "\n");
}
#endif

struct questions {
	char name[64];
	char command[4];
	char output[64];
	void (*parser)(const char *output, uint8_t *msg, size_t len);
};

static void dump(const char *msg, const void *buf, unsigned int len)
{
	const unsigned char *b = buf;
	unsigned int i;

	if (verbose < 1)
		return;

	fprintf(stderr, "%s:", msg);
	for (i = 0; i < len; i++)
		fprintf(stderr, " %02x", b[i]);
	fprintf(stderr, "\n");
}

static ssize_t write_all(int fd, void *buf, size_t len)
{
	char *t = buf;
	ssize_t ret, total = 0;

	dump("write", buf, len);
	while (total < len) {
		ret = write(fd, t + total, len - total);
		if (ret <= 0) {
			if (!total)
				return ret;
			break;
		}
		total += ret;
	}
	return total;
}

int getdata(void)
{
	struct questions q_data[] = {
		{"Get mine name", "\xFF\x0F\x01\x00", "nazwa kopalni",
			parse_string},
		{"Get shaft name", "\xFF\x0F\x02\x00", "nazwa szybu",
			parse_string},
		{"Get binary signals name", "\xFF\x0F\x03\x00",
			"nazwy dwustanow", parse_bin_names},
		{"Get analog signals name", "\xFF\x0F\x04\x00",
			"nazwy analogow", parse_analog_names},
		{"Get function table for binary signals", "\xFF\x0F\x05\x00",
			"tablica funkcji dwustanow", parse_bin_fun_table},
		{"Get map table for analog signals", "\xFF\x0F\x06\x00",
			"tablica mapowania analogow", parse_analog_map},
		{"Get velocity measure information", "\xFF\x0F\x07\x00",
			"informacja pomiaru predkosci", parse_vel_info},
		{"Get offset table for analog signals", "\xFF\x0F\x09\x00",
			"tablica offsetow analogow", parse_analog_offset},
		{"Get WC8 card blocks configuration", "\xFF\x0F\x0B\x00",
			"konfiguracja blokow WC8", parse_wc8_blocks},
		{"Get communication cards configuration", "\xFF\x0F\x0C\x00",
			"konfiguracja kart komunik.", parse_dp_blocks},
		{"Get analog dividers for OKI", "\xFF\x0F\x0D\x00",
			"dzielniki analogow dla OKI", parse_analog_divs}};
	int i;
	unsigned char buf[4096];
	unsigned int retry = 0;
	int ret;
	int len;

	for (i = 0; i < ARRAY_SIZE(q_data);) {
		retry++;
		if (!((1 << i) & get_bitmap) || (retry > 3)) {
			retry = 0;
			i++;
			continue;
		}

		if ((write_all(ofd, (void*)q_data[i].command, 4)) < 4) {
			error("Cannot write command: %s", q_data[i].name);
			do_resync(ifd, ofd);
			continue;
		}

		if ((ret = t_read(ifd, buf, 3, cmd_timeout)) < 3) {
			error("Cannot read after command: %s\n",
					q_data[i].name);
			do_resync(ifd, ofd);
			continue;
		}
		if (buf[0] != q_data[i].command[2]) {
			error("invalid magic in response, should be %02hhx\n",
					q_data[i].command[2]);
			dump("response", buf, ret);
			do_resync(ifd, ofd);
			continue;
		}
		len = (buf[1] * 256) + buf[2];
		if (len > sizeof(buf) - 3) {
			error("too large response length: %d\n", len);
			dump("response", buf, ret);
			do_resync(ifd, ofd);
			continue;
		}
		if ((ret = t_read(ifd, buf + 3, len, cmd_timeout)) < len) {
			error("partial read %d, expected %d\n", ret, len);
			do_resync(ifd, ofd);
			continue;
		}
		dump("read", buf, ret + 3);
		q_data[i].parser(q_data[i].output, buf, ret + 3);

		i++;
		retry = 0;
	}
	return 0;
}

int main(int argc, char **args)
{
	int ret;

	signal(SIGALRM, SIG_IGN);
	parse_command_line(argc, args);
	switch (targ_type) {
		case TT_STDIN:
			ifd = 0;
			ofd = 1;
			out = stderr;
			cmd_timeout = cmd_timeout ? : 2000;
			break;
		case TT_FILE:
			if (targ_opt == NULL) {
				fprintf(stderr, "fatal: null targ_opt.\n");
				exit(1);
			}
			if ((ifd = open(targ_opt, O_RDWR)) < 0) {
				fprintf(stderr, "error: Can't open %s:%s\n",
						targ_opt, strerror(errno));
				exit(1);
			}
			ofd = ifd;
			out = stdout;
			cmd_timeout = cmd_timeout ? : 2000;
			break;
		case TT_TCP:
			if (targ_opt == NULL) {
				fprintf(stderr, "fatal: null targ_opt.\n");
				exit(1);
			}
			if ((ifd = pcomm_open(targ_opt, port)) < 0) {
				fprintf(stderr, "error: Can't connect to %s\n",
						targ_opt);
				exit(1);
			}
			ofd = ifd;
			out = stdout;
			cmd_timeout = cmd_timeout ? : 10000;
			break;
		case TT_FIFO:
			if (targ_opt == NULL) {
				fprintf(stderr, "fatal: null targ_opt.\n");
				exit(1);
			}
			if (targ_opt2 == NULL) {
				fprintf(stderr, "fatal: null targ_opt2.\n");
				exit(1);
			}
			/* order is important */
			if ((ofd = open(targ_opt2, O_WRONLY)) == -1) {
				fprintf(stderr, "fatal: Can't open wr_fifo.\n");
				exit(1);
			}
			if ((ifd = open(targ_opt, O_RDONLY)) == -1) {
				fprintf(stderr, "fatal: Can't open rd_fifo.\n");
				exit(1);
			}
			out = stdout;
			cmd_timeout = cmd_timeout ? : 500;
			break;
		case TT_SERIAL:
			if (targ_opt == NULL) {
				fprintf(stderr, "fatal: null targ_opt.\n");
				exit(1);
			}
			if ((ifd = open(targ_opt, O_RDWR)) < 0) {
				fprintf(stderr, "error: Can't open %s:%s\n",
						targ_opt, strerror(errno));
				exit(1);
			}
			ofd = ifd;
			serial_set_speed(ifd, bspeed);
			if (ar2cspeedtospeed(speed) != bspeed)
				if (ar2c_set_speed(ifd, speed)) {
					fprintf(stderr, "Setting serial speed "
							"error\n");
					exit(1);
				}
			out = stdout;
			cmd_timeout = cmd_timeout ? : 500;
			break;
		default:
			fprintf(stderr, "fatal: Internal error.\n");
			exit(1);
	}

	ret = do_resync(ifd, ofd);
	if (ret) {
		fprintf(stderr, "resync error: %s\n", strerror(-ret));
		exit(EXIT_FAILURE);
	}

	if ((ret = ar2c_get_id(ifd, ofd)) < 0)
		fprintf(stderr, "error: Can't read remote rej id.\n");
	else
		fprintf(out, "par:0:numer rejestratora:%d\n", ret);

	if (getdata() < 0)
		fprintf(stderr, "error: Cannot get data.\n");

	if (speed) {
		if (rspeed == 0) {
			rspeed = speedtoar2cspeed(bspeed);
			if (rspeed < 0) {
				fprintf(stderr, "Unable to return to initial "
						"serial speed.\n");
				exit(1);
			}
		}
		if ((speed != rspeed) && ar2c_set_speed(ifd, rspeed)) {
			fprintf(stderr, "Setting serial speed error\n");
			exit(1);
		}
	}
	return 0;
}
