/*
 * ar2c-par-utils - ar2cpar.c
 * Copyright (C) 2008  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Fundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ar2cpar.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>

static struct ar2c_param *ar2c_param_alloc(void)
{
	return malloc(sizeof(struct ar2c_param));
}

static void ar2c_param_free(struct ar2c_param *p)
{
	free(p);
}

static int ar2c_default_error_handler(struct ar2c_config *c, int priority,
		int line, const char *fmt, ...)
{
	return priority <= AR2C_CONFIG_ERR;
}

int ar2c_config_init(struct ar2c_config *c)
{
	INIT_LIST_HEAD(&c->params);
	c->error = ar2c_default_error_handler;
	return 0;
}

int ar2c_config_set_error_handler(struct ar2c_config *c,
		ar2c_config_error_t handler)
{
	c->error = handler;
	return 0;
}

int ar2c_config_clear(struct ar2c_config *c)
{
	struct ar2c_param *p, *p2;

	list_for_each_entry_safe(p, p2, &c->params, list) {
		list_del(&p->list);
		ar2c_param_free(p);
	}
	return 0;
}

int ar2c_config_destroy(struct ar2c_config *c)
{
	return ar2c_config_clear(c);
}

struct ar2c_param *ar2c_config_lookup(struct ar2c_config *c,
		const char *name)
{
	struct ar2c_param *p;

	list_for_each_entry(p, &c->params, list)
		if (!strcmp(p->name, name))
			return p;

	return NULL;
}

const char *__ar2c_config_get(struct ar2c_config *c, const char *name)
{
	struct ar2c_param *p = ar2c_config_lookup(c, name);

	if (p == NULL)
		return NULL;

	return p->value;
}

int ar2c_config_get(struct ar2c_config *c, const char *name,
		char **value, size_t *len)
{
	const char *v = __ar2c_config_get(c, name);
	size_t vlen;

	if (v == NULL)
		return -ENOENT;
	vlen = strlen(v) + 1;
	if (*len < vlen) {
		*value = realloc(*value, vlen);
		if (*value == NULL)
			return -ENOMEM;
		*len = vlen;
	}
	memcpy(*value, v, vlen);
	return vlen - 1;
}

int __ar2c_config_set(struct ar2c_config *c, const char *name,
		const char *value)
{
	struct ar2c_param *p = ar2c_config_lookup(c, name);

	if (p == NULL)
		return -ENOENT;

	if (p->value != NULL)
		free(p->value);

	p->value = strdup(value);
	if (p->value == NULL)
		return -ENOMEM;

	return 0;
}

int ar2c_config_set(struct ar2c_config *c, const char *name,
		const char *value)
{
	int ret;

	ret = __ar2c_config_set(c, name, value);
	if (ret == -ENOENT) {
		struct ar2c_param *p = ar2c_param_alloc();
		
		ret = -ENOMEM;
		if (p != NULL) {
			p->name = strdup(name);
			p->value = strdup(value);
			list_add_tail(&p->list, &c->params);
			ret = 0;
		}
	}
	return ret;
}

int __ar2c_config_set_all(struct ar2c_config *c, const char *name,
		const char *value, const char *source, int prio)
{
	struct ar2c_param *p = ar2c_config_lookup(c, name);

	if (p == NULL)
		return -ENOENT;

	if (p->value != NULL)
		free(p->value);
	if (p->source != NULL)
		free(p->source);

	p->value = strdup(value);
	if (p->value == NULL)
		return -ENOMEM;
	
	p->source = strdup(source);
	if (p->value == NULL) {
		free(p->value);
		return -ENOMEM;
	}

	p->prio = prio;
	return 0;
}

int ar2c_config_set_all(struct ar2c_config *c, const char *name,
		const char *value, const char *source, int prio)
{
	struct ar2c_param *p;
	int ret;

	ret = __ar2c_config_set_all(c, name, value, source, prio);
	if (ret == -ENOENT) {
		p = ar2c_param_alloc();
		
		ret = -ENOMEM;
		if (p != NULL) {
			p->name = strdup(name);
			if (p->name == NULL)
				goto err;
			p->value = strdup(value);
			if (p->name == NULL)
				goto err_name;
			p->source = strdup(source);
			if (p->name == NULL)
				goto err_value;
			p->prio = prio;
			list_add_tail(&p->list, &c->params);
			ret = 0;
		}
	}
	return ret;

err_value:
	free(p->value);
err_name:	
	free(p->name);
err:
	free(p);
	return ret;
}

int ar2c_config_set_all_fmt(struct ar2c_config *c, const char *name,
		const char *source, int prio, const char *fmt, ...)
{
	size_t size;
	char *value = NULL;
	char *p;
	va_list ap;
	int n;
	int ret;

	size = 256;
	value = malloc(size);
	ret = -ENOMEM;
	if (value == NULL)
		goto out;

	while (1) {
		va_start(ap, fmt);
		n = vsnprintf(value, size, fmt, ap);
		va_end(ap);

		if (n > -1 && n < size)
			break;

		if (n > -1)
			size = n + 1;
		else
			size *= 2;

		p = realloc(value, size);
		if (p == NULL)
			goto out_free;
	}

	ret = ar2c_config_set_all(c, name, value, source, prio);
out_free:
	free(value);
out:
	return ret;
}

int ar2c_config_rm(struct ar2c_config *c, const char *name)
{
	struct ar2c_param *p = ar2c_config_lookup(c, name);

	if (!p)
		return -ENOENT;

	list_del(&p->list);
	free(p->value);
	free(p->name);
	free(p);
	return 0;
}

#define log_error(fmt...) c->error(c, AR2C_CONFIG_ERR, lines, fmt)

int ar2c_config_load(struct ar2c_config *c, FILE *f)
{
	char *line = NULL, *source, *prio, *name, *value, *endptr;
	size_t len = 0;
	ssize_t rlen;
	int lines = 0;
	int ret = 0;
	int prio_value;
	
	while ((rlen = getline(&line, &len, f)) != -1) {
		lines++;
		line[rlen - 1] = '\0';
		source = line;
		prio = strchr(line, ':');
		if (prio == NULL) {
			ret = -EINVAL;
			if (log_error("missing ':' after source\n"))
				break;
			continue;
		}
		*(prio++) = '\0';
		name = strchr(prio, ':');
		if (name == NULL) {
			ret = -EINVAL;
			if (log_error("missing ':' after name\n"))
				break;
			continue;
		}
		*(name++) = '\0';
		value = strchr(name, ':');
		if (value == NULL) {
			ret = -EINVAL;
			if (log_error("missing ':' after value\n"))
				break;
			continue;
		}
		*(value++) = '\0';

		prio_value = strtol(prio, &endptr, 0);
		if (*endptr != '\0') {
			ret = -EINVAL;
			if (log_error("invalid priority '%s'\n", prio))
				break;
			continue;
		}
		ar2c_config_set_all(c, name, value, source, prio_value);
	}
	if (line != NULL)
		free(line);
	return ret;
}

#undef log_error

int ar2c_config_store(struct ar2c_config *c, FILE *f)
{
	struct ar2c_param *p;
	int ret;

	ar2c_config_for_each_param(p, c) {
		ret = fprintf(f, "%s:%d:%s:%s\n", p->source, p->prio, p->name,
				p->value);
		if (ret < 0)
			return errno;
	}
	return 0;
}
