/*
 * ar2c-lp-utils - ar2cputdata.c
 *
 * 	AR-2c LP putdata
 *
 * Copyright (C) 2008  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <comm.h>
#include <time.h>
#include <ar2c.h>
#include <signal.h>
#include <tread.h>
#include <inttypes.h>
#include <ar2cpar.h>
#include <ar2cparproto.h>
#include <crc16.h>
#include <ctype.h>

#define TT_STDIN	0
#define TT_FILE		1
#define TT_TCP		2
#define TT_FIFO		3
#define TT_SERIAL	4

#ifdef HAVE_ENDIAN_H
#include <endian.h>
#endif

#ifdef HAVE_BYTESWAP_H
#include <byteswap.h>
#endif

#if BYTE_ORDER == BIG_ENDIAN

#define cpu_to_be16(x)		(x)
#define cpu_to_be32(x)		(x)
#define cpu_to_le16(x)		(bswap_16(x))
#define cpu_to_le32(x)		(bswap_32(x))

#define be16_to_cpu(x)		(x)
#define be32_to_cpu(x)		(x)
#define le16_to_cpu(x)		(bswap_16(x))
#define le32_to_cpu(x)		(bswap_32(x))

#else

#if BYTE_ORDER == LITTLE_ENDIAN

#define cpu_to_be16(x)		(bswap_16(x))
#define cpu_to_be32(x)		(bswap_32(x))
#define cpu_to_le16(x)		(x)
#define cpu_to_le32(x)		(x)

#define be16_to_cpu(x)		(bswap_16(x))
#define be32_to_cpu(x)		(bswap_32(x))
#define le16_to_cpu(x)		(x)
#define le32_to_cpu(x)		(x)

#else

#error "Only little and big endian systems are supported."

#endif
#endif

#define get_unaligned(ptr)						\
	({								\
		struct __str {						\
			typeof(*(ptr)) __v;				\
	 	} __attribute__((packed)) *__p				\
	 		= (struct __str *) (ptr);			\
		__p->__v;						\
	})

#define put_unaligned(val, ptr)						\
	do {								\
		struct __str {						\
			typeof(*(ptr)) __v;				\
		} __attribute__((packed)) *__p				\
			= (struct __str *) (ptr);			\
		__p->__v = (val);					\
	} while(0)

static int targ_type = TT_STDIN;
static char *targ_opt = NULL;
static char *targ_opt2 = NULL;
static const char *port = "2000";
static int ifd, ofd;
static int verbose = 0;
static FILE *out;
static speed_t bspeed = B9600;
static int speed = 0, rspeed = 0;
int cmd_timeout = 0;
static int synchronize = 1;

static void show_usage(char *name)
{
	fprintf(stderr, "ar2cputdata " PACKAGE_VERSION ".\n"
"Copyright (C) 2008  Krzysztof Mazur <krzysiek@podlesie.net>.\n");

	fprintf(stderr,	"ar2cputdata.\n"
			"usage: %s [options]\n"
			"Options:\n"
			"      --bbitrate	base bitrate(default 9600bps)\n"
			"  -a, --no-sync\t	don't use 0xff 0xd2 sync\n"
			"  -b, --bitrate\t	set bitrate\n"
			"  -d, --device		use selected device\n"
			"  -f, --fifo		use dual fifo(for wr)\n"
			"  -h, --help		show this help\n"
			"  -p, --port		remote port(2000)\n"
			"  -q, --quiet		be quiet\n"
			"      --rbitrate	return bitrate\n"
			"  -s, --host		remote host(localhost)\n"
			"  -t, --timeout\t	set timeout\n"
			"  -v, --verbose\t	be verbose\n"
			"  -V, --version\t	show version\n",
			(name == NULL) ? "ar2cputdata" : name);
	exit(EXIT_SUCCESS);
}

static void show_version(void)
{
	fprintf(stderr, "ar2cputdata " PACKAGE_VERSION ".\n"
"Copyright (C) 2008  Krzysztof Mazur <krzysiek@podlesie.net>.\n\n");

	fprintf(stderr,	"This program comes with ABSOLUTELY NO WARRANTY; "
			"see COPYING for details.\n"
			"This is free software, and you are welcome to "
			"redistribute it\n"
			"under certain conditions; see COPYING for details.\n");
	exit(EXIT_SUCCESS);
}

static void parse_command_line(int argc, char **args)
{
	int c;
	int option_index = 0;

	static struct option long_options[] = {
		{"bbitrate", 1, 0, 1},
		{"bitrate", 1, 0, 'b'},
		{"device", 1, 0, 'd'},
		{"fifo", 1, 0, 'f'},
		{"help", 0, 0, 'h'},
		{"host", 1, 0, 's'},
		{"no-sync", 0, 0, 'a'},
		{"port", 1, 0, 'p'},
		{"quiet", 0, 0, 'q'},
		{"rbitrate", 1, 0, 2},
		{"timeout", 1, 0, 't'},
		{"verbose", 0, 0, 'v'},
		{"version", 0, 0, 'V'},
		{0, 0, 0, 0},
	};

	while (1) {
		c = getopt_long(argc, args, "ab:d:f:g:p:hs:t:vqV",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 1:
			bspeed = itospeed(atoi(optarg));
			if (bspeed < 0) {
				fprintf(stderr, "Unsupported initial "
						"speed setting.\n");
				exit(EXIT_FAILURE);
			}
			if (rspeed == 0) {
				rspeed = speedtoar2cspeed(bspeed);
				if (rspeed < 0)
					rspeed = 0;
			}
			targ_type = TT_SERIAL;
			break;
		case 2:
			rspeed = itoar2cspeed(atoi(optarg));
			if (rspeed < 0) {
				fprintf(stderr, "Invalid return speed "
							"setting.");
				exit(EXIT_FAILURE);
			}
			targ_type = TT_SERIAL;
			break;
		case 'a':
			synchronize = 0;
			break;
		case 'b':
			speed = itoar2cspeed(atoi(optarg));
			if (speed < 0) {
				fprintf(stderr, "Invalid speed setting.");
				exit(EXIT_FAILURE);
			}
			targ_type = TT_SERIAL;
			break;
		case 'd':
			targ_opt = optarg;
			if ((targ_type == TT_TCP) || (targ_type == TT_STDIN))
				targ_type = TT_FILE;
			break;
		case 'f':
			targ_opt2 = optarg;
			targ_type = TT_FIFO;
			break;
		case 'h':
			show_usage(args[0]);
			break;
		case 'p':
			port = strdup(optarg);
			targ_type = TT_TCP;
			break;
		case 'v':
			verbose++;
			break;
		case 'q':
			verbose--;
			break;
		case 's':
			targ_opt = optarg;
			targ_type = TT_TCP;
			break;
		case 't':
			cmd_timeout = atoi(optarg);
			break;
		case 'V':
			show_version();
			break;
		default:
			fprintf(stderr, "Unknown option.\n");
			exit(EXIT_FAILURE);
		}
	}
	if (optind < argc)
		fprintf(stderr, "warning: unknown parameters detected.\n");
}

#define log_warning(fmt...)	fprintf(stderr, fmt)

static inline void lp_par_checksum(struct lp_par_pkt *p)
{
	uint16_t crc = crc16(0xffff, (void *) p->payload, p->plen - 2);

	p->payload[p->plen - 2] = crc >> 8;
	p->payload[p->plen - 1] = crc & 0xff;
}

static ssize_t write_all(int fd, void *buf, size_t len)
{
	char *t = buf;
	ssize_t ret, total = 0;

	while (total < len) {
		ret = write(fd, t + total, len - total);
		if (ret <= 0) {
			if (!total)
				return ret;
			break;
		}
		total += ret;
	}
	return total;
}

static int __ar2c_wait(void)
{
	static char __sync_seq[] = { 0xff, 0xd2, };
	int ret;
	char buf[2];

	do {
		ret = write_all(ofd, __sync_seq, sizeof(__sync_seq));
		if (ret < sizeof(__sync_seq))
			return -errno;

		ret = t_read(ifd, buf, 2, 300);
		if (ret < 0)
			return -errno;
	} while (ret < 2);
	return 0;
}

static void ar2c_wait(void)
{
	if (synchronize)
		__ar2c_wait();
}

static int put_nazwa_kopalni(void *priv, const char *name, const char *value)
{
	struct lp_par_pkt *p;
	size_t value_len = strlen(value);
	int p_len = sizeof(*p) + value_len + LP_PAR_CRC_LEN;
	ssize_t ret;

	p = malloc(sizeof(*p) + p_len);
	if (p == NULL)
		return -ENOMEM;

	p->sync = LP_PAR_SYNC;
	p->cmd = LP_PAR_CMD;
	p->param_id = LP_PARID_NAZWA_KOPALNI;
	p->plen = value_len + LP_PAR_CRC_LEN;
	memcpy(p->payload, value, value_len);
	lp_par_checksum(p);
	ar2c_wait();
	ret = write_all(ofd, p, p_len);
	if (ret != p_len) {
		if (ret < 0) {
			perror("write");
			return -errno;
		}
		fprintf(stderr, "can't write.\n");
		return -EINVAL;
	}

	free(p);
	return 0;
}

static int put_nazwa_szybu(void *priv, const char *name, const char *value)
{
	struct lp_par_pkt *p;
	size_t value_len = strlen(value);
	int p_len = sizeof(*p) + value_len + LP_PAR_CRC_LEN;
	ssize_t ret;

	p = malloc(sizeof(*p) + p_len);
	if (p == NULL)
		return -ENOMEM;

	p->sync = LP_PAR_SYNC;
	p->cmd = LP_PAR_CMD;
	p->param_id = LP_PARID_NAZWA_SZYBU;
	p->plen = value_len + LP_PAR_CRC_LEN;
	memcpy(p->payload, value, value_len);
	lp_par_checksum(p);
	ar2c_wait();
	ret = write_all(ofd, p, p_len);
	if (ret != p_len) {
		if (ret < 0) {
			perror("write");
			return -errno;
		}
		fprintf(stderr, "can't write.\n");
		return -EINVAL;
	}

	free(p);
	return 0;
}

static int send_binary_names(struct ar2c_config *c)
{
	int i, j;
	struct lp_par_pkt *p;
	size_t p_len = sizeof(*p) + 1 + LP_PAR_CRC_LEN
		+ LP_PAR_BNAME_LENGTH * LP_PAR_BNAMES_PER_CARD;
	char *value = NULL;
	size_t v_len = 0;
	int ret;

	p = malloc(p_len);
	if (p == NULL)
		return -ENOMEM;
	p->sync = LP_PAR_SYNC;
	p->cmd = LP_PAR_CMD;
	p->param_id = LP_PARID_BNAMES;
	p->plen = p_len - sizeof(*p);

	for (i = 0; i <= LP_PAR_BINARY_CARD_MAX; i++) {
		p->payload[0] = i;

		for (j = 0; j < LP_PAR_BNAMES_PER_CARD; j++) {
			char buf[64];

			snprintf(buf, 64, "dwustan %d",
					i * LP_PAR_BNAMES_PER_CARD + j);
			ret = ar2c_config_get(c, buf, &value, &v_len);
			if (ret == -ENOENT)
				break;
			if (ret < 0) {
				fprintf(stderr, "ar2c_config_get: %s\n",
						strerror(-ret));
				break;
			}
			if (ret != 2) {
				fprintf(stderr, "binary signal %d have "
						"incorrect length\n",
						i * LP_PAR_BNAMES_PER_CARD + j);
				break;
			}
			memcpy(p->payload + 1 + j * LP_PAR_BNAME_LENGTH,
					value, LP_PAR_BNAME_LENGTH);
		}
		if (j == LP_PAR_BNAMES_PER_CARD) {
			lp_par_checksum(p);
			ar2c_wait();
			ret = write_all(ofd, p, p_len);
			if (ret != p_len) {
				if (ret < 0) {
					perror("write");
					return -errno;
				}
				fprintf(stderr, "can't write.\n");
				return -EINVAL;
			}
		}
	}
	return 0;
}

static int send_analog_descriptions(struct ar2c_config *c)
{
	int i;
	struct lp_par_pkt *p;
	size_t p_len = sizeof(*p) + 1 + LP_PAR_CRC_LEN + LP_PAR_ADESC_LEN;
	char *value = NULL;
	size_t v_len = 0;
	int ret;

	p = malloc(p_len);
	if (p == NULL)
		return -ENOMEM;
	p->sync = LP_PAR_SYNC;
	p->cmd = LP_PAR_CMD;
	p->param_id = LP_PARID_ADESC;
	p->plen = p_len - sizeof(*p);

	for (i = 0; i <= LP_PAR_ADESC_PAGE_MAX; i++) {
		char buf[64];

		p->payload[0] = i;
		snprintf(buf, 64, "opis skali %d", i);
		ret = ar2c_config_get(c, buf, &value, &v_len);
		if (ret == -ENOENT)
			break;
		if (ret < 0) {
			fprintf(stderr, "ar2c_config_get: %s\n",
					strerror(-ret));
			break;
		}
		if (ret != LP_PAR_ADESC_LEN + 2) {
			fprintf(stderr, "analog description %d have "
					"incorrect length\n", i);
			break;
		}
		if ((value[0] != '"') || (value[LP_PAR_ADESC_LEN + 1] != '"')) {
			fprintf(stderr, "analog description %d is "
					"incorrect.\n", i);
			break;
		}
		memcpy(p->payload + 1, value + 1, LP_PAR_ADESC_LEN);
		lp_par_checksum(p);
		ar2c_wait();
		ret = write_all(ofd, p, p_len);
		if (ret != p_len) {
			if (ret < 0) {
				perror("write");
				return -errno;
			}
			fprintf(stderr, "can't write.\n");
			return -EINVAL;
		}
	}
	return 0;
}

#define ARRAY_SIZE(x)	(sizeof(x) / sizeof(x[0]))

static int send_binary_functions(struct ar2c_config *c)
{
	int i, j;
	struct lp_par_pkt *p;
	size_t p_len = sizeof(*p) + 1 + LP_PAR_CRC_LEN
		+ LP_PAR_BFUNCTIONS_COUNT;
	char *value = NULL;
	size_t v_len = 0;
	int ret;
	static const char *tabways[] = {
		"noop",
		"obs_alarmu",
		"obs_hamulca",
		"hamulec_bezp",
		"korekcja_k1",
		"korekcja_k2"
	};

	p = malloc(p_len);
	if (p == NULL)
		return -ENOMEM;
	p->sync = LP_PAR_SYNC;
	p->cmd = LP_PAR_CMD;
	p->param_id = LP_PARID_BFUNCTIONS;
	p->plen = p_len - sizeof(*p);

	ret = ar2c_config_get(c, "ilosc WC8", &value, &v_len);
	if (ret == -ENOENT)
		return ret;
	if (ret < 0) {
		fprintf(stderr, "ar2c_config_get: %s\n", strerror(-ret));
		return ret;
	}

	p->payload[0] = atoi(value);
	for (i = 0; i < LP_PAR_BFUNCTIONS_COUNT; i++) {
		char buf[64];
		uint8_t fun = 0;

		snprintf(buf, 64, "tabway %d", i);
		ret = ar2c_config_get(c, buf, &value, &v_len);
		if (ret == -ENOENT)
			break;
		if (ret < 0) {
			fprintf(stderr, "ar2c_config_get: %s\n",
					strerror(-ret));
			break;
		}

		for (j = 0; j < ARRAY_SIZE(tabways); j++)
			if (!strcmp(value, tabways[j])) {
				fun = j;
				break;
			}
		if (j == ARRAY_SIZE(tabways)) {
			/*
			 * Some tabways are noops in kz.
			 */
			if (strcmp(value, "napiecie_sterujace")
					&& strcmp(value, "zmiana_ark2")) {
				fprintf(stderr, "unknown tabway: %s\n", value);
				break;
			}
		}

		p->payload[1 + i] = fun;
	}
	if (i == LP_PAR_BFUNCTIONS_COUNT) {
		lp_par_checksum(p);
		ar2c_wait();
		ret = write_all(ofd, p, p_len);
		if (ret != p_len) {
			if (ret < 0) {
				perror("write");
				return -errno;
			}
			fprintf(stderr, "can't write.\n");
			return -EINVAL;
		}
		return 0;
	} else if (verbose > 0)
		fprintf(stderr, "tabway %d undefined\n", i);
	return 1;
}

static int asm_digittoi(int c, int base)
{
	if (c >= '0' && c <= '9')
		return c - '0';
	if (base == 10)
		return -EINVAL;

	c = tolower(c);
	if (c >= 'a' && c <= 'f')
		return c - 'a' + 10;
	return -EINVAL;
}

static int asm_atoui(const char *s)
{
	int len = strlen(s);
	int base = 10;
	int ret = 0;
	int digit;

	if (s[len - 1] == 'h') {
		base = 16;
		len--;
	} else if (s[len - 1] == 'd')
		len--;

	for (; len > 0; len--) {
		digit = asm_digittoi(*s++, base);
		if (digit < 0)
			return digit;
		ret = base * ret + digit;
	}
	return ret;
}

static int asm_atoi(const char *s, int *v)
{
	int len = strlen(s);
	int base = 10;
	int ret = 0;
	int digit;
	int neg = 0;

	if (s[len - 1] == 'h') {
		base = 16;
		len--;
	} else if (s[len - 1] == 'd')
		len--;
	if (s[0] == '-') {
		neg = 1;
		s++;
		len--;
	}

	for (; len > 0; len--) {
		digit = asm_digittoi(*s++, base);
		if (digit < 0)
			return digit;
		ret = base * ret + digit;
	}
	if (neg)
		ret = -ret;
	*v = ret;
	return 0;
}

static int send_nlogpom(struct ar2c_config *c)
{
	int i;
	struct lp_par_pkt *p;
	size_t p_len = sizeof(*p) + LP_PAR_CRC_LEN + LP_PAR_NLOGPOM_COUNT;
	char *value = NULL;
	size_t v_len = 0;
	int ret;

	p = malloc(p_len);
	if (p == NULL)
		return -ENOMEM;
	p->sync = LP_PAR_SYNC;
	p->cmd = LP_PAR_CMD;
	p->param_id = LP_PARID_NLOGPOM;
	p->plen = p_len - sizeof(*p);

	for (i = 0; i < LP_PAR_NLOGPOM_COUNT; i++) {
		char buf[64];
		int v;

		snprintf(buf, 64, "analog map %d", i);
		ret = ar2c_config_get(c, buf, &value, &v_len);
		if (ret == -ENOENT)
			break;
		if (ret < 0) {
			fprintf(stderr, "ar2c_config_get: %s\n",
					strerror(-ret));
			break;
		}

		v = asm_atoui(value);
		if (v < 0) {
			fprintf(stderr, "error in '%s' number conversion: %s\n",
					value,	strerror(-v));
			break;
		}
		if (v > 0xff) {
			fprintf(stderr, "analog map index (%d) out of range\n",
					v);
			break;
		}

		p->payload[i] = v;
	}
	if (i == LP_PAR_NLOGPOM_COUNT) {
		lp_par_checksum(p);
		ar2c_wait();
		ret = write_all(ofd, p, p_len);
		if (ret != p_len) {
			if (ret < 0) {
				perror("write");
				return -errno;
			}
			fprintf(stderr, "can't write.\n");
			return -EINVAL;
		}
		return 0;
	} else if (verbose > 0)
		fprintf(stderr, "analog map %d undefined\n", i);
	return 1;
}

static int send_analog_offsets(struct ar2c_config *c)
{
	int i;
	struct lp_par_pkt *p;
	size_t p_len = sizeof(*p) + LP_PAR_CRC_LEN + 2 * LP_PAR_AOFFSETS_COUNT;
	char *value = NULL;
	size_t v_len = 0;
	int ret;
	int t;

	p = malloc(p_len);
	if (p == NULL)
		return -ENOMEM;
	p->sync = LP_PAR_SYNC;
	p->cmd = LP_PAR_CMD;
	p->param_id = LP_PARID_AOFFSETS;
	p->plen = p_len - sizeof(*p);

	for (i = 0; i < LP_PAR_AOFFSETS_COUNT; i++) {
		char buf[64];
		int v;
		uint16_t a;

		snprintf(buf, 64, "analog offset %d", i);
		ret = ar2c_config_get(c, buf, &value, &v_len);
		if (ret == -ENOENT)
			break;
		if (ret < 0) {
			fprintf(stderr, "ar2c_config_get: %s\n",
					strerror(-ret));
			break;
		}

		v = asm_atoi(value, &t);
		if (v < 0) {
			fprintf(stderr, "error in '%s' number conversion: %s\n",
					value,	strerror(-v));
			break;
		}

		a = (uint16_t)(int16_t) t;
		p->payload[2 * i] = a & 0xff;
		p->payload[2 * i + 1] = a >> 8;
	}
	if (i == LP_PAR_AOFFSETS_COUNT) {
		lp_par_checksum(p);
		ar2c_wait();
		ret = write_all(ofd, p, p_len);
		if (ret != p_len) {
			if (ret < 0) {
				perror("write");
				return -errno;
			}
			fprintf(stderr, "can't write.\n");
			return -EINVAL;
		}
		return 0;
	} else if (verbose > 0)
		fprintf(stderr, "analog offset %d undefined\n", i);
	return 1;
}

static int send_analog_divs(struct ar2c_config *c)
{
	int i;
	struct lp_par_pkt *p;
	size_t p_len = sizeof(*p) + LP_PAR_CRC_LEN + LP_PAR_ADIVS_COUNT;
	char *value = NULL;
	size_t v_len = 0;
	int ret;
	int t;

	p = malloc(p_len);
	if (p == NULL)
		return -ENOMEM;
	p->sync = LP_PAR_SYNC;
	p->cmd = LP_PAR_CMD;
	p->param_id = LP_PARID_ADIVS;
	p->plen = p_len - sizeof(*p);

	for (i = 0; i < LP_PAR_ADIVS_COUNT; i++) {
		char buf[64];
		int v;

		snprintf(buf, 64, "analog div %d", i);
		ret = ar2c_config_get(c, buf, &value, &v_len);
		if (ret == -ENOENT)
			break;
		if (ret < 0) {
			fprintf(stderr, "ar2c_config_get: %s\n",
					strerror(-ret));
			break;
		}

		v = asm_atoi(value, &t);
		if (v < 0) {
			fprintf(stderr, "error in '%s' number conversion: %s\n",
					value,	strerror(-v));
			break;
		}

		p->payload[i] = (int8_t) t;
	}
	if (i == LP_PAR_ADIVS_COUNT) {
		lp_par_checksum(p);
		ar2c_wait();
		ret = write_all(ofd, p, p_len);
		if (ret != p_len) {
			if (ret < 0) {
				perror("write");
				return -errno;
			}
			fprintf(stderr, "can't write.\n");
			return -EINVAL;
		}
		return 0;
	} else if (verbose > 0)
		fprintf(stderr, "analog div %d undefined\n", i);
	return 1;
}

static int config_get_uint8(struct ar2c_config *c, const char *buf, uint8_t *v)
{
	int ret;
	char *value = NULL;
	size_t v_len = 0;

	ret = ar2c_config_get(c, buf, &value, &v_len);
	if (ret == -ENOENT)
		return ret;
	if (ret < 0) {
		fprintf(stderr, "ar2c_config_get: %s\n",
				strerror(-ret));
		return ret;
	}

	ret = asm_atoui(value);
	if (ret < 0) {
		fprintf(stderr, "error in '%s' number conversion: %s\n",
				value,	strerror(-ret));
		return ret;
	}

	if (ret > 0xff) {
		fprintf(stderr, "value (%d) out of unsigned byte range\n",
				ret);
		return -ERANGE;
	}
	*v = ret;
	return 0;
}

static int config_get_int32le(struct ar2c_config *c, const char *buf,
		int32_t *v)
{
	int ret;
	char *value = NULL;
	size_t v_len = 0;
	int t;

	ret = ar2c_config_get(c, buf, &value, &v_len);
	if (ret == -ENOENT)
		return ret;
	if (ret < 0) {
		fprintf(stderr, "ar2c_config_get: %s\n",
				strerror(-ret));
		return ret;
	}

	ret = asm_atoi(value, &t);
	if (ret < 0) {
		fprintf(stderr, "error in '%s' number conversion: %s\n",
				value,	strerror(-ret));
		return ret;
	}

	put_unaligned(cpu_to_le32(t), v);
	return 0;
}

static int config_get_uint16le(struct ar2c_config *c, const char *buf,
		uint16_t *v)
{
	int ret;
	char *value = NULL;
	size_t v_len = 0;
	int t;

	ret = ar2c_config_get(c, buf, &value, &v_len);
	if (ret == -ENOENT)
		return ret;
	if (ret < 0) {
		fprintf(stderr, "ar2c_config_get: %s\n",
				strerror(-ret));
		return ret;
	}

	ret = asm_atoi(value, &t);
	if (ret < 0) {
		fprintf(stderr, "error in '%s' number conversion: %s\n",
				value,	strerror(-ret));
		return ret;
	}

	put_unaligned(cpu_to_le16(t), v);
	return 0;
}

static int send_wc8_config(struct ar2c_config *c)
{
	int i;
	struct lp_par_pkt *p;
	struct lp_par_wc8_config *payload;
	size_t p_len = sizeof(*p) + LP_PAR_CRC_LEN + sizeof(*payload);
	int ret;

	p = malloc(p_len);
	if (p == NULL)
		return -ENOMEM;
	p->sync = LP_PAR_SYNC;
	p->cmd = LP_PAR_CMD;
	p->param_id = LP_PARID_WC8;
	p->plen = p_len - sizeof(*p);
	payload = (struct lp_par_wc8_config *) p->payload;

	for (i = 0; i < LP_PAR_WC8_COUNT; i++) {
		char buf[64];

		payload->id = i;

		snprintf(buf, 64, "WC8_dest_offset %d", i);
		if (config_get_uint8(c, buf, &payload->dest_offset)) {
			if (verbose > 0)
				fprintf(stderr, "warning: invalid/undefined "
						"`WC8_dest_offset %d'", i);
			break;
		}

		snprintf(buf, 64, "WC8_block_lth %d", i);
		if (config_get_uint8(c, buf, &payload->length)) {
			if (verbose > 0)
				fprintf(stderr, "warning: invalid/undefined "
						"`WC8_block_lth %d'", i);
			break;
		}

		snprintf(buf, 64, "WC8_block_base %d", i);
		if (config_get_uint16le(c, buf, &payload->base)) {
			if (verbose > 0)
				fprintf(stderr, "warning: invalid/undefined "
						"`WC8_block_base %d'", i);
			break;
		}

		lp_par_checksum(p);
		ar2c_wait();
		ret = write_all(ofd, p, p_len);
		if (ret != p_len) {
			if (ret < 0) {
				perror("write");
				return -errno;
			}
			fprintf(stderr, "can't write.\n");
			return -EINVAL;
		}
	}
	return 0;
}

static int send_card_bus_config(struct ar2c_config *c)
{
	int i;
	struct lp_par_pkt *p;
	struct lp_par_card_bus_config *payload;
	size_t p_len = sizeof(*p) + LP_PAR_CRC_LEN + sizeof(*payload);
	int ret;

	p = malloc(p_len);
	if (p == NULL)
		return -ENOMEM;
	p->sync = LP_PAR_SYNC;
	p->cmd = LP_PAR_CMD;
	p->param_id = LP_PARID_CARD_BUS;
	p->plen = p_len - sizeof(*p);
	payload = (struct lp_par_card_bus_config *) p->payload;

	for (i = 0; i < LP_PAR_CARD_BUS_COUNT; i++) {
		char buf[64];

		payload->id = i;

#define GET_VARIABLE(namefmt, type, dest)				\
	snprintf(buf, 64, namefmt" %d", i);				\
	if (config_get_##type(c, buf, dest)) {				\
		if (verbose > 0)					\
			fprintf(stderr, "warning: invalid/undefined `"	\
					namefmt " %d'", i);		\
		break;							\
	}

#define GET_VAR_UINT8(namefmt, dest) GET_VARIABLE(namefmt, uint8, dest)
#define GET_VAR_UINT16(namefmt, dest) GET_VARIABLE(namefmt, uint16le, dest)

		GET_VAR_UINT16("card_bus_base", &payload->base);
		GET_VAR_UINT16("card_bus_sbin_offset", &payload->bin_offset);
		GET_VAR_UINT16("card_bus_dest_offset", &payload->bin_dest);
		GET_VAR_UINT16("card_bus_bin_lth", &payload->bin_length);
		GET_VAR_UINT16("card_bus_sanal_offset", &payload->anal_offset);
		GET_VAR_UINT16("card_bus_anal_offset", &payload->anal_dest);
		GET_VAR_UINT16("card_bus_anal_lth", &payload->anal_length);
		GET_VAR_UINT16("card_bus_sHB_offset", &payload->hb_offset);
		GET_VAR_UINT16("card_bus_HB_offset", &payload->hb_dest);
		GET_VAR_UINT16("card_bus_HB_lth", &payload->hb_length);
		GET_VAR_UINT16("card_bus_stat_adr", &payload->diag_offset);
		GET_VAR_UINT8("card_bus_stat_ok", &payload->diag_ok);
		GET_VAR_UINT8("card_bus_stat_bad", &payload->diag_trfail);
		GET_VAR_UINT8("card_bus_chk_value", &payload->chk_value);

		lp_par_checksum(p);
		ar2c_wait();
		ret = write_all(ofd, p, p_len);
		if (ret != p_len) {
			if (ret < 0) {
				perror("write");
				return -errno;
			}
			fprintf(stderr, "can't write.\n");
			return -EINVAL;
		}
	}
	return 0;
}

static int send_posvel(struct ar2c_config *c)
{
	struct lp_par_posvel *payload;
	struct lp_par_pkt *p;
	size_t p_len = sizeof(*p) + LP_PAR_CRC_LEN + sizeof(*payload);
	int ret;

	p = malloc(p_len);
	if (p == NULL)
		return -ENOMEM;

	p->sync = LP_PAR_SYNC;
	p->cmd = LP_PAR_CMD;
	p->param_id = LP_PARID_POSVEL;
	p->plen = p_len - sizeof(*p);
	payload = (struct lp_par_posvel *) p->payload;

	if (config_get_uint8(c, "tacho", &payload->tacho))
		return 1;

	if (config_get_uint8(c, "szybowskaz", &payload->szybowskaz))
		return 1;

	if (config_get_uint8(c, "typ impulsatora", &payload->imp))
		return 1;

	if (config_get_uint8(c, "wa2", &payload->wa2))
		return 1;

	if (config_get_uint8(c, "pommax", &payload->pommax))
		return 1;

	if (config_get_uint8(c, "dzw. 22", &payload->dzw_flag))
		return 1;

	if (config_get_int32le(c, "pkk1", &payload->pkk1))
		return 1;

	if (config_get_int32le(c, "pkk2", &payload->pkk2))
		return 1;

	lp_par_checksum(p);
	ar2c_wait();
	ret = write_all(ofd, p, p_len);
	if (ret != p_len) {
		if (ret < 0) {
			perror("write");
			return -errno;
		}
		fprintf(stderr, "can't write.\n");
		return -EINVAL;
	}
	return 0;
}

struct {
	const char *name;
	int (*put)(void *priv, const char *name, const char *value);
	void *priv;
} param_table[] = {
	{ "nazwa kopalni", put_nazwa_kopalni, NULL },
	{ "nazwa szybu", put_nazwa_szybu, NULL },
};

#define ARRAY_SIZE(x)		(sizeof(x) / sizeof(x[0]))
#define PARAM_TABLE_SIZE	ARRAY_SIZE(param_table)

static void send_parameters(struct ar2c_config *c)
{
	char *value = NULL;
	size_t len = 0;
	int ret;
	int i;

	for (i = 0; i < PARAM_TABLE_SIZE; i++) {
		const char *name = param_table[i].name;
		ret = ar2c_config_get(c, name, &value, &len);
		if (ret == -ENOENT) {
			if (verbose > 0)
				fprintf(stderr, "skiping %s parameter\n", name);
		} else if (ret < 0) {
			fprintf(stderr, "ar2c_config_get: %s\n",
					strerror(-ret));
		} else
			param_table[i].put(param_table[i].priv, name, value);
	}

	send_binary_names(c);
	send_analog_descriptions(c);
	send_binary_functions(c);
	send_nlogpom(c);
	send_posvel(c);
	send_analog_offsets(c);
	send_wc8_config(c);
	send_card_bus_config(c);
	send_analog_divs(c);
}

static void do_putdata(FILE *f)
{
	struct ar2c_config c;
	int ret;

	ret = ar2c_config_init(&c);
	if (ret) {
		fprintf(stderr, "ar2c_config_init: %s\n", strerror(-ret));
		exit(EXIT_FAILURE);
	}

	ret = ar2c_config_load(&c, f);
	if (ret) {
		fprintf(stderr, "ar2c_config_load: %s\n", strerror(-ret));
		exit(EXIT_FAILURE);
	}

	send_parameters(&c);

	ret = ar2c_config_destroy(&c);
	if (ret) {
		fprintf(stderr, "ar2c_config_destroy: %s\n", strerror(-ret));
		exit(EXIT_FAILURE);
	}
}

static void spd_return(void)
{
	if (speed) {
		if (rspeed == 0) {
			rspeed = speedtoar2cspeed(bspeed);
			if (rspeed < 0) {
				fprintf(stderr, "Unable to return to initial "
						"serial speed.\n");
				exit(EXIT_FAILURE);
			}
		}
		if ((speed != rspeed) && ar2c_set_speed(ifd, rspeed)) {
			fprintf(stderr, "Setting serial speed error\n");
			exit(EXIT_FAILURE);
		}
	}
}

int main(int argc, char **args)
{
	signal(SIGALRM, SIG_IGN);
	parse_command_line(argc, args);
	switch (targ_type) {
	case TT_STDIN:
		ifd = 0;
		ofd = 1;
		out = stderr;
		cmd_timeout = cmd_timeout ? : 2000;
		break;
	case TT_FILE:
		if (targ_opt == NULL) {
			fprintf(stderr, "fatal: null targ_opt.\n");
			exit(EXIT_FAILURE);
		}
		if ((ifd = open(targ_opt, O_RDWR)) < 0) {
			fprintf(stderr, "error: Can't open %s:%s\n",
					targ_opt, strerror(errno));
			exit(EXIT_FAILURE);
		}
		ofd = ifd;
		out = stdout;
		cmd_timeout = cmd_timeout ? : 2000;
		break;
	case TT_TCP:
		if (targ_opt == NULL) {
			fprintf(stderr, "fatal: null targ_opt.\n");
			exit(EXIT_FAILURE);
		}
		if ((ifd = pcomm_open(targ_opt, port)) < 0) {
			fprintf(stderr, "error: Can't connect to %s\n",
					targ_opt);
			exit(EXIT_FAILURE);
		}
		ofd = ifd;
		out = stdout;
		cmd_timeout = cmd_timeout ? : 10000;
		break;
	case TT_FIFO:
		if (targ_opt == NULL) {
			fprintf(stderr, "fatal: null targ_opt.\n");
			exit(EXIT_FAILURE);
		}
		if (targ_opt2 == NULL) {
			fprintf(stderr, "fatal: null targ_opt2.\n");
			exit(EXIT_FAILURE);
		}
		/* order is important */
		if ((ofd = open(targ_opt2, O_WRONLY)) == -1) {
			fprintf(stderr, "fatal: Can't open wr_fifo.\n");
			exit(EXIT_FAILURE);
		}
		if ((ifd = open(targ_opt, O_RDONLY)) == -1) {
			fprintf(stderr, "fatal: Can't open rd_fifo.\n");
			exit(EXIT_FAILURE);
		}
		out = stdout;
		cmd_timeout = cmd_timeout ? : 500;
		break;
	case TT_SERIAL:
		if (targ_opt == NULL) {
			fprintf(stderr, "fatal: null targ_opt.\n");
			exit(EXIT_FAILURE);
		}
		if ((ifd = open(targ_opt, O_RDWR)) < 0) {
			fprintf(stderr, "error: Can't open %s:%s\n",
					targ_opt, strerror(errno));
			exit(EXIT_FAILURE);
		}
		ofd = ifd;
		serial_set_speed(ifd, bspeed);
		if (ar2cspeedtospeed(speed) != bspeed)
			if (ar2c_set_speed(ifd, speed)) {
				fprintf(stderr, "Setting serial speed "
						"error\n");
				exit(EXIT_FAILURE);
			}
		out = stdout;
		cmd_timeout = cmd_timeout ? : 500;
		break;
	default:
		fprintf(stderr, "fatal: Internal error.\n");
		exit(EXIT_FAILURE);
	}

	atexit(spd_return);
	do_putdata(stdin);
	return EXIT_SUCCESS;
}
