/*
 * ar2c-par-utils - arbusctl.c
 *
 * Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <comm.h>
#include <time.h>
#include <ar2c.h>
#include <signal.h>
#include <tread.h>
#include <ctype.h>
#include <lpde.h>
#include <pci_regs.h>
#include <arbusctl.h>
#include <ar2cpar.h>
#include <stdarg.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <mclock.h>

#ifndef PATH_MAX
#define PATH_MAX	4096
#endif

#define ARBUS_BASE		0xd8000
#define ARBUS_MAGIC		0xf0
#define ARBUS_MAGIC_VALUE	0x53425241

#define ARBUS_REG(dev, fun, reg) \
	(ARBUS_BASE + (((dev) * 8 + (fun)) << 8) + (reg))

#define TT_STDIN	0
#define TT_FILE		1
#define TT_TCP		2
#define TT_SERIAL	4

#ifdef HAVE_ENDIAN_H
#include <endian.h>
#endif

#ifdef HAVE_BYTESWAP_H
#include <byteswap.h>
#endif

#if BYTE_ORDER == BIG_ENDIAN

#define cpu_to_be16(x)		(x)
#define cpu_to_be32(x)		(x)
#define cpu_to_le16(x)		(bswap_16(x))
#define cpu_to_le32(x)		(bswap_32(x))

#define be16_to_cpu(x)		(x)
#define be32_to_cpu(x)		(x)
#define le16_to_cpu(x)		(bswap_16(x))
#define le32_to_cpu(x)		(bswap_32(x))

#else

#if BYTE_ORDER == LITTLE_ENDIAN

#define cpu_to_be16(x)		(bswap_16(x))
#define cpu_to_be32(x)		(bswap_32(x))
#define cpu_to_le16(x)		(x)
#define cpu_to_le32(x)		(x)

#define be16_to_cpu(x)		(bswap_16(x))
#define be32_to_cpu(x)		(bswap_32(x))
#define le16_to_cpu(x)		(x)
#define le32_to_cpu(x)		(x)

#else

#error "Only little and big endian systems are supported."

#endif
#endif

#define get_unaligned(ptr)						\
	({								\
		struct __str {						\
			typeof(*(ptr)) __v;				\
		} __attribute__((packed)) *__p				\
			= (struct __str *) (ptr);			\
		__p->__v;						\
	})

#define put_unaligned(val, ptr)						\
	do {								\
		struct __str {						\
			typeof(*(ptr)) __v;				\
		} __attribute__((packed)) *__p				\
			= (struct __str *) (ptr);			\
		__p->__v = (val);					\
	} while (0)

#define fatal(fmt...) \
	do { \
		fprintf(stderr, fmt); \
		exit(EXIT_FAILURE); \
	} while (0)

static int targ_type = TT_STDIN;
static char *targ_opt;
static const char *port = "2000";
static int ifd, ofd;
static int verbose;
static FILE *out;
static speed_t bspeed = B9600;
static int speed, rspeed;
int cmd_timeout;
static int synchronize;

static int pci_domain;
static int pci_bus;
static int pci_dev;
static int pci_fun;

static int lspci_hexdump;

static const char *command;
static const char *config_file;

static int exit_code = EXIT_SUCCESS;

void dev_error(struct arbus_dev *dev, const char *fmt, ...)
{
	va_list ap;

	fprintf(stderr, "%s error: ", dev->slot);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);

	exit_code = EXIT_FAILURE;
}

static void show_usage(char *name)
{
	fprintf(stderr, "ar2c-arbusctl " PACKAGE_VERSION ".\n"
"Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>.\n");

	fprintf(stderr,	"ar2c-arbusctl.\n"
			"usage: %s [options]\n"
			"Options:\n"
			"      --bbitrate	base bitrate(default 9600bps)\n"
			"  -a, --no-sync\t	don't use 0xff 0xd2 sync\n"
			"  -b, --bitrate\t	set bitrate\n"
			"  -c, --config		use configuration file\n"
			"  -d, --device		use selected device\n"
			"  -h, --help		show this help\n"
			"  -p, --port		remote port(2000)\n"
			"  -q, --quiet		be quiet\n"
			"      --rbitrate	return bitrate\n"
			"  -s, --host		remote host(localhost)\n"
			"  -t, --timeout\t	set timeout\n"
			"  -v, --verbose\t	be verbose\n"
			"  -V, --version\t	show version\n",
			(name == NULL) ? "ar2c-arbusctl" : name);
	exit(EXIT_SUCCESS);
}

static void show_version(void)
{
	fprintf(stderr, "ar2c-arbusctl " PACKAGE_VERSION ".\n"
"Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>.\n\n");

	fprintf(stderr,	"This program comes with ABSOLUTELY NO WARRANTY; "
			"see COPYING for details.\n"
			"This is free software, and you are welcome to "
			"redistribute it\n"
			"under certain conditions; see COPYING for details.\n");
	exit(EXIT_SUCCESS);
}

static void parse_device(const char *device)
{
	char *s;
	long a, b;

	a = strtol(device, &s, 0);
	if (*s == '\0') {
		pci_dev = a;
		return;
	} else if (*s == '.') {
		pci_dev = a;

		s++;
		if (*s == '\0')
			return;

		a = strtol(s, &s, 0);
		if (*s != '\0')
			fatal("unexpected '%s'\n", s);
		pci_fun = a;
	} else if (*s == ':') {
		s++;
		if (*s == '\0')
			fatal("missing PCI device number\n");

		b = strtol(s, &s, 0);
		if (*s == '\0') {
			pci_bus = a;
			pci_dev = b;
			return;
		}

		if (*s == ':') {
			pci_domain = a;
			pci_bus = b;

			s++;
			if (*s == '\0')
				return;

			a = strtol(s, &s, 0);
			if (*s == '\0') {
				pci_dev = a;
				return;
			}
			if (*s != '.')
				fatal("unexpected '%s'\n", s);

			pci_dev = a;

			s++;
			if (*s == '\0')
				return;

			a = strtol(s, &s, 0);
			if (*s != '\0')
				fatal("unexpected '%s'\n", s);
			pci_fun = a;
		} else if (*s == '.') {
			pci_bus = a;
			pci_dev = b;

			s++;
			if (*s == '\0')
				return;

			a = strtol(s, &s, 0);
			if (*s != '\0')
				fatal("unexpected '%s'\n", s);
			pci_fun = a;
			return;
		} else {
			fatal("unexpected '%s'\n", s);
		}
	} else {
		fatal("unexpected '%s'\n", s);
	}
}

static void parse_command_line(int argc, char **argv)
{
	int c;
	int option_index = 0;

	static struct option long_options[] = {
		{"bbitrate", 1, 0, 1},
		{"bitrate", 1, 0, 'b'},
		{"config", 1, 0, 'c'},
		{"device", 1, 0, 'd'},
		{"devfn", 1, 0, 'i'},
		{"help", 0, 0, 'h'},
		{"host", 1, 0, 's'},
		{"no-sync", 0, 0, 3},
		{"port", 1, 0, 'p'},
		{"quiet", 0, 0, 'q'},
		{"rbitrate", 1, 0, 2},
		{"timeout", 1, 0, 't'},
		{"verbose", 0, 0, 'v'},
		{"version", 0, 0, 'V'},
		{0, 0, 0, 0},
	};

	while (1) {
		c = getopt_long(argc, argv, "b:c:d:g:i:p:hs:t:vxqV",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 1:
			bspeed = itospeed(atoi(optarg));
			if (bspeed < 0) {
				fprintf(stderr, "Unsupported initial "
						"speed setting.\n");
				exit(EXIT_FAILURE);
			}
			if (rspeed == 0) {
				rspeed = speedtoar2cspeed(bspeed);
				if (rspeed < 0)
					rspeed = 0;
			}
			targ_type = TT_SERIAL;
			break;
		case 2:
			rspeed = itoar2cspeed(atoi(optarg));
			if (rspeed < 0) {
				fprintf(stderr, "Invalid return speed "
							"setting.");
				exit(EXIT_FAILURE);
			}
			targ_type = TT_SERIAL;
			break;
		case 3:
			synchronize = 0;
			break;
		case 'b':
			speed = itoar2cspeed(atoi(optarg));
			if (speed < 0) {
				fprintf(stderr, "Invalid speed setting.");
				exit(EXIT_FAILURE);
			}
			targ_type = TT_SERIAL;
			break;
		case 'c':
			config_file = strdup(optarg);
			break;
		case 'd':
			targ_opt = optarg;
			if ((targ_type == TT_TCP) || (targ_type == TT_STDIN))
				targ_type = TT_FILE;
			break;
		case 'h':
			show_usage(argv[0]);
			break;
		case 'i':
			parse_device(optarg);
			break;
		case 'p':
			port = strdup(optarg);
			targ_type = TT_TCP;
			break;
		case 'v':
			verbose++;
			break;
		case 'q':
			verbose--;
			break;
		case 's':
			targ_opt = optarg;
			targ_type = TT_TCP;
			break;
		case 't':
			cmd_timeout = atoi(optarg);
			break;
		case 'x':
			lspci_hexdump = 1;
			break;
		case 'V':
			show_version();
			break;
		default:
			fprintf(stderr, "Unknown option.\n");
			exit(EXIT_FAILURE);
		}
	}
	if (optind < argc)
		command = strdup(argv[optind++]);
}

#define log_warning(fmt...)	fprintf(stderr, fmt)

static ssize_t write_all(int fd, void *buf, size_t len)
{
	char *t = buf;
	ssize_t ret, total = 0;

	while (total < len) {
		ret = write(fd, t + total, len - total);
		if (ret <= 0) {
			if (!total)
				return ret;
			break;
		}
		total += ret;
	}
	return total;
}

static int __ar2c_wait(void)
{
	static char __sync_seq[] = { 0xff, 0xd2, };
	int ret;
	char buf[2];

	do {
		ret = write_all(ofd, __sync_seq, sizeof(__sync_seq));
		if (ret < sizeof(__sync_seq))
			return -errno;

		ret = t_read(ifd, buf, 2, 300);
		if (ret < 0)
			return -errno;
	} while (ret < 2);
	return 0;
}

static void ar2c_wait(void)
{
	if (synchronize)
		__ar2c_wait();
}

static int ar2c_exec(unsigned int cmd, void *arg, size_t arg_len)
{
	u8 *c;
	int ret;

	c = malloc(2 + arg_len);
	if (c == NULL)
		return -ENOMEM;

	c[0] = 0xff;
	c[1] = (~cmd << 4) | cmd;
	memcpy(c + 2, arg, arg_len);

	ret = write_all(ofd, c, arg_len + 2);
	free(c);
	if (ret != arg_len + 2)
		return -EIO;
	return 0;
}

static unsigned long current_segment;

/*
 * dump 256 bytes of from selected location
 */
static int __ar2c_dump_memory(unsigned long addr, u8 *data)
{
	struct lpde_dump_memory d;
	int ret;
	static u8 dummy[256];

	addr &= 0xfff00;
	d.offset = cpu_to_be16(addr & 0xff00);
	d.segment = cpu_to_be16((addr >> 4) & 0xf000);
	ar2c_wait();

	if (data == NULL)
		data = dummy;

	current_segment = addr >> 16;

	ret = ar2c_exec(LPDE_DUMP_MEMORY, &d, sizeof(d));
	if (ret)
		return ret;

	ret = t_read(ifd, data, 256, cmd_timeout);
	if (ret != 256) {
		fprintf(stderr, "error: read %d bytes instead of %d\n",
				ret, 256);
		return -EIO;
	}
	return 0;
}

int ar2c_dump_memory(unsigned long addr, u8 *data)
{
	int ret;
	int i;

	if (verbose > 0)
		fprintf(stderr, "dump(%05lx, 256)\n", addr);
	for (i = 0; i < 5; i++) {
		ret = __ar2c_dump_memory(addr, data);
		if (!ret)
			break;
	}
	return ret;
}

#if 0
static int ar2c_dump(unsigned long addr)
{
	u8 data[256];
	unsigned int i, j;
	int ret;

	ret = ar2c_dump_memory(addr, data);
	if (ret) {
		fprintf(stderr, "dump: error %d\n", ret);
		return ret;
	}
	for (i = 0; i < 256; i += 16) {
		fprintf(stderr, "%05lx:", addr + i);
		for (j = 0; j < 16; j++)
			fprintf(stderr, " %02x", data[i + j]);
		fprintf(stderr, "\n");
	}
	return ret;
}
#endif

static void select_segment(unsigned long addr)
{
	unsigned long segment = addr >> 16;

	if (segment != current_segment)
		ar2c_dump_memory(addr & 0xf0000, NULL);
}

static int __ar2c_readb(unsigned long addr)
{
	struct lpde_readb d;
	struct lpde_readb_r r;
	int ret;

	select_segment(addr);
	d.offset = cpu_to_be16(addr & 0xffff);
	ar2c_wait();

	ret = ar2c_exec(LPDE_READB, &d, sizeof(d));
	if (ret)
		return ret;

	ret = t_read(ifd, &r, sizeof(r), cmd_timeout);
	if (ret != sizeof(r)) {
		fprintf(stderr, "error: read %d bytes instead of %d\n",
				ret, (unsigned int) sizeof(r));
		return -EIO;
	}
	if (be16_to_cpu(r.offset) != (addr & 0xffff)) {
		fprintf(stderr, "error: offset error %04x read:%04lx\n",
				be16_to_cpu(r.offset), addr & 0xffff);
		return -EIO;
	}
	return r.value;
}

int ar2c_readb(unsigned long addr)
{
	int ret;
	int i;

	for (i = 0; i < 5; i++) {
		ret = __ar2c_readb(addr);
		if (ret >= 0)
			break;
		do_resync(ifd, ofd);
	}
	if (verbose > 0)
		fprintf(stderr, "readb(%05lx) = %d\n", addr, ret);
	return ret;
}

static int __ar2c_ioread8_copy_pipelined(u8 *data, unsigned long addr,
		unsigned int count)
{
	struct lpde_readb d;
	struct lpde_readb_r r;
	unsigned int i;
	int ret;

	select_segment(addr);
	ar2c_wait();

	for (i = 0; i < count; i++) {
		d.offset = cpu_to_be16((addr + i) & 0xffff);
		ret = ar2c_exec(LPDE_READB, &d, sizeof(d));
		if (ret)
			return ret;
	}

	for (i = 0; i < count; i++) {
		ret = t_read(ifd, &r, sizeof(r), cmd_timeout);
		if (ret != sizeof(r)) {
			fprintf(stderr, "error: read %d bytes instead of "
					"%zd\n", ret, sizeof(r));
			return -EIO;
		}
		if (be16_to_cpu(r.offset) != ((addr + i) & 0xffff)) {
			fprintf(stderr, "error: offset error %04x read:%04lx\n",
					be16_to_cpu(r.offset), addr & 0xffff);
			return -EIO;
		}
		data[i] = r.value;
	}
	return 0;
}

static int ar2c_ioread8_copy_pipelined(u8 *data, unsigned long addr,
		unsigned int count)
{
	int ret;
	int i;

	for (i = 0; i < 5; i++) {
		ret = __ar2c_ioread8_copy_pipelined(data, addr, count);
		if (!ret)
			break;
		do_resync(ifd, ofd);
	}
	return ret;
}

static int ar2c_ioread8_copy_slow(u8 *data, unsigned long addr,
		unsigned int count)
{
	int ret;
	int i;

	for (i = 0; i < count; i++) {
		ret = ar2c_readb(addr + i);
		if (ret < 0)
			return ret;
		data[i] = ret;
	}
	return 0;
}

static int allow_pipelining = 1;

int ar2c_ioread8_copy(u8 *data, unsigned long addr,
		unsigned int count)
{
	if (verbose > 0)
		fprintf(stderr, "ioread8_copy(%05lx, %u)\n", addr, count);

	if (allow_pipelining)
		return ar2c_ioread8_copy_pipelined(data, addr, count);
	return ar2c_ioread8_copy_slow(data, addr, count);
}

int ar2c_readw(unsigned long addr)
{
	u8 b[2];
	int ret;

	ret = ar2c_ioread8_copy(b, addr, 2);
	if (ret)
		return -1;
	return (b[1] << 8) | b[0];
}

int ar2c_readl(unsigned long addr, u32 *x)
{
	u8 b[4];
	int ret;

	ret = ar2c_ioread8_copy(b, addr, 4);
	if (ret)
		return -1;
	*x = (b[3] << 24) | (b[2] << 16) | (b[1] << 8) | b[0];
	return 0;
}

static u8 __arbus_read_byte(struct arbus_dev *dev, int pos)
{
	return dev->config_regs[pos];
}

static u16 __arbus_read_word(struct arbus_dev *dev, int pos)
{
	u16 *x;

	x = (u16 *) &dev->config_regs[pos];
	return le16_to_cpu(get_unaligned(x));
}

static u32 __arbus_read_long(struct arbus_dev *dev, int pos)
{
	u32 *x;

	x = (u32 *) &dev->config_regs[pos];
	return le32_to_cpu(get_unaligned(x));
}

static void arbus_update_cache_or_die(struct arbus_dev *dev)
{
	unsigned int slot = PCI_SLOT(dev->devfn);
	unsigned int fun = PCI_FUNC(dev->devfn);
	unsigned int retry = 5;
	int ret;
	u32 x;

	do {
		ret = __ar2c_dump_memory(ARBUS_REG(slot, fun, 0),
				dev->config_regs);
		if (ret >= 0) {
			x = __arbus_read_long(dev, ARBUS_MAGIC);
			if (x == ARBUS_MAGIC_VALUE)
				break;

			ret = -ENXIO;
		}
		do_resync(ifd, ofd);
	} while (retry-- > 0);
 
 	if (ret < 0)
		fatal("PCI device %x:%02x: config read error %s\n",
				slot, fun, strerror(-ret));
 	dev->valid = 1;
}

u8 arbus_read_byte(struct arbus_dev *dev, int pos)
{
	if (!dev->valid)
		arbus_update_cache_or_die(dev);

	return __arbus_read_byte(dev, pos);
}

u16 arbus_read_word(struct arbus_dev *dev, int pos)
{
	if (!dev->valid)
		arbus_update_cache_or_die(dev);

	return __arbus_read_word(dev, pos);
}

u32 arbus_read_long(struct arbus_dev *dev, int pos)
{
	if (!dev->valid)
		arbus_update_cache_or_die(dev);

	return __arbus_read_long(dev, pos);
}

static int __ar2c_writeb(u8 data, unsigned long addr)
{
	struct lpde_writeb d;
	struct lpde_writeb_r r;
	int ret;

	/*
	 * The LPDE is highly non-ortogonal. We need to perform read to
	 * write.
	 */
	ret = ar2c_readb(addr);
	if (ret < 0)
		return ret;

	if (addr <= 0xa0000 || addr >= 0x100000)
		fatal("invalid address %05lx\n", addr);

	d.value = data;
	ar2c_wait();

	ret = ar2c_exec(LPDE_WRITEB, &d, sizeof(d));
	if (ret)
		return ret;

	ret = t_read(ifd, &r, sizeof(r), cmd_timeout);
	if (ret != sizeof(r)) {
		fprintf(stderr, "error: read %d bytes instead of %d\n",
				ret, (unsigned int) sizeof(r));
		return -EIO;
	}
	if (be16_to_cpu(r.next_offset) != ((addr + 1) & 0xffff)) {
		fprintf(stderr, "error: offset error %04x read:%04lx\n",
				be16_to_cpu(r.next_offset), addr & 0xffff);
		return -EIO;
	}
#if 0
	fprintf(stderr, "writeb(%2x, %05x): next: %04x next_v: %02x\n", data,
			addr, be16_to_cpu(r.next_offset),
			r.next_value);
#endif
	return 0;
}

int ar2c_writeb(u8 x, unsigned long addr)
{
	int ret;
	int i;

	for (i = 0; i < 5; i++) {
		ret = __ar2c_writeb(x, addr);
		if (ret >= 0)
			break;
		do_resync(ifd, ofd);
	}
	return ret;
}

static int __ar2c_iowrite8_copy_pipelined(unsigned long addr, const u8 *data,
		unsigned int count)
{
	struct lpde_writeb d;
	struct lpde_writeb_r r;
	unsigned int i;
	int ret;

	/*
	 * The LPDE is highly non-ortogonal. We need to perform read to
	 * write.
	 */
	ret = ar2c_readb(addr);
	if (ret < 0)
		return ret;

	for (i = 0; i < count; i++) {
		d.value = data[i];
		ret = ar2c_exec(LPDE_WRITEB, &d, sizeof(d));
		if (ret)
			return ret;
	}

	for (i = 0; i < count; i++) {
		ret = t_read(ifd, &r, sizeof(r), cmd_timeout);
		if (ret != sizeof(r)) {
			fprintf(stderr, "error: read %d bytes instead of "
					"%zd\n", ret, sizeof(r));
			return -EIO;
		}
		if (be16_to_cpu(r.next_offset) != ((addr + 1 + i) & 0xffff)) {
			fprintf(stderr, "error: offset error %04x read:%04lx\n",
					be16_to_cpu(r.next_offset),
					addr & 0xffff);
			return -EIO;
		}
	}
	return 0;
}

static int ar2c_iowrite8_copy_pipelined(unsigned long addr, const u8 *data,
		unsigned int count)
{
	int ret;
	int i;

	for (i = 0; i < 5; i++) {
		ret = __ar2c_iowrite8_copy_pipelined(addr, data, count);
		if (!ret)
			break;
		do_resync(ifd, ofd);
	}
	return ret;
}

static int ar2c_iowrite8_copy_slow(unsigned long addr, const u8 *data,
		unsigned int count)
{
	int ret;
	int i;

	for (i = 0; i < count; i++) {
		ret = ar2c_writeb(data[i], addr + i);
		if (ret < 0)
			return ret;
	}
	return 0;
}

int ar2c_iowrite8_copy(unsigned long addr, const u8 *data, unsigned int count)
{
	if (allow_pipelining)
		return ar2c_iowrite8_copy_pipelined(addr, data, count);
	return ar2c_iowrite8_copy_slow(addr, data, count);
}

int ar2c_writew(u16 x, unsigned long addr)
{
	u8 data[2];

	data[0] = x >> 0;
	data[1] = x >> 8;
	return ar2c_iowrite8_copy(addr, data, 2);
}

int ar2c_writel(u32 x, unsigned long addr)
{
	u8 data[4];

	data[0] = x >> 0;
	data[1] = x >> 8;
	data[2] = x >> 16;
	data[3] = x >> 24;
	return ar2c_iowrite8_copy(addr, data, 4);
}

int arbus_write_byte(struct arbus_dev *dev, int pos, u8 data)
{
	int ret;

	dev->valid = 0;
	ret = ar2c_writeb(data, ARBUS_BASE + (dev->devfn << 8) + pos);
	if (ret < 0)
		fatal("arbus_write_byte: error reading %s at %d\n",
				dev->slot, pos);
	return ret;
}

int arbus_write_word(struct arbus_dev *dev, int pos, u16 data)
{
	int ret;

	dev->valid = 0;
	ret = ar2c_writew(data, ARBUS_BASE + (dev->devfn << 8) + pos);
	if (ret < 0)
		fatal("arbus_write_word: error reading %s at %d\n",
				dev->slot, pos);
	return ret;
}

int arbus_write_long(struct arbus_dev *dev, int pos, u32 data)
{
	int ret;

	dev->valid = 0;
	ret = ar2c_writel(data, ARBUS_BASE + (dev->devfn << 8) + pos);
	if (ret < 0)
		fatal("arbus_write_long: error reading %s at %d\n",
				dev->slot, pos);
	return ret;
}

static void spd_return(void)
{
	if (speed) {
		if (rspeed == 0) {
			rspeed = speedtoar2cspeed(bspeed);
			if (rspeed < 0) {
				fprintf(stderr, "Unable to return to initial "
						"serial speed.\n");
				exit(EXIT_FAILURE);
			}
		}
		if ((speed != rspeed) && ar2c_set_speed(ifd, rspeed)) {
			fprintf(stderr, "Setting serial speed error\n");
			exit(EXIT_FAILURE);
		}
	}
}

static int arbus_card_present(unsigned int dev, unsigned int fun)
{
	int ret;
	u32 x;

	ret = ar2c_readb(ARBUS_REG(dev, fun, ARBUS_MAGIC));
	if (ret != 'A')
		return 0;

	ret = ar2c_readl(ARBUS_REG(dev, fun, ARBUS_MAGIC), &x);
	if (ret)
		return 0;
	return x == ARBUS_MAGIC_VALUE;
}

static void do_lspci_hexdump(unsigned long addr)
{
	unsigned char config[256];
	int ret;
	int i, j;

	ret = ar2c_dump_memory(addr, config);
	if (ret < 0)
		return;

	for (i = 0; i < sizeof(config); i += 16) {
		printf("%02x:", i);
		for (j = 0; j < 16; j++)
			printf(" %02x", config[i + j]);
		printf("\n");
	}
	printf("\n");
}

static void cmd_lspci(void)
{
	unsigned int dev;
	unsigned int fun;
	int ret;
	int vendor_id;
	int device_id;
	u32 class_rev;

	for (dev = 0; dev < 16; dev++) {
		for (fun = 0; fun < 8; fun++) {
			ret = arbus_card_present(dev, fun);
			if (!ret)
				break;

			vendor_id = ar2c_readw(ARBUS_REG(dev, fun,
						PCI_VENDOR_ID));
			if (vendor_id < 0)
				break;
			device_id = ar2c_readw(ARBUS_REG(dev, fun,
						PCI_DEVICE_ID));
			if (device_id < 0)
				break;

			ret = ar2c_readl(ARBUS_REG(dev, fun,
						PCI_CLASS_REVISION),
					&class_rev);
			if (ret)
				break;

			printf("%02x:%02x.%d %04x: %04x:%04x", 0, dev, fun,
					class_rev >> 16, vendor_id, device_id);
			if (class_rev & 0xff)
				printf(" (rev %02x)", class_rev & 0xff);
			printf("\n");

			ar2c_writel(0xd8000, ARBUS_REG(dev, fun,
						PCI_BASE_ADDRESS_1));
			ar2c_writew(PCI_COMMAND_MEMORY, ARBUS_REG(dev, fun,
						PCI_COMMAND));
			ar2c_writeb(0, 0xd8000);

			if (lspci_hexdump)
				do_lspci_hexdump(ARBUS_REG(dev, fun, 0));
		}
	}
}

static void setup_device(struct arbus_dev *dev,
		unsigned int domain, unsigned int bus,
		unsigned int device, unsigned int fun)
{
	unsigned int i;
	int vendor_id;
	int device_id;
	char *slot;

	if (domain || bus)
		fatal("Only bus 0000:00 is supported\n");

	dev->valid = 0;
	slot = malloc(64);
	if (slot == NULL)
		fatal("malloc error");
	dev->slot = slot;
	sprintf(slot, "%02x:%02x.%d", bus, device, fun);
	dev->devfn = (device << 3) | fun;

	vendor_id = arbus_read_word(dev, PCI_VENDOR_ID);
	if (vendor_id < 0)
		fatal("I/O error\n");
	device_id = arbus_read_word(dev, PCI_DEVICE_ID);
	if (device_id < 0)
		fatal("I/O error\n");

	fprintf(stderr, "Card %02x.%d  %04x:%04x\n", device, fun,
			vendor_id, device_id);
	dev->vendor_id = vendor_id;
	dev->device_id = device_id;

	for (i = 0; i < 6; i++)
		dev->base_addr[i] = arbus_read_long(dev,
				PCI_BASE_ADDRESS_0 + 4 * i);
	for (i = 0; i < 6; i++)
		if (dev->base_addr[i])
			fprintf(stderr, "BAR%d: %08lx\n", i, dev->base_addr[i]);
}

static void do_lspci_hexdumpx(struct arbus_dev *dev)
{
	unsigned int i, j;
	u16 vendor_id;
	u16 device_id;
	u32 class_rev;

	if (!dev->valid)
		return;

	vendor_id = arbus_read_word(dev, PCI_VENDOR_ID);
	device_id = arbus_read_word(dev, PCI_DEVICE_ID);
	class_rev = arbus_read_long(dev, PCI_CLASS_REVISION);
			
	printf("%02x:%02x.%d %04x: %04x:%04x", 0, PCI_SLOT(dev->devfn),
			PCI_FUNC(dev->devfn), class_rev >> 16,
			vendor_id, device_id);
	if (class_rev & 0xff)
		printf(" (rev %02x)", class_rev & 0xff);
	printf("\n");

	for (i = 0; i < sizeof(dev->config_regs); i += 16) {
		printf("%02x:", i);
		for (j = 0; j < 16; j++)
			printf(" %02x", dev->config_regs[i + j]);
		printf("\n");
	}
	printf("\n");
}

static void cmd_lspcix(void)
{
	struct arbus_dev dev;
	unsigned int device;
	unsigned int fun;
	int ret;
	pid_t pid;
	int p[2];

	ret = pipe(p);
	if (ret) {
		perror("pipe");
		exit(EXIT_FAILURE);
	}

	pid = fork();
	if (pid < 0) {
		perror("fork");
		exit(EXIT_FAILURE);
	}

	if (!pid) {
		char pci_ids[PATH_MAX];

		close(p[1]);
		dup2(p[0], STDIN_FILENO);

		snprintf(pci_ids, sizeof(pci_ids), "%s/pci.ids",
				AR2CPARUTILSDIR);
		execlp("lspci", "lspci", "-i", pci_ids, "-F", "/dev/stdin",
				"-v", NULL);
		exit(EXIT_FAILURE);
	}
	close(p[0]);
	fflush(stdout);

	dup2(p[1], STDOUT_FILENO);
	close(p[1]);
	for (device = 0; device < 16; device++) {
		for (fun = 0; fun < 8; fun++) {
			ret = arbus_card_present(device, fun);
			if (!ret)
				break;

			setup_device(&dev, pci_domain, pci_bus, device, fun);
			do_lspci_hexdumpx(&dev);
		}
	}
	fflush(stdout);
	close(STDOUT_FILENO);
	wait(NULL);
}

static unsigned int card_id;

static void get_config_dev(struct arbus_dev *dev, struct ar2c_config *c)
{
	unsigned int id;
	char name[128];
	char slot[128];
	unsigned int i;
	u16 x;

	id = card_id++;

	sprintf(slot, "%02x.%d", PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn));

	snprintf(name, sizeof(name), "card %d", id);
	ar2c_config_set_all_fmt(c, name, "arbus", 1, "PCI %s", slot);

	snprintf(name, sizeof(name), "PCI %s VendorID", slot);
	ar2c_config_set_all_fmt(c, name, "arbus", 1, "0x%04x", dev->vendor_id);
	snprintf(name, sizeof(name), "PCI %s DeviceID", slot);
	ar2c_config_set_all_fmt(c, name, "arbus", 1, "0x%04x", dev->device_id);

	x = arbus_read_word(dev, PCI_SUBSYSTEM_VENDOR_ID);
	if (x) {
		snprintf(name, sizeof(name), "PCI %s SubsystemVendorID", slot);
		ar2c_config_set_all_fmt(c, name, "arbus", 1, "0x%04x", x);
		x = arbus_read_word(dev, PCI_SUBSYSTEM_ID);
		snprintf(name, sizeof(name), "PCI %s SubsystemID", slot);
		ar2c_config_set_all_fmt(c, name, "arbus", 1, "0x%04x", x);
	}

	for (i = 0; i < 6; i++) {
		snprintf(name, sizeof(name), "PCI %s BAR%d", slot, i);
		ar2c_config_set_all_fmt(c, name, "arbus", 1, "0x%08x",
				dev->base_addr[i]);
	}

	if (dev->vendor_id == PCI_VENDOR_ID_MICROSTER
			&& dev->device_id == PCI_DEVICE_ID_MICROSTER_POMMAX2) {
		pommax3_getconfig(dev, c);
	} else {
		fprintf(stderr, "warning: unsupported device %04x:%04x\n",
				dev->vendor_id, dev->device_id);
	}
}

static void put_config_dev(struct arbus_dev *dev, struct ar2c_config *c)
{
	if (dev->vendor_id == PCI_VENDOR_ID_MICROSTER
			&& dev->device_id == PCI_DEVICE_ID_MICROSTER_POMMAX2) {
		pommax3_putconfig(dev, c);
	} else {
		fprintf(stderr, "warning: unsupported device %04x:%04x\n",
				dev->vendor_id, dev->device_id);
	}
}

static void get_all_config(void)
{
	struct arbus_dev dev;
	struct ar2c_config c;
	unsigned int device;
	unsigned int fun;
	FILE *f;
	int ret;
	unsigned int number;

	ret = ar2c_config_init(&c);
	if (ret) {
		fprintf(stderr, "ar2c_config_init: %s\n", strerror(-ret));
		exit(EXIT_FAILURE);
	}

	ret = ar2c_get_id(ifd, ofd);
	if (ret < 0) {
		fprintf(stderr, "ar2c_get_id: %s\n", strerror(-ret));
		exit(EXIT_FAILURE);
	}
	number = ret;

	for (device = 0; device < 16; device++) {
		for (fun = 0; fun < 8; fun++) {
			if (!arbus_card_present(device, fun))
				break;

			setup_device(&dev, pci_domain, pci_bus, device, fun);
			get_config_dev(&dev, &c);
		}
	}

	ret = ar2c_get_id(ifd, ofd);
	if (ret < 0) {
		fprintf(stderr, "ar2c_get_id: %s\n", strerror(-ret));
		exit(EXIT_FAILURE);
	}
	if (ret != number) {
		fprintf(stderr, "error: Inconsistent datalogger number "
				"%d != %d\n", number, ret);
		exit(EXIT_FAILURE);
	}
	ar2c_config_set_all_fmt(&c, "number", "arbus", 3, "%d", number);

	if (config_file) {
		f = fopen(config_file, "w");
		if (f == NULL) {
			fprintf(stderr, "cannot open %s: %s\n", config_file,
					strerror(errno));
		}
	} else {
		f = stdout;
	}
	ret = ar2c_config_store(&c, f);
	if (ret) {
		fprintf(stderr, "ar2c_config_store: %s\n", strerror(-ret));
		exit(EXIT_FAILURE);
	}

	if (config_file) {
		ret = fclose(f);
		if (ret) {
			fprintf(stderr, "fclose: %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
	}
}

static void get_config(void)
{
	struct arbus_dev dev;
	struct ar2c_config c;
	FILE *f;
	int ret;

	ret = ar2c_config_init(&c);
	if (ret) {
		fprintf(stderr, "ar2c_config_init: %s\n", strerror(-ret));
		exit(EXIT_FAILURE);
	}

	if (config_file) {
		f = fopen(config_file, "r");
		if (f != NULL) {
			ret = ar2c_config_load(&c, f);
			if (ret) {
				fprintf(stderr, "ar2c_config_load: %s\n",
						strerror(-ret));
				exit(EXIT_FAILURE);
			}
			fclose(f);
		}
	}

	fprintf(stderr, "%04x:%02x:%02x.%d\n: get config",
			pci_domain, pci_bus, pci_dev, pci_fun);

	if (!arbus_card_present(pci_dev, pci_fun))
		fatal("Card %02x.%d not present\n", pci_dev, pci_fun);

	setup_device(&dev, pci_domain, pci_bus, pci_dev, pci_fun);

	get_config_dev(&dev, &c);
	if (config_file) {
		f = fopen(config_file, "w");
		if (f == NULL) {
			fprintf(stderr, "cannot open %s: %s\n", config_file,
					strerror(errno));
		}
	} else {
		f = stdout;
	}
	ret = ar2c_config_store(&c, f);
	if (ret) {
		fprintf(stderr, "ar2c_config_store: %s\n", strerror(-ret));
		exit(EXIT_FAILURE);
	}

	if (config_file) {
		ret = fclose(f);
		if (ret) {
			fprintf(stderr, "fclose: %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
	}
}

static void put_all_config(void)
{
	struct arbus_dev dev;
	struct ar2c_config c;
	unsigned int device;
	unsigned int fun;
	FILE *f;
	int ret;

	ret = ar2c_config_init(&c);
	if (ret) {
		fprintf(stderr, "ar2c_config_init: %s\n", strerror(-ret));
		exit(EXIT_FAILURE);
	}

	if (config_file) {
		f = fopen(config_file, "r");
		if (f == NULL) {
			fprintf(stderr, "cannot open %s: %s\n", config_file,
					strerror(errno));
		}
	} else
		f = stdin;

	ret = ar2c_config_load(&c, f);
	if (ret) {
		fprintf(stderr, "ar2c_config_load: %s\n",
				strerror(-ret));
		exit(EXIT_FAILURE);
	}
	if (config_file)
		fclose(f);


	for (device = 0; device < 16; device++) {
		for (fun = 0; fun < 8; fun++) {
			if (!arbus_card_present(device, fun))
				break;

			setup_device(&dev, pci_domain, pci_bus, device, fun);
			put_config_dev(&dev, &c);
		}
	}
}

static void put_config(void)
{
	struct arbus_dev dev;
	struct ar2c_config c;
	FILE *f;
	int ret;

	ret = ar2c_config_init(&c);
	if (ret) {
		fprintf(stderr, "ar2c_config_init: %s\n", strerror(-ret));
		exit(EXIT_FAILURE);
	}

	if (config_file) {
		f = fopen(config_file, "r");
		if (f == NULL) {
			fprintf(stderr, "cannot open %s: %s\n", config_file,
					strerror(errno));
		}
	} else
		f = stdin;

	ret = ar2c_config_load(&c, f);
	if (ret) {
		fprintf(stderr, "ar2c_config_load: %s\n",
				strerror(-ret));
		exit(EXIT_FAILURE);
	}
	if (config_file)
		fclose(f);

	fprintf(stderr, "%04x:%02x:%02x.%d\n: get config",
			pci_domain, pci_bus, pci_dev, pci_fun);
	setup_device(&dev, pci_domain, pci_bus, pci_dev, pci_fun);

	put_config_dev(&dev, &c);
}

static void get_data_dev(struct arbus_dev *dev, struct ar2c_config *c)
{
	if (dev->vendor_id == PCI_VENDOR_ID_MICROSTER
			&& dev->device_id == PCI_DEVICE_ID_MICROSTER_POMMAX2) {
		pommax3_get_data(dev, c);
	} else {
		fprintf(stderr, "warning: unsupported device %04x:%04x\n",
				dev->vendor_id, dev->device_id);
	}
}

static void get_all_data(void)
{
	struct arbus_dev dev;
	struct ar2c_config c;
	unsigned int device;
	unsigned int fun;
	FILE *f;
	int ret;

	ret = ar2c_config_init(&c);
	if (ret) {
		fprintf(stderr, "ar2c_config_init: %s\n", strerror(-ret));
		exit(EXIT_FAILURE);
	}

	for (device = 0; device < 16; device++) {
		for (fun = 0; fun < 8; fun++) {
			if (!arbus_card_present(device, fun))
				break;

			setup_device(&dev, pci_domain, pci_bus, device, fun);
			get_data_dev(&dev, &c);
		}
	}

	if (config_file) {
		f = fopen(config_file, "w");
		if (f == NULL) {
			fprintf(stderr, "cannot open %s: %s\n", config_file,
					strerror(errno));
		}
	} else {
		f = stdout;
	}
	ret = ar2c_config_store(&c, f);
	if (ret) {
		fprintf(stderr, "ar2c_config_store: %s\n", strerror(-ret));
		exit(EXIT_FAILURE);
	}

	if (config_file) {
		ret = fclose(f);
		if (ret) {
			fprintf(stderr, "fclose: %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
	}
}

static void configure_device(void)
{
	struct arbus_dev dev;

	fprintf(stderr, "configuring device %04x:%02x:%02x.%d\n",
			pci_domain, pci_bus, pci_dev, pci_fun);
	setup_device(&dev, pci_domain, pci_bus, pci_dev, pci_fun);

	if (dev.vendor_id == PCI_VENDOR_ID_MICROSTER
			&& dev.device_id == PCI_DEVICE_ID_MICROSTER_POMMAX2) {
		configure_pommax3(&dev);
	} else {
		fprintf(stderr, "unsupported device model %04x:%04x\n",
				dev.vendor_id, dev.device_id);
	}
}
	
static void save_config_dev(struct arbus_dev *dev)
{
	if (dev->vendor_id == PCI_VENDOR_ID_MICROSTER
			&& dev->device_id == PCI_DEVICE_ID_MICROSTER_POMMAX2) {
		pommax3_save(dev);
	} else {
		fprintf(stderr, "unsupported device model %04x:%04x\n",
				dev->vendor_id, dev->device_id);
	}
}

static void save_config(void)
{
	struct arbus_dev dev;

	fprintf(stderr, "saving configuration of device %04x:%02x:%02x.%d\n",
			pci_domain, pci_bus, pci_dev, pci_fun);
	setup_device(&dev, pci_domain, pci_bus, pci_dev, pci_fun);

	save_config_dev(&dev);
}

static void save_all_config(void)
{
	struct arbus_dev dev;
	unsigned int device;
	unsigned int fun;

	for (device = 0; device < 16; device++) {
		for (fun = 0; fun < 8; fun++) {
			if (!arbus_card_present(device, fun))
				break;

			setup_device(&dev, pci_domain, pci_bus, device, fun);
			save_config_dev(&dev);
		}
	}
}

int main(int argc, char **argv)
{
	signal(SIGALRM, SIG_IGN);
	parse_command_line(argc, argv);
	pcomm_use_nodelay(1);

	mclock_init();

	switch (targ_type) {
	case TT_STDIN:
		ifd = 0;
		ofd = 1;
		out = stderr;
		cmd_timeout = cmd_timeout ? : 2000;
		break;
	case TT_FILE:
		if (targ_opt == NULL) {
			fprintf(stderr, "fatal: null targ_opt.\n");
			exit(EXIT_FAILURE);
		}
		ifd = open(targ_opt, O_RDWR);
		if (ifd < 0) {
			fprintf(stderr, "error: Can't open %s:%s\n",
					targ_opt, strerror(errno));
			exit(EXIT_FAILURE);
		}
		ofd = ifd;
		out = stdout;
		cmd_timeout = cmd_timeout ? : 2000;
		break;
	case TT_TCP:
		if (targ_opt == NULL) {
			fprintf(stderr, "fatal: null targ_opt.\n");
			exit(EXIT_FAILURE);
		}
		ifd = pcomm_open(targ_opt, port);
		if (ifd < 0) {
			fprintf(stderr, "error: Can't connect to %s\n",
					targ_opt);
			exit(EXIT_FAILURE);
		}
		ofd = ifd;
		out = stdout;
		cmd_timeout = cmd_timeout ? : 10000;
		break;
	case TT_SERIAL:
		if (targ_opt == NULL) {
			fprintf(stderr, "fatal: null targ_opt.\n");
			exit(EXIT_FAILURE);
		}
		ifd = open(targ_opt, O_RDWR);
		if (ifd == -1) {
			fprintf(stderr, "error: Can't open %s:%s\n",
					targ_opt, strerror(errno));
			exit(EXIT_FAILURE);
		}
		ofd = ifd;
		serial_set_speed(ifd, bspeed);
		if (ar2cspeedtospeed(speed) != bspeed)
			if (ar2c_set_speed(ifd, speed)) {
				fprintf(stderr, "Setting serial speed "
						"error\n");
				exit(EXIT_FAILURE);
			}
		out = stdout;
		cmd_timeout = cmd_timeout ? : 500;
		break;
	default:
		fprintf(stderr, "fatal: Internal error.\n");
		exit(EXIT_FAILURE);
	}

	atexit(spd_return);

	do_resync(ifd, ofd);

	if (command == NULL) {
		cmd_lspci();
	} else if (!strcmp(command, "lspci")) {
		cmd_lspci();
	} else if (!strcmp(command, "lspcix")) {
		cmd_lspcix();
	} else if (!strcmp(command, "get-config-dev")) {
		get_config();
	} else if (!strcmp(command, "get-config")) {
		get_all_config();
	} else if (!strcmp(command, "put-config-dev")) {
		put_config();
	} else if (!strcmp(command, "put-config")) {
		put_all_config();
	} else if (!strcmp(command, "save-dev")) {
		save_config();
	} else if (!strcmp(command, "save")) {
		save_all_config();
	} else if (!strcmp(command, "config")) {
		configure_device();
	} else if (!strcmp(command, "get-data")) {
		get_all_data();
	} else {
		fprintf(stderr, "error: unknown command %s\n", command);
		exit(EXIT_FAILURE);
	}
	return exit_code;
}
