/*
 * freetrans - comm.c
 * 	version 0.1
 * 	
 * 	Common TCP/IP Communication
 * 
 * Copyright (C) 2004, 2005, 2006  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <comm.h>

static int use_nodelay = 0;

void pcomm_use_nodelay(int nodelay)
{
	use_nodelay = nodelay;
}

static int do_connect(struct addrinfo *ai)
{
	int ret;
	int fd;

	fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
	if (fd == -1) {
		ret = errno;
		goto err;
	}

	ret = connect(fd, ai->ai_addr, ai->ai_addrlen);
	if (ret < 0) {
		ret = -errno;
		fprintf(stderr, "connect: %s\n", strerror(errno));
		goto err_close;
	}
	return fd;

err_close:
	close(fd);
err:
	return ret;
}

static int my_connect(const char *host, const char *port)
{
	static struct addrinfo *ai, *aitop;
	struct addrinfo hints;
	int ret;

	if (!ai) {
		if (aitop) {
			freeaddrinfo(aitop);
			aitop = NULL;
		}

		memset(&hints, 0, sizeof(hints));
		hints.ai_family = PF_UNSPEC;
		hints.ai_socktype = SOCK_STREAM;
		hints.ai_flags = 0;
		ret = getaddrinfo(host, port, &hints, &aitop);
		if (ret < 0) {
			fprintf(stderr, "getaddrinfo: %s\n", strerror(errno));
			return -errno;
		}
		ai = aitop;
	}

	ret = -ENOENT;
	for (; ai != NULL; ai = ai->ai_next) {
		ret = do_connect(ai);
		if (ret >= 0)
			break;
	}
	return ret;
}

int pcomm_open(const char *hostname, const char *port)
{
	int fd;
	int op;

	fd = my_connect(hostname, port);
	if (fd < 0)
		return fd;

	if (use_nodelay) {
		op = 1;
		setsockopt(fd, SOL_TCP, TCP_NODELAY, &op, sizeof(op));
	}

	return fd;
}

int pcomm_close(int fd)
{
	return close(fd);
}

#ifndef HAVE_CFMAKERAW
static void cfmakeraw(struct termios *termios_p)
{
	termios_p->c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR
			| IGNCR | ICRNL | IXON);
	termios_p->c_oflag &= ~OPOST;
	termios_p->c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
	termios_p->c_cflag &= ~(CSIZE | PARENB);
	termios_p->c_cflag |= CS8;
}
#endif

void serial_set_speed_o(int fd, speed_t speed, int opts)
{
	struct termios mode;

	memset(&mode, 0, sizeof(struct termios));
	cfmakeraw(&mode);
	mode.c_cflag |= CLOCAL | CREAD | CS8;
	if (opts & COMM_PAR_PAR) {
		mode.c_cflag |= PARENB;
		if (COMM_PAR_EVEN)
			mode.c_cflag &= ~PARODD;
	}
	if (opts & COMM_PAR_CSTOPB)
		mode.c_cflag &= ~CSTOPB;
	mode.c_oflag = 0;
	cfsetispeed(&mode, speed);
	cfsetospeed(&mode, speed);
	tcflush(fd, TCIFLUSH);
	tcsetattr(fd, TCSANOW, &mode);
}

void serial_set_speed(int fd, speed_t speed)
{
	serial_set_speed_o(fd, speed, COMM_PAR_EVEN | COMM_PAR_CSTOPB);
}

speed_t itospeed(int speed)
{
	switch (speed) {
		case 50: return B50;
		case 75: return B75;
		case 110: return B110;
		case 134: return B134;
		case 150: return B150;
		case 200: return B200;
		case 300: return B300;
		case 600: return B600;
		case 1200: return B1200;
		case 2400: return B2400;
		case 4800: return B4800;
		case 9600: return B9600;
		case 19200: return B19200;
		case 38400: return B38400;
		case 57600: return B57600;
		case 115200: return B115200;
		case 230400: return B230400;
		case 460800: return B460800;
	}
	return -EINVAL;
}
