/*
 * ar2c-par-utils - gconf.c
 *
 * Copyright (C) 2012..2015  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <signal.h>
#include <inttypes.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <ar2cpar.h>
#include <math.h>
#include <locale.h>
#include <gettext.h>

#if HAVE_LIBPCI
#include <pci/pci.h>
#else
typedef uint16_t u16;

struct pci_access;

enum pci_lookup_mode {
	PCI_LOOKUP_VENDOR = 1,
	PCI_LOOKUP_DEVICE = 2,
	PCI_LOOKUP_SUBSYSTEM = 8,
};

#define pci_alloc() (NULL)
#define pci_init(pacc) do { } while (0)
#define pci_set_name_list_path(pacc, path, to_free)

static char *pci_lookup_name(struct pci_access *pacc, char *buf, size_t len,
		...)
{
	snprintf(buf, len, "unknown");
	return buf;
}
#endif

static struct pci_access *pacc;

#define GETTEXT_PACKAGE "ar2c-par-utils"
#include <glib/gi18n.h>

#define GTK_DISABLE_DEPRECATED	1
#include <gtk/gtk.h>
#include <gtk/gtkfilesel.h>
#include <glib.h>
#include <gdk/gdkkeysyms.h>

#ifndef likely
#if __GNUC__ > 2
#define likely(x)	__builtin_expect(!!(x), 1)
#define unlikely(x)	__builtin_expect(!!(x), 0)
#else
#define likely(x)	(x)
#define unlikely(x)	(x)
#endif
#endif

#ifndef ARRAY_SIZE
#define ARRAY_SIZE(x) (sizeof((x)) / sizeof((x)[0]))
#endif

#if defined _MSC_VER || defined __MINGW32__

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

/* The return value of spawnvp() is really a process handle as returned
   by CreateProcess().  Therefore we can kill it using TerminateProcess.  */
#define kill(pid,sig) TerminateProcess ((HANDLE) (pid), sig)

#endif

#define POMMAX3_FLAGS_GAIN		1
#define POMMAX3_FLAGS_OFFSET		2

#define ADC_CHANNELS		8

static int silent_level = 1;
static struct ar2c_config ar2c_config;
static struct ar2c_config ar2c_data;
static char *config_fname;

#define ERROR_ON(cond, msg...)						\
	do {								\
		if (unlikely(cond)) {					\
			fprintf(stderr, msg);				\
			exit(EXIT_FAILURE);				\
		}							\
	} while (0)

enum conn_type {
	CONN_NONE,
	CONN_TCP,
	CONN_FILE,
};

static char *tcp_port;
static char *tcp_host;
static char *file_path;
static char *file_baudrate;
static enum conn_type conn_type = CONN_NONE;

GtkWidget *main_wnd;
GtkWidget *vendor_id_w;
GtkWidget *device_id_w;
GtkWidget *subsystem_vendor_id_w;
GtkWidget *subsystem_id_w;
GtkWidget *pci_bar_w[6];

GtkWidget *comm_w;

GtkWidget *pommax3_wnd;
GtkWidget *pommax3_card_id_w;
GtkWidget *pommax3_freq_w;
GtkWidget *pommax3_adc_offset_w;
GtkWidget *pommax3_serial_w;
GtkWidget *pommax3_gain_w[ADC_CHANNELS];
GtkWidget *pommax3_init_gain_w[ADC_CHANNELS];
int pommax3_init_ofix[ADC_CHANNELS];
GtkWidget *pommax3_data_w[ADC_CHANNELS];
GtkWidget *pommax3_adata_w[ADC_CHANNELS];
GtkWidget *pommax3_range_w[ADC_CHANNELS];
GtkWidget *pommax3_unit_w[ADC_CHANNELS];
GtkWidget *pommax3_offset_w[ADC_CHANNELS];
GtkWidget *pommax3_ar2co_w[ADC_CHANNELS];
GtkWidget *pommax3_value_w[ADC_CHANNELS];
GtkWidget *pommax3_gain_enable;
GtkWidget *pommax3_offset_enable;
GtkWidget *pommax3_adc_combo;

GtkWidget *xfer_log_view;
GtkWidget *xfer_ok;
GtkTextBuffer *xfer_log;

GtkWidget *preferences_wnd;
GObject *conn_tcp_host_w;
GObject *conn_tcp_port_w;
GObject *conn_file_path_w;
GObject *conn_file_baudrate_w;
GObject *conn_tcp_w;
GObject *conn_file_w;

GtkWidget *xfer_w;
GtkWidget *xfer_spinner_w;

static char *pommax3_card_id;

#if defined _MSC_VER || defined __MINGW32__
static char *mingw_getenv(const char *name)
{
	char *result = getenv(name);
	if (!result && !strcmp(name, "TMPDIR")) {
		/* on Windows it is TMP and TEMP */
		result = getenv("TMP");
		if (!result)
			result = getenv("TEMP");
	}
	return result;
}
#define getenv mingw_getenv
#endif

/* git_mkstemp() - create tmp file honoring TMPDIR variable */
static int git_mkstemp(char *path, size_t len, const char *template)
{
	const char *tmp;
	size_t n;

	tmp = getenv("TMPDIR");
	if (!tmp)
		tmp = "/tmp";
	n = snprintf(path, len, "%s/%s", tmp, template);
	if (len <= n) {
		errno = ENAMETOOLONG;
		return -1;
	}
	return mkstemp(path);
}

static void show_usage(char *name)
{
	fprintf(stderr,	"usage: %s [options]\n"
			"Options:\n"
			"  -A, --ascii		ASCII mode\n"
			"  -B, --bbitrate	base bitrate(default 9600bps)\n"
			"  -b, --bitrate\t	set bitrate\n"
			"  -d, --device		use selected device\n"
			"  -h, --help		show this help\n"
			"      --parity		set parity (none, odd, even)\n"
			"  -p, --port		remote port(2000)\n"
			"  -q, --quiet		be quiet\n"
			"      --rbitrate	return bitrate\n"
			"  -s, --host		remote host(localhost)\n"
			"      --stopbits	stop bits (1/2)\n"
			"  -u, --udp		use UDP\n"
			"  -v, --verbose		be verbose\n"
			"  -V, --version		show version\n",
			(name == NULL) ? "ar2c-gconf" : name);
	exit(EXIT_SUCCESS);
}

static void show_version(void)
{
	fprintf(stderr,	"This program comes with ABSOLUTELY NO WARRANTY; "
			"see COPYING for details.\n"
			"This is free software, and you are welcome to "
			"redistribute it\n"
			"under certain conditions; see COPYING for details.\n");
}

static void parse_command_line(int argc, char **args)
{
	int c;
	int option_index = 0;

	static struct option long_options[] = {
		{"help", 0, 0, 'h'},
		{"quiet", 0, 0, 'q'},
		{"verbose", 0, 0, 'v'},
		{"version", 0, 0, 'V'},
		{0, 0, 0, 0},
	};

	while (1) {
		c = getopt_long(argc, args, "hqvV",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'h':
			show_usage(args[0]);
			break;
		case 'v':
			silent_level++;
			break;
		case 'q':
			silent_level--;
			break;
		case 'V':
			show_version();
			exit(EXIT_SUCCESS);
			break;
		default:
			fprintf(stderr, "Unknown option.\n");
			exit(EXIT_FAILURE);
		}
	}
	if (optind < argc)
		fprintf(stderr, "warning: unknown parameters detected.\n");
}

GtkWidget *cards;
GtkListStore *cards_store;

static char *make_stringv(const char *fmt, va_list ap)
{
	size_t size;
	char *value = NULL;
	va_list ap2;
	char *p;
	int n;

	size = 256;
	value = malloc(size);
	if (value == NULL)
		return NULL;

	while (1) {
		va_copy(ap2, ap);
		n = vsnprintf(value, size, fmt, ap2);
		va_end(ap2);

		if (n > -1 && n < size)
			break;

		if (n > -1)
			size = n + 1;
		else
			size *= 2;

		p = realloc(value, size);
		if (p == NULL) {
			free(value);
			return NULL;
		}
		value = p;
	}
	return value;
}

static int remake_string(char **x, const char *fmt, ...)
{
	size_t size;
	char *value = NULL;
	char *p;
	va_list ap;
	int n;

	size = 256;
	value = malloc(size);
	if (value == NULL)
		return -ENOMEM;

	while (1) {
		va_start(ap, fmt);
		n = vsnprintf(value, size, fmt, ap);
		va_end(ap);

		if (n > -1 && n < size)
			break;

		if (n > -1)
			size = n + 1;
		else
			size *= 2;

		p = realloc(value, size);
		if (p == NULL) {
			free(value);
			return -ENOMEM;
		}
		value = p;
	}
	free(*x);
	*x = value;
	return 0;
}

static void gconf_error(const char *title, const char *fmt, ...)
{
	GtkWidget *dialog;
	va_list ap;
	char *msg;

	if (title)
		fprintf(stderr, "%s ", title);

	fprintf(stderr, "error: ");
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fprintf(stderr, "\n");

	if (title == NULL)
		title = _("Error");

	va_start(ap, fmt);
	msg = make_stringv(fmt, ap);
	va_end(ap);

	dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL,
			GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
			"%s", msg);
	gtk_window_set_title(GTK_WINDOW(dialog), title);
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);

	free(msg);
}

#define AR2C_GCONF_CONFIG_FILENAME		"ar2c-gconf.conf"
static GKeyFile *gconf_config;

static void gconf_load_config(void)
{
	gchar *filename;
	gchar *type;

	gconf_config = g_key_file_new();
	filename = g_build_filename(g_get_user_config_dir(),
			AR2C_GCONF_CONFIG_FILENAME, NULL);
	g_key_file_load_from_file(gconf_config, filename,
			G_KEY_FILE_KEEP_COMMENTS, NULL);
	g_free(filename);

	tcp_host = g_key_file_get_value(gconf_config, "comm", "tcp_host",
			NULL);
	tcp_port = g_key_file_get_value(gconf_config, "comm", "tcp_port",
			NULL);
	file_path = g_key_file_get_value(gconf_config, "comm", "file_path",
			NULL);
	file_baudrate = g_key_file_get_value(gconf_config,
			"comm", "file_baudrate", NULL);

	type = g_key_file_get_value(gconf_config, "comm", "type", NULL);
	if (type == NULL) {
		/* force user to configure link */
	} else if (!strcmp(type, "tcp")) {
		conn_type = CONN_TCP;
	} else if (!strcmp(type, "file")) {
		conn_type = CONN_FILE;
	}
	g_free(type);
}

#define GCONF_WRITE_ERR _("ar2c-gconfig configuration write error")
/*
 * from Glade 3.8.3
 */
gint gconf_config_save(void)
{
	GIOChannel *channel;
	GIOStatus status;
	gchar *data = NULL, *filename;
	const gchar *config_dir = g_get_user_config_dir();
	GError *error = NULL;
	gsize size, written, bytes_written = 0;
	static gboolean error_shown = FALSE;

	/*
	 * If we had any errors; wait untill next session to retry.
	 */
	if (error_shown)
		return -1;

	/* Just in case... try to create the config directory */
	if (g_file_test(config_dir, G_FILE_TEST_IS_DIR) == FALSE) {
		if (g_file_test(config_dir, G_FILE_TEST_EXISTS)) {
			/* Config dir exists but is not a directory */
			gconf_error(GCONF_WRITE_ERR,
				_("Trying to save private data to %s "
				"directory but it is a regular file.\n"
				"No private data will be saved in this "
				"session"), config_dir);
			error_shown = TRUE;
			return -1;
		} else if (mkdir(config_dir, S_IRWXU) != 0) {
			/* Doesnt exist; failed to create */
			gconf_error(GCONF_WRITE_ERR,
				_("Failed to create directory %s to save "
				"private data.\n"
				"No private data will be saved in this "
				"session"), config_dir);
			error_shown = TRUE;
			return -1;
		}
	}

	filename = g_build_filename(config_dir, AR2C_GCONF_CONFIG_FILENAME,
			NULL);

	channel = g_io_channel_new_file(filename, "w", &error);
	if (channel != NULL) {
		data = g_key_file_to_data(gconf_config, &size, &error);
		if (data != NULL) {
			/* Implement loop here */
			while ((status = g_io_channel_write_chars(channel,
				data + bytes_written, /* Offset of write */
				size - bytes_written, /* Size left to write */
				&written, &error)) != G_IO_STATUS_ERROR
					&& (bytes_written + written) < size)
				bytes_written += written;

			if (status == G_IO_STATUS_ERROR) {
				gconf_error(GCONF_WRITE_ERR,
					_("Error writing private data to %s "
					"(%s).\n"
					"No private data will be saved in "
					"this session"),
					filename, error->message);
				 error_shown = TRUE;
			}
			g_free(data);
		} else {
			gconf_error(GCONF_WRITE_ERR,
				_("Error serializing configuration data "
				"to save (%s).\n"
				"No private data will be saved in this "
				"session"), error->message);
			error_shown = TRUE;
		}
		g_io_channel_shutdown(channel, TRUE, NULL);
		g_io_channel_unref(channel);
	} else {
		gconf_error(GCONF_WRITE_ERR,
			_("Error opening %s to write private data (%s).\n"
			"No private data will be saved in this session"),
			filename, error->message);
		error_shown = TRUE;
	}
	g_free(filename);

	if (error) {
		g_error_free(error);
		return -1;
	}
	return 0;
}

int ar2c_config_get_int(struct ar2c_config *c, const char *name)
{
	const char *v;
	char *end = NULL;
	int ret;

	v = __ar2c_config_get(c, name);
	if (!v)
		return -1;
	ret = strtol(v, &end, 0);
	if (*end != '\0')
		return -1;
	return ret;
}

static int make_card_name(struct ar2c_config *c, const char *card_id,
		char **card_name)
{
	void *p;
	char *name = NULL;
	int vendor_id;
	int device_id;
	int subvendor_id;
	int subsystem_id;

	remake_string(&name, "%s VendorID", card_id);
	vendor_id = ar2c_config_get_int(c, name);
	remake_string(&name, "%s DeviceID", card_id);
	device_id = ar2c_config_get_int(c, name);
	remake_string(&name, "%s SubsystemVendorID", card_id);
	subvendor_id = ar2c_config_get_int(c, name);
	remake_string(&name, "%s SubsystemID", card_id);
	subsystem_id = ar2c_config_get_int(c, name);
	free(name);

	if (vendor_id == -1 || device_id == -1
			|| subvendor_id == -1 || subsystem_id == -1) {
		return remake_string(card_name, "unknown");
	}

	p = malloc(256);
	if (!p)
		return -ENOMEM;

	pci_lookup_name(pacc, p, 256,
			PCI_LOOKUP_SUBSYSTEM | PCI_LOOKUP_VENDOR
			| PCI_LOOKUP_DEVICE,
			vendor_id, device_id, subvendor_id, subsystem_id);
	free(*card_name);
	*card_name = p;
	return 0;
}

static void reload_cards(void)
{
	GtkTreeIter iter;
	char card[128];
	const char *card_id;
	char *name = NULL;
	char *dev_name = NULL;
	unsigned int i;

	gtk_list_store_clear(cards_store);

	for (i = 0; ; i++) {
		snprintf(card, sizeof(card), "card %d", i);
		card_id = __ar2c_config_get(&ar2c_config, card);
		if (!card_id)
			break;

		make_card_name(&ar2c_config, card_id, &name);

		remake_string(&dev_name, "%s (%s)",
				card_id, name);

		gtk_list_store_append(cards_store, &iter);
		gtk_list_store_set(cards_store, &iter, 0, dev_name,
				1, card_id, -1);
	}

	if (i)
		gtk_combo_box_set_active(GTK_COMBO_BOX(cards), 0);

	free(name);
	free(dev_name);
}

static void init_cards_list(void)
{
	GtkCellRenderer *cell;

	cards_store = gtk_list_store_new(2, G_TYPE_STRING, G_TYPE_STRING);
	reload_cards();

	gtk_combo_box_set_model(GTK_COMBO_BOX(cards),
			GTK_TREE_MODEL(cards_store));
	g_object_unref(G_OBJECT(cards_store));

	cell = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(cards), cell, TRUE);
	gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(cards), cell,
			"text", 0, NULL);
}

void open_file(const char *filename)
{
	FILE *f;

	f = fopen(filename, "r");
	if (f == NULL) {
		fprintf(stderr, "error\n");
		return;
	}

	ar2c_config_load(&ar2c_config, f);
	fclose(f);

	reload_cards();
}

static int pommax3_get_offset(unsigned int ch, double *offset)
{
	const char *value;
	char *end;

	value = gtk_entry_get_text(GTK_ENTRY(pommax3_offset_w[ch]));
	if (value == NULL)
		return -ENOENT;

	*offset = strtod(value, &end);
	if (*end != '\0')
		return -EINVAL;
	return 0;
}

static int pommax3_get_range(unsigned int ch, double *range)
{
	const char *value;
	char *end;

	value = gtk_entry_get_text(GTK_ENTRY(pommax3_range_w[ch]));
	if (value == NULL)
		return -ENOENT;

	*range = strtod(value, &end);
	if (*end != '\0')
		return -EINVAL;
	return 0;
}

static int pommax3_get_unitvalue(unsigned int ch, double *unit)
{
	const char *value;
	char *end;

	value = gtk_entry_get_text(GTK_ENTRY(pommax3_unit_w[ch]));
	if (value == NULL)
		return -ENOENT;

	*unit = strtod(value, &end);
	if (*end != '\0')
		return -EINVAL;
	return 0;
}

static int pommax3_update_gain(unsigned int i)
{
	GtkTreeModel *model;
	GtkTreeIter iter;
	double range;
	double offset;
	double unitvalue;
	double gain;
	char ystr[32];
	int adc;
	int ret;
	int ar2c_offset;

	if (!gtk_combo_box_get_active_iter(GTK_COMBO_BOX(pommax3_adc_combo),
				&iter))
		return -ENOENT;
	model = gtk_combo_box_get_model(GTK_COMBO_BOX(pommax3_adc_combo));
	gtk_tree_model_get(GTK_TREE_MODEL(model), &iter, 1, &adc, -1);

	ret = pommax3_get_offset(i, &offset);
	if (ret)
		return ret;

	ret = pommax3_get_range(i, &range);
	if (ret)
		return ret;

	ret = pommax3_get_unitvalue(i, &unitvalue);
	if (ret)
		return ret;

	if (offset < -32760 || offset > 32760) {
		gain = 0;
	} else if (unitvalue == 0) {
		gain = 0;
	} else {
		gain = range * 210 * 128 / (21 * unitvalue * (32768 - offset));
		if (gain < -8)
			gain = -0;
		else if (gain > 8)
			gain = 0;
	}
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(pommax3_gain_w[i]),
			gain);

	ar2c_offset = round(offset / 16 * gain);
	snprintf(ystr, 32, "%d\n", ar2c_offset);
	gtk_label_set_text(GTK_LABEL(pommax3_ar2co_w[i]), ystr);
	return 0;
}

static void pommax3_update_data(void)
{
	unsigned int i;
	char *name = NULL;
	const char *value;
	GtkTreeModel *model;
	GtkTreeIter iter;
	int adc;
	int ret;

	if (!gtk_combo_box_get_active_iter(GTK_COMBO_BOX(pommax3_adc_combo),
				&iter))
		return;
	model = gtk_combo_box_get_model(GTK_COMBO_BOX(pommax3_adc_combo));
	gtk_tree_model_get(GTK_TREE_MODEL(model), &iter, 1, &adc, -1);

	for (i = 0; i < ADC_CHANNELS; i++) {
		char ystr[32];
		double x = 0;
		double range;
		double offset;
		double init_gain = 1;
		char *end;
		double y;
		int have_x = 0;

		remake_string(&name, "%s adc%d value%d", pommax3_card_id,
				adc, i);
		value = __ar2c_config_get(&ar2c_data, name);
		fprintf(stderr, "%s, value%d: %s\n", name, i, value);
		if (value == NULL) {
			gtk_label_set_text(GTK_LABEL(pommax3_data_w[i]),
					"-");
			gtk_label_set_text(GTK_LABEL(pommax3_adata_w[i]),
					"-");
		} else {
			gtk_label_set_text(GTK_LABEL(pommax3_data_w[i]),
					value);

			have_x = 1;
			x = strtod(value, &end);
			if (*end != '\0')
				continue;

			value = gtk_label_get_text(GTK_LABEL(
						pommax3_init_gain_w[i]));
			init_gain = strtod(value, &end);
			if (*end != '\0')
				continue;

			x /= init_gain;
			x -= pommax3_init_ofix[i];
			snprintf(ystr, 32, "%d", (int) x);
			gtk_label_set_text(GTK_LABEL(pommax3_adata_w[i]),
					ystr);
		}

		ret = pommax3_get_offset(i, &offset);
		if (ret)
			continue;

		ret = pommax3_get_range(i, &range);
		if (ret)
			continue;

		if (have_x) {
			y = range * (x - offset) / (32768 - offset);
			snprintf(ystr, 32, "%.3f", y);
			gtk_entry_set_text(GTK_ENTRY(pommax3_value_w[i]),
					ystr);
		}
	}
}

void load_data(const char *filename)
{
	FILE *f;

	f = fopen(filename, "r");
	if (f == NULL) {
		fprintf(stderr, "error\n");
		return;
	}

	ar2c_config_clear(&ar2c_data);
	ar2c_config_load(&ar2c_data, f);
	fclose(f);

	pommax3_update_data();
}

void save_file(const char *filename)
{
	FILE *f;
	int ret;

	f = fopen(filename, "w");
	if (f == NULL) {
		fprintf(stderr, "error\n");
		return;
	}

	ar2c_config_store(&ar2c_config, f);
	ret = fclose(f);
	if (ret)
		perror("fclose");
}

void load_config_button(GtkWidget *menuitem, gpointer user_data)
{
	GtkWidget *dialog;

	dialog = gtk_file_chooser_dialog_new(_("Open parameters"),
			GTK_WINDOW(main_wnd),
			GTK_FILE_CHOOSER_ACTION_OPEN,
			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
			NULL);

	if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
		char *filename;

		filename = gtk_file_chooser_get_filename(
				GTK_FILE_CHOOSER(dialog));

		ar2c_config_clear(&ar2c_config);
		open_file(filename);

		free(config_fname);
		config_fname = strdup(filename);

		g_free(filename);
	}
	gtk_widget_destroy(dialog);
}

void save_as_config_button(GtkWidget *menuitem, gpointer user_data)
{
	GtkWidget *dialog;

	dialog = gtk_file_chooser_dialog_new(_("Save parameters"),
			GTK_WINDOW(main_wnd),
			GTK_FILE_CHOOSER_ACTION_SAVE,
			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
			NULL);

	if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
		char *filename;

		filename = gtk_file_chooser_get_filename(
				GTK_FILE_CHOOSER(dialog));
		save_file(filename);

		free(config_fname);
		config_fname = strdup(filename);

		g_free(filename);
	}
	gtk_widget_destroy(dialog);
}

void save_config_button(GtkWidget *menuitem, gpointer user_data)
{
	if (config_fname)
		save_file(config_fname);
	else
		save_as_config_button(menuitem, user_data);
}

#if 0
static char *make_string(const char *fmt, ...)
{
	size_t size;
	char *value = NULL;
	char *p;
	va_list ap;
	int n;

	size = 256;
	value = malloc(size);
	if (value == NULL)
		return NULL;

	while (1) {
		va_start(ap, fmt);
		n = vsnprintf(value, size, fmt, ap);
		va_end(ap);

		if (n > -1 && n < size)
			break;

		if (n > -1)
			size = n + 1;
		else
			size *= 2;

		p = realloc(value, size);
		if (p == NULL) {
			free(value);
			return NULL;
		}
		value = p;
	}
	return value;
}
#endif

static void card_changed(GtkWidget *widget, gpointer user_data)
{
	GtkTreeIter iter;
	gchar *card_id = NULL;
	char *name = NULL;
	const char *value;
	char *end;
	unsigned int i;
	char buf[128];
	u16 vendor_id = 0xffff;
	u16 device_id = 0xffff;
	u16 subvendor_id = 0xffff;
	u16 subsystem_id = 0xffff;

	if (!gtk_combo_box_get_active_iter(GTK_COMBO_BOX(cards), &iter))
		return;
	gtk_tree_model_get(GTK_TREE_MODEL(cards_store), &iter, 1, &card_id,
			-1);

	fprintf(stderr, "card_id: %s\n", card_id);

	remake_string(&name, "%s VendorID", card_id);
	value = __ar2c_config_get(&ar2c_config, name);
	if (value != NULL) {
		vendor_id = strtol(value, &end, 0);
		if (*end != '\0')
			vendor_id = 0xffff;

		remake_string(&name, "%s (0x%04x)",
				pci_lookup_name(pacc, buf, sizeof(buf),
					PCI_LOOKUP_VENDOR,
					vendor_id), vendor_id);
		gtk_label_set_text(GTK_LABEL(vendor_id_w), name);
	} else {
		gtk_label_set_text(GTK_LABEL(vendor_id_w), "undefined");
	}

	remake_string(&name, "%s DeviceID", card_id);
	value = __ar2c_config_get(&ar2c_config, name);
	if (value != NULL) {
		device_id = strtol(value, &end, 0);
		if (*end != '\0')
			device_id = 0xffff;

		remake_string(&name, "%s (0x%04x)",
				pci_lookup_name(pacc, buf, sizeof(buf),
					PCI_LOOKUP_VENDOR | PCI_LOOKUP_DEVICE,
					vendor_id, device_id), device_id);
		gtk_label_set_text(GTK_LABEL(device_id_w), name);
	} else {
		gtk_label_set_text(GTK_LABEL(device_id_w), "undefined");
	}

	remake_string(&name, "%s SubsystemVendorID", card_id);
	value = __ar2c_config_get(&ar2c_config, name);
	if (value != NULL) {
		subvendor_id = strtol(value, &end, 0);
		if (*end != '\0')
			subvendor_id = 0xffff;

		remake_string(&name, "%s (0x%04x)",
				pci_lookup_name(pacc, buf, sizeof(buf),
					PCI_LOOKUP_SUBSYSTEM
					| PCI_LOOKUP_VENDOR,
					subvendor_id), subvendor_id);
		gtk_label_set_text(GTK_LABEL(subsystem_vendor_id_w), name);
	} else {
		gtk_label_set_text(GTK_LABEL(subsystem_vendor_id_w),
				"undefined");
	}

	remake_string(&name, "%s SubsystemID", card_id);
	value = __ar2c_config_get(&ar2c_config, name);
	if (value != NULL) {
		subsystem_id = strtol(value, &end, 0);
		if (*end != '\0')
			subsystem_id = 0xffff;

		remake_string(&name, "%s (0x%04x)",
				pci_lookup_name(pacc, buf, sizeof(buf),
				PCI_LOOKUP_SUBSYSTEM | PCI_LOOKUP_VENDOR
				| PCI_LOOKUP_DEVICE,
				vendor_id, device_id, subvendor_id,
				subsystem_id), subsystem_id);

		gtk_label_set_text(GTK_LABEL(subsystem_id_w), name);
	} else {
		gtk_label_set_text(GTK_LABEL(subsystem_id_w), "undefined");
	}

	for (i = 0; i < 6; i++) {
		remake_string(&name, "%s BAR%d", card_id, i);
		value = __ar2c_config_get(&ar2c_config, name);
		if (value != NULL) {
			gtk_label_set_text(GTK_LABEL(pci_bar_w[i]), value);
		} else {
			gtk_label_set_text(GTK_LABEL(pci_bar_w[i]),
					"undefined");
		}
	}

	g_free(card_id);
}

static void pommax3_update_adc(void)
{
	unsigned int version_major, version_minor;
	unsigned int i;
	char *name = NULL;
	const char *value;
	GtkTreeModel *model;
	GtkTreeIter iter;
	int adc;
	unsigned int flags = 0;
	int ret;

	if (!gtk_combo_box_get_active_iter(GTK_COMBO_BOX(pommax3_adc_combo),
				&iter))
		return;
	model = gtk_combo_box_get_model(GTK_COMBO_BOX(pommax3_adc_combo));
	gtk_tree_model_get(GTK_TREE_MODEL(model), &iter, 1, &adc, -1);

	remake_string(&name, "%s adc%d freq", pommax3_card_id, adc);
	value = __ar2c_config_get(&ar2c_config, name);
	if (value != NULL) {
		int freq = atoi(value);

		if (!freq)
			freq = 1280;
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(pommax3_freq_w),
				(gdouble) freq);
	} else {
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(pommax3_freq_w),
				(gdouble) 1280);
	}

	remake_string(&name, "%s adc%d version", pommax3_card_id, adc);
	value = __ar2c_config_get(&ar2c_config, name);
	if (value) {
		ret = sscanf(value, "%d.%d", &version_major, &version_minor);
		if (ret < 2) {
			version_major = 0;
			version_minor = 0;
		}
	}

	remake_string(&name, "%s adc%d offset", pommax3_card_id, adc);
	value = __ar2c_config_get(&ar2c_config, name);
	if (value != NULL) {
		int offset = atoi(value);

		if (!offset)
			offset = 16384;
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(pommax3_adc_offset_w),
				(gdouble) offset);
	} else {
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(pommax3_adc_offset_w),
				(gdouble) 16384);
	}

	remake_string(&name, "%s adc%d serial", pommax3_card_id, adc);
	value = __ar2c_config_get(&ar2c_config, name);
	gtk_label_set_text(GTK_LABEL(pommax3_serial_w),	value ? value : "?");

	remake_string(&name, "%s adc%d flags", pommax3_card_id, adc);
	value = __ar2c_config_get(&ar2c_config, name);
	if (value != NULL)
		flags = atoi(value);

	if (version_major == 0 && version_minor < 6) {
		flags &= ~POMMAX3_FLAGS_OFFSET;
		gtk_toggle_button_set_inconsistent(
				GTK_TOGGLE_BUTTON(pommax3_offset_enable),
					TRUE);
	} else {
		gtk_toggle_button_set_inconsistent(
				GTK_TOGGLE_BUTTON(pommax3_offset_enable),
					FALSE);
	}

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pommax3_gain_enable),
				(flags & POMMAX3_FLAGS_GAIN) ? TRUE : FALSE);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pommax3_offset_enable),
				(flags & POMMAX3_FLAGS_OFFSET) ? TRUE : FALSE);

	for (i = 0; i < ADC_CHANNELS; i++) {
		gdouble v;
		int ofix;
		char vstr[32];

		remake_string(&name, "%s adc%d gain%d", pommax3_card_id,
				adc, i);
		value = __ar2c_config_get(&ar2c_config, name);
		fprintf(stderr, "gain%d: %s\n", i, value);

		v = 1.0;
		if (value != NULL && (flags & POMMAX3_FLAGS_GAIN))
			v = atoi(value) / 4096.0;
		snprintf(vstr, 32, "%.5f", v);
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(pommax3_gain_w[i]),
				v);
		gtk_label_set_text(GTK_LABEL(pommax3_init_gain_w[i]), vstr);

		remake_string(&name, "%s adc%d ofix%d", pommax3_card_id,
				adc, i);
		ofix = 0;
		value = __ar2c_config_get(&ar2c_config, name);
		if (value != NULL && (flags & POMMAX3_FLAGS_GAIN))
			ofix = atoi(value);
		pommax3_init_ofix[i] = ofix;

		remake_string(&name, "%s adc%d offset%d", pommax3_card_id,
				adc, i);
		value = __ar2c_config_get(&ar2c_config, name);
		value = (value != NULL) ? value : "0";
		gtk_entry_set_text(GTK_ENTRY(pommax3_offset_w[i]), value);

		remake_string(&name, "%s adc%d range%d", pommax3_card_id,
				adc, i);
		value = __ar2c_config_get(&ar2c_config, name);
		value = (value != NULL) ? value : "20";
		gtk_entry_set_text(GTK_ENTRY(pommax3_range_w[i]), value);

		remake_string(&name, "%s adc%d unit%d", pommax3_card_id,
				adc, i);
		value = __ar2c_config_get(&ar2c_config, name);
		value = (value != NULL) ? value : "1";
		gtk_entry_set_text(GTK_ENTRY(pommax3_unit_w[i]), value);
	}

	pommax3_update_data();
}

static void pommax3_commit(GtkWidget *widget, gpointer user_data)
{
	unsigned int i;
	char *name = NULL;
	GtkTreeModel *model;
	GtkTreeIter iter;
	const char *s;
	int adc;
	double v;
	int x;
	unsigned int flags;

	if (!gtk_combo_box_get_active_iter(GTK_COMBO_BOX(pommax3_adc_combo),
				&iter))
		return;
	model = gtk_combo_box_get_model(GTK_COMBO_BOX(pommax3_adc_combo));
	gtk_tree_model_get(GTK_TREE_MODEL(model), &iter, 1, &adc, -1);

	v = gtk_spin_button_get_value(GTK_SPIN_BUTTON(pommax3_freq_w));
	remake_string(&name, "%s adc%d freq", pommax3_card_id, adc);
	ar2c_config_set_all_fmt(&ar2c_config, name, "gconf", 1, "%d",
			(int) v);

	v = gtk_spin_button_get_value(GTK_SPIN_BUTTON(pommax3_adc_offset_w));
	remake_string(&name, "%s adc%d offset", pommax3_card_id, adc);
	ar2c_config_set_all_fmt(&ar2c_config, name, "gconf", 1, "%d",
			(int) v);

	flags = 0;
	if (gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(pommax3_gain_enable)))
		flags |= POMMAX3_FLAGS_GAIN;
	if (gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(pommax3_offset_enable)))
		flags |= POMMAX3_FLAGS_OFFSET;
	remake_string(&name, "%s adc%d flags", pommax3_card_id, adc);
	ar2c_config_set_all_fmt(&ar2c_config, name, "gconf", 1, "%d",
			(int) flags);

	for (i = 0; i < ADC_CHANNELS; i++) {
		v = gtk_spin_button_get_value(GTK_SPIN_BUTTON(
					pommax3_gain_w[i]));
		remake_string(&name, "%s adc%d gain%d",
				pommax3_card_id, adc, i);

		x = (int) round(4096 * v);
		if (x > 32767)
			x = 32767;
		else if (x < -32768)
			x = -32768;

		ar2c_config_set_all_fmt(&ar2c_config, name, "gconf", 1, "%d",
				x);

		s = gtk_entry_get_text(GTK_ENTRY(pommax3_offset_w[i]));
		remake_string(&name, "%s adc%d offset%d", pommax3_card_id,
				adc, i);
		ar2c_config_set_all(&ar2c_config, name, s, "gconf", 1);

		x = -atoi(s);
		if (x > 32767)
			x = 32767;
		else if (x < -32768)
			x = -32768;
		if (!(flags & POMMAX3_FLAGS_OFFSET))
			x = 0;
		remake_string(&name, "%s adc%d ofix%d", pommax3_card_id,
				adc, i);
		ar2c_config_set_all_fmt(&ar2c_config, name, "gconf", 1, "%d",
				x & ~1);

		s = gtk_entry_get_text(GTK_ENTRY(pommax3_range_w[i]));
		remake_string(&name, "%s adc%d range%d", pommax3_card_id,
				adc, i);
		ar2c_config_set_all(&ar2c_config, name, s, "gconf", 1);

		s = gtk_entry_get_text(GTK_ENTRY(pommax3_unit_w[i]));
		remake_string(&name, "%s adc%d unit%d", pommax3_card_id,
				adc, i);
		ar2c_config_set_all(&ar2c_config, name, s, "gconf", 1);
	}
}

static void pommax3_adc_changed(GtkWidget *widget, gpointer user_data)
{
	pommax3_update_adc();
}

static void pommax3_range_offset_changed(GtkWidget *widget,
		gpointer user_data)
{
	unsigned int i = (unsigned long) user_data;

	pommax3_update_gain(i);
}

static gboolean pommax3_value_key_press(GtkWidget *widget, GdkEvent *event,
		gpointer user_data)
{
	GdkEventKey *k = (GdkEventKey *) event;
	unsigned int i;
	const char *value;
	GtkTreeModel *model;
	GtkTreeIter iter;
	int adc;

	if (k->keyval != GDK_Return)
		return FALSE;

	if (!gtk_combo_box_get_active_iter(GTK_COMBO_BOX(pommax3_adc_combo),
				&iter))
		return FALSE;
	model = gtk_combo_box_get_model(GTK_COMBO_BOX(pommax3_adc_combo));
	gtk_tree_model_get(GTK_TREE_MODEL(model), &iter, 1, &adc, -1);

	for (i = 0; i < ADC_CHANNELS; i++) {
		char ystr[32];
		double x = 0;
		double y;
		double range;
		double offset;
		char *end;

		if (widget != pommax3_value_w[i])
			continue;

		value = gtk_label_get_text(GTK_LABEL(pommax3_adata_w[i]));
		if (value == NULL)
			continue;
		x = strtod(value, &end);
		if (*end != '\0')
			continue;

		value = gtk_entry_get_text(GTK_ENTRY(pommax3_value_w[i]));
		if (value == NULL)
			continue;
		y = strtod(value, &end);
		if (*end != '\0')
			continue;

		value = gtk_entry_get_text(GTK_ENTRY(pommax3_offset_w[i]));
		if (value == NULL)
			continue;
		offset = strtod(value, &end);
		if (*end != '\0')
			continue;

		if (fabs(x - offset) < 1000) {
			fprintf(stderr, "too low value for range estimation\n");
			range = 9999.9;
		} else {
			range = y * (32768 - offset) / (x - offset);
		}
		snprintf(ystr, 32, "%.3f", range);
		gtk_entry_set_text(GTK_ENTRY(pommax3_range_w[i]), ystr);
	}
	pommax3_update_data();
	return FALSE;
}

static void pommax3_configure(const char *card_id)
{
	gtk_combo_box_set_active(GTK_COMBO_BOX(pommax3_adc_combo), 0);

	free(pommax3_card_id);
	pommax3_card_id = strdup(card_id);
	gtk_widget_show(pommax3_wnd);

	gtk_label_set_text(GTK_LABEL(pommax3_card_id_w), pommax3_card_id);

	pommax3_update_adc();
}

static void configure_card(GtkWidget *widget, gpointer user_data)
{
	GtkTreeIter iter;
	gchar *card_id = NULL;
	char *name = NULL;
	const char *vendor_id;
	const char *device_id;

	if (!gtk_combo_box_get_active_iter(GTK_COMBO_BOX(cards), &iter))
		return;
	gtk_tree_model_get(GTK_TREE_MODEL(cards_store), &iter, 1, &card_id,
			-1);

	remake_string(&name, "%s VendorID", card_id);
	vendor_id = __ar2c_config_get(&ar2c_config, name);
	remake_string(&name, "%s DeviceID", card_id);
	device_id = __ar2c_config_get(&ar2c_config, name);

	if (vendor_id == NULL || device_id == NULL)
		return;
	if (!strcmp(vendor_id, "0xff00") && !strcmp(device_id, "0x0003"))
		pommax3_configure(card_id);
}

static void pommax3_hide(GtkWidget *widget, gpointer user_data)
{
	gtk_widget_hide(pommax3_wnd);
}

enum xfer_op {
	XFER_IDLE,
	XFER_DOWNLOAD,
	XFER_UPLOAD,
	XFER_FLASH,
	XFER_GET_DATA,
};
static enum xfer_op xfer_op = XFER_IDLE;
static GPid xfer_pid;
static gint xfer_stderr;
static gint xfer_itag;

static void xfer_log_printf(const char *fmt, ...)
{
	char buf[256];
	va_list ap;

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);

	gtk_text_buffer_insert_at_cursor(xfer_log, buf, -1);
}

static void cleanup_cards(struct ar2c_config *c)
{
	char name[128];
	unsigned int i;

	for (i = 0; i < 256; i++) {
		snprintf(name, sizeof(name), "card %d", i);
		ar2c_config_rm(c, name);
	}
}

static void xfer_watch_child(GPid pid, gint status, gpointer user_data)
{
	char *tmp = user_data;

	gtk_spinner_stop(GTK_SPINNER(xfer_spinner_w));

	g_spawn_close_pid(pid);

	if (WIFEXITED(status)) {
		if (WEXITSTATUS(status))
			xfer_log_printf(_("child exited with error %d\n"),
					WEXITSTATUS(status));
		else
			xfer_log_printf(_("Operation completed successfully\n"));
	} else if (WIFSIGNALED(status)) {
		xfer_log_printf(_("child killed with signal %d\n"),
				WTERMSIG(status));
	}
	gtk_widget_show(xfer_ok);

	if (tmp != NULL) {
		if (xfer_op == XFER_DOWNLOAD) {
			cleanup_cards(&ar2c_config);
			open_file(tmp);
		} else if (xfer_op == XFER_GET_DATA) {
			load_data(tmp);
		}
		unlink(tmp);
		free(tmp);
	}

	xfer_op = XFER_IDLE;
}

static void xfer_read(gpointer data, gint source, GdkInputCondition condition)
{
	ssize_t ret, i;
	char buff[1024];

	if (condition == GDK_INPUT_EXCEPTION) {
		gdk_input_remove(xfer_itag);
		close(xfer_stderr);
		xfer_stderr = -1;
		return;
	}

	ret = read(xfer_stderr, buff, sizeof(buff) - 1);
	if (ret < 0)
		return;
	if (!ret) {
		gdk_input_remove(xfer_itag);
		close(xfer_stderr);
		xfer_stderr = -1;
		return;
	}

	for (i = 0; i < ret; i++)
		if (buff[i] == '\0')
			buff[i] = '\n';
	buff[ret] = '\0';

	fputs(buff, stderr);
	gtk_text_buffer_insert_at_cursor(xfer_log, buff, ret);
}


static int setup_conn(char **argv, unsigned int max)
{
	int ret;

	if (max < 6) {
		fprintf(stderr, "invalid max argument\n");
		return -EINVAL;
	}

	switch (conn_type) {
	case CONN_TCP:
		if (!tcp_host || !tcp_port) {
			fprintf(stderr, "invalid host/port: %s/%s\n",
					tcp_host, tcp_port);
			return -EINVAL;
		}

		argv[0] = strdup("-s");
		argv[1] = strdup(tcp_host);
		argv[2] = strdup("-p");
		argv[3] = strdup(tcp_port);
		ret = 4;
		break;
	case CONN_FILE:
		if (!file_path) {
			fprintf(stderr, "invalid path: %s\n", file_path);
			return -EINVAL;
		}

		argv[0] = strdup("-d");
		argv[1] = strdup(file_path);
		ret = 2;
		if (file_baudrate) {
			argv[2] = strdup("--bitrate");
			argv[3] = strdup(file_baudrate);
			argv[4] = strdup("--bbitrate");
			argv[5] = strdup(file_baudrate);
			ret = 6;
		}
		break;
	default:
		return -EINVAL;
	}
	return ret;
}


static void setup_conn_error(void)
{
	gconf_error(_("Connection error"),
			_("Connection configuration error.\n"
			"Please check connection configuration."));
}

static void download_all(GtkWidget *widget, gpointer user_data)
{
	GError *error = NULL;
	char *argv[11];
	char **t;
	char tmp[PATH_MAX];
	int ret;

	if (xfer_op != XFER_IDLE)
		return;

	ret = git_mkstemp(tmp, PATH_MAX, ".gconf.XXXXXX");
	if (ret == -1) {
		perror("mkstemp");
		return;
	}
	close(ret);

	fprintf(stderr, "tmp %s\n", tmp);

	t = argv;
	*t++ = strdup(UTILSDIR "ar2c-arbusctl");
	ret = setup_conn(t, 6);
	if (ret < 0) {
		setup_conn_error();
		unlink(tmp);
		*t = NULL;
		goto out_free;
	}
	t += ret;
	*t++ = strdup("-c");
	*t++ = strdup(tmp);
	*t++ = strdup("get-config");
	*t++ = NULL;

	gtk_text_buffer_set_text(xfer_log, "", -1);
	gtk_widget_hide(xfer_ok);
	gtk_spinner_start(GTK_SPINNER(xfer_spinner_w));
	gtk_widget_show(xfer_w);

	if (!g_spawn_async_with_pipes(NULL, argv, NULL,
				G_SPAWN_DO_NOT_REAP_CHILD,
			NULL, NULL, &xfer_pid, NULL, NULL, &xfer_stderr,
			&error)) {
		gconf_error(_("Download error"),
				"g_spawn_async_with_pipes error: %s\n",
				error->message);
		g_error_free(error);
		gtk_spinner_stop(GTK_SPINNER(xfer_spinner_w));
		gtk_widget_hide(xfer_w);
		unlink(tmp);
		*t = NULL;
		goto out_free;
	}
	xfer_itag = gdk_input_add(xfer_stderr, GDK_INPUT_READ, xfer_read, NULL);
	g_child_watch_add(xfer_pid, xfer_watch_child, strdup(tmp));

	xfer_op = XFER_DOWNLOAD;

out_free:
	for (t = argv; *t != NULL; t++)
		free(*t);
}

static void upload_all(GtkWidget *widget, gpointer user_data)
{
	GError *error = NULL;
	char *argv[12];
	char **t;
	char tmp[PATH_MAX];
	int ret;

	if (xfer_op != XFER_IDLE)
		return;

	ret = git_mkstemp(tmp, PATH_MAX, ".gconf.XXXXXX");
	if (ret == -1) {
		perror("mkstemp");
		return;
	}
	close(ret);
	save_file(tmp);

	t = argv;
	*t++ = strdup(UTILSDIR "ar2c-arbusctl");
	ret = setup_conn(t, 6);
	if (ret < 0) {
		setup_conn_error();
		unlink(tmp);
		*t = NULL;
		goto out_free;
	}
	t += ret;
	*t++ = strdup("-c");
	*t++ = strdup(tmp);
	*t++ = strdup("put-config");
	*t++ = NULL;

	gtk_text_buffer_set_text(xfer_log, "", -1);
	gtk_widget_hide(xfer_ok);
	gtk_spinner_start(GTK_SPINNER(xfer_spinner_w));
	gtk_widget_show(xfer_w);

	if (!g_spawn_async_with_pipes(NULL, argv, NULL,
				G_SPAWN_DO_NOT_REAP_CHILD,
			NULL, NULL, &xfer_pid, NULL, NULL, &xfer_stderr,
			&error)) {
		gconf_error(_("Upload error"),
				"g_spawn_async_with_pipes error: %s\n",
				error->message);
		g_error_free(error);
		gtk_spinner_stop(GTK_SPINNER(xfer_spinner_w));
		gtk_widget_hide(xfer_w);
		unlink(tmp);
		*t = NULL;
		goto out_free;
	}
	xfer_itag = gdk_input_add(xfer_stderr, GDK_INPUT_READ, xfer_read, NULL);
	g_child_watch_add(xfer_pid, xfer_watch_child, strdup(tmp));

	xfer_op = XFER_UPLOAD;
out_free:
	for (t = argv; *t != NULL; t++)
		free(*t);
}

static void flash_all(GtkWidget *widget, gpointer user_data)
{
	GError *error = NULL;
	char *argv[10];
	char **t;
	int ret;

	if (xfer_op != XFER_IDLE)
		return;

	t = argv;
	*t++ = strdup(UTILSDIR "ar2c-arbusctl");
	ret = setup_conn(t, 6);
	if (ret < 0) {
		setup_conn_error();
		*t = NULL;
		goto out_free;
	}
	t += ret;
	*t++ = strdup("save");
	*t++ = NULL;

	gtk_text_buffer_set_text(xfer_log, "", -1);
	gtk_widget_hide(xfer_ok);
	gtk_spinner_start(GTK_SPINNER(xfer_spinner_w));
	gtk_widget_show(xfer_w);

	if (!g_spawn_async_with_pipes(NULL, argv, NULL,
				G_SPAWN_DO_NOT_REAP_CHILD,
			NULL, NULL, &xfer_pid, NULL, NULL, &xfer_stderr,
			&error)) {
		gconf_error(_("Flash error"),
				"g_spawn_async_with_pipes error: %s\n",
				error->message);
		g_error_free(error);
		gtk_spinner_stop(GTK_SPINNER(xfer_spinner_w));
		gtk_widget_hide(xfer_w);
		*t = NULL;
		goto out_free;
	}
	xfer_itag = gdk_input_add(xfer_stderr, GDK_INPUT_READ, xfer_read, NULL);
	g_child_watch_add(xfer_pid, xfer_watch_child, NULL);

	xfer_op = XFER_FLASH;

out_free:
	for (t = argv; *t != NULL; t++)
		free(*t);
}

static void get_data_all(GtkWidget *widget, gpointer user_data)
{
	GError *error = NULL;
	char *argv[11];
	char **t;
	char tmp[PATH_MAX];
	int ret;

	if (xfer_op != XFER_IDLE)
		return;

	ret = git_mkstemp(tmp, PATH_MAX, ".gconf.XXXXXX");
	if (ret == -1) {
		perror("mkstemp");
		return;
	}
	close(ret);

	t = argv;
	*t++ = strdup(UTILSDIR "ar2c-arbusctl");
	ret = setup_conn(t, 6);
	if (ret < 0) {
		setup_conn_error();
		unlink(tmp);
		*t = NULL;
		goto out_free;
	}
	t += ret;
	*t++ = strdup("-c");
	*t++ = strdup(tmp);
	*t++ = strdup("get-data");
	*t++ = NULL;

	gtk_text_buffer_set_text(xfer_log, "", -1);
	gtk_widget_hide(xfer_ok);
	gtk_spinner_start(GTK_SPINNER(xfer_spinner_w));
	gtk_widget_show(xfer_w);

	if (!g_spawn_async_with_pipes(NULL, argv, NULL,
				G_SPAWN_DO_NOT_REAP_CHILD,
			NULL, NULL, &xfer_pid, NULL, NULL, &xfer_stderr,
			&error)) {
		gconf_error(_("Download data error"),
				"g_spawn_async_with_pipes error: %s\n",
				error->message);
		g_error_free(error);
		gtk_spinner_stop(GTK_SPINNER(xfer_spinner_w));
		gtk_widget_hide(xfer_w);
		unlink(tmp);
		*t = NULL;
		goto out_free;
	}
	xfer_itag = gdk_input_add(xfer_stderr, GDK_INPUT_READ, xfer_read, NULL);
	g_child_watch_add(xfer_pid, xfer_watch_child, strdup(tmp));

	xfer_op = XFER_GET_DATA;

out_free:
	for (t = argv; *t != NULL; t++)
		free(*t);
}

static void xfer_ok_clicked(GtkWidget *widget, gpointer user_data)
{
	gtk_widget_hide(xfer_w);
}

static void xfer_abort(GtkWidget *widget, gpointer user_data)
{
	if (xfer_op != XFER_IDLE)
		kill(xfer_pid, SIGTERM);
	else
		xfer_ok_clicked(widget, user_data);
}

static void show_preferences(GtkWidget *widget, gpointer user_data)
{
	if (tcp_host)
		gtk_entry_set_text(GTK_ENTRY(conn_tcp_host_w), tcp_host);
	if (tcp_port)
		gtk_entry_set_text(GTK_ENTRY(conn_tcp_port_w), tcp_port);
	if (file_path)
		gtk_entry_set_text(GTK_ENTRY(conn_file_path_w), file_path);
	if (file_baudrate)
		gtk_entry_set_text(GTK_ENTRY(conn_file_baudrate_w),
				file_baudrate);

	if (conn_type == CONN_TCP) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(conn_tcp_w),
				TRUE);
	} else if (conn_type == CONN_FILE) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(conn_file_w),
				TRUE);
	}
	gtk_widget_show(preferences_wnd);
}

static void update_preferences(GtkWidget *widget, gpointer user_data)
{
	free(tcp_host);
	tcp_host = strdup(gtk_entry_get_text(GTK_ENTRY(conn_tcp_host_w)));
	free(tcp_port);
	tcp_port = strdup(gtk_entry_get_text(GTK_ENTRY(conn_tcp_port_w)));
	free(file_path);
	file_path = strdup(gtk_entry_get_text(GTK_ENTRY(conn_file_path_w)));
	free(file_baudrate);
	file_baudrate = strdup(gtk_entry_get_text(GTK_ENTRY(
					conn_file_baudrate_w)));

	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(conn_tcp_w))) {
		conn_type = CONN_TCP;
	} else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(conn_file_w))) {
		conn_type = CONN_FILE;
	} else {
		fprintf(stderr, "error: unknown conn method\n");
	}
	gtk_widget_hide(preferences_wnd);
}

static void save_preferences(GtkWidget *widget, gpointer user_data)
{
	const char *type = "unknown";

	update_preferences(widget, user_data);

	g_key_file_set_value(gconf_config, "comm", "tcp_host", tcp_host);
	g_key_file_set_value(gconf_config, "comm", "tcp_port", tcp_port);
	g_key_file_set_value(gconf_config, "comm", "file_path", file_path);
	g_key_file_set_value(gconf_config, "comm", "file_baudrate",
			file_baudrate);
	switch (conn_type) {
	case CONN_TCP:
		type = "tcp";
		break;
	case CONN_FILE:
		type = "file";
		break;
	default:
		break;
	}
	g_key_file_set_value(gconf_config, "comm", "type", type);

	gconf_config_save();
}

static void hide_preferences(GtkWidget *widget, gpointer user_data)
{
	gtk_widget_hide(preferences_wnd);
}

static const char *gconf_authors[] = {
	"Krzysztof Mazur <krzysiek@podlesie.net>",
	NULL };

static const char gconf_copyright[] =
	"Copyright (C) 2012..2015  Krzysztof Mazur";

static const char gconf_comments[] =
	N_("ar2c-gconf is a graphical interface for ar2c-par-utils,\n"
	"a set of utilities for configuration of AR-2c dataloggers.");

static const char gconf_license[] =
	N_("This program is free software; you can redistribute it and/or\n"
	"modify it under the terms of the GNU General Public License\n"
	"as published by the Free Software Foundation; either version 2\n"
	"of the License, or (at your option) any later version.\n"
	"\n"
	"This program is distributed in the hope that it will be useful,\n"
	"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
	"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
	"GNU General Public License for more details.\n");

static const char gconf_version[] = PACKAGE_VERSION;

static void show_about(GtkWidget *widget, gpointer user_data)
{
	gtk_show_about_dialog(NULL,
			"program-name", "ar2c-gconf",
			"title", _("ar2c-gconf"),
			"authors", gconf_authors,
			"copyright", gconf_copyright,
			"comments", gconf_comments,
			"license", gconf_license,
			"version", gconf_version,
			NULL);
}

void init_main_window(const gchar *glade_file)
{
	char name[256];
	GError *error = NULL;
	GtkBuilder *builder = gtk_builder_new();
	GtkWidget *w;
	GtkCellRenderer *cell;
	unsigned int i;

	if (!gtk_builder_add_from_file(builder, glade_file, &error)) {
		gconf_error(_("gtk builder error"),
				_("Couldn't load builder file: %s."),
				error->message, glade_file);
		g_error_free(error);
	}

	main_wnd = GTK_WIDGET(gtk_builder_get_object(builder, "window1"));
	g_signal_connect(G_OBJECT(main_wnd), "destroy",
			G_CALLBACK(gtk_main_quit), NULL);

	pommax3_wnd = GTK_WIDGET(gtk_builder_get_object(builder, "pommax3"));
	g_signal_connect(G_OBJECT(pommax3_wnd), "delete-event",
			G_CALLBACK(pommax3_hide), NULL);
	pommax3_card_id_w = GTK_WIDGET(gtk_builder_get_object(builder,
				"pommax3_card_id"));
	pommax3_freq_w = GTK_WIDGET(gtk_builder_get_object(builder,
				"pommax3_freq"));
	pommax3_adc_offset_w = GTK_WIDGET(gtk_builder_get_object(builder,
				"pommax3_offset"));
	pommax3_serial_w = GTK_WIDGET(gtk_builder_get_object(builder,
				"pommax3_serial"));

	for (i = 0; i < ADC_CHANNELS; i++) {
		snprintf(name, sizeof(name), "pommax3_gain%d", i);
		pommax3_gain_w[i] = GTK_WIDGET(gtk_builder_get_object(
					builder, name));

		snprintf(name, sizeof(name), "pommax3_init_gain%d", i);
		pommax3_init_gain_w[i] = GTK_WIDGET(gtk_builder_get_object(
					builder, name));

		snprintf(name, sizeof(name), "pommax3_data%d", i);
		pommax3_data_w[i] = GTK_WIDGET(gtk_builder_get_object(
					builder, name));

		snprintf(name, sizeof(name), "pommax3_adata%d", i);
		pommax3_adata_w[i] = GTK_WIDGET(gtk_builder_get_object(
					builder, name));
		
		snprintf(name, sizeof(name), "pommax3_range%d", i);
		pommax3_range_w[i] = GTK_WIDGET(gtk_builder_get_object(
					builder, name));

		snprintf(name, sizeof(name), "pommax3_unit%d", i);
		pommax3_unit_w[i] = GTK_WIDGET(gtk_builder_get_object(
					builder, name));
		
		snprintf(name, sizeof(name), "pommax3_value%d", i);
		pommax3_value_w[i] = GTK_WIDGET(gtk_builder_get_object(
					builder, name));
		
		snprintf(name, sizeof(name), "pommax3_offset%d", i);
		pommax3_offset_w[i] = GTK_WIDGET(gtk_builder_get_object(
					builder, name));
		
		snprintf(name, sizeof(name), "pommax3_ar2co%d", i);
		pommax3_ar2co_w[i] = GTK_WIDGET(gtk_builder_get_object(
					builder, name));
	}

	pommax3_gain_enable = GTK_WIDGET(gtk_builder_get_object(builder,
				"pommax3_gain_enable"));
	pommax3_offset_enable = GTK_WIDGET(gtk_builder_get_object(builder,
				"pommax3_offset_enable"));
	pommax3_adc_combo = GTK_WIDGET(gtk_builder_get_object(builder,
				"pommax3_adc_combo"));
	g_signal_connect(G_OBJECT(pommax3_adc_combo), "changed",
			G_CALLBACK(pommax3_adc_changed), NULL);

	for (i = 0; i < ADC_CHANNELS; i++) {
		void *p = (void *)(unsigned long)i;
		g_signal_connect(G_OBJECT(pommax3_range_w[i]), "changed",
				G_CALLBACK(pommax3_range_offset_changed), p);
		g_signal_connect(G_OBJECT(pommax3_unit_w[i]), "changed",
				G_CALLBACK(pommax3_range_offset_changed), p);
		g_signal_connect(G_OBJECT(pommax3_offset_w[i]), "changed",
				G_CALLBACK(pommax3_range_offset_changed), p);
		g_signal_connect(G_OBJECT(pommax3_value_w[i]),
				"key-press-event",
				G_CALLBACK(pommax3_value_key_press), NULL);
	}

	w = GTK_WIDGET(gtk_builder_get_object(builder, "pommax3_commit"));
	g_signal_connect(G_OBJECT(w), "clicked",
			G_CALLBACK(pommax3_commit), NULL);

	w = GTK_WIDGET(gtk_builder_get_object(builder, "pommax3_reset"));
	g_signal_connect(G_OBJECT(w), "clicked",
			G_CALLBACK(pommax3_adc_changed), NULL);

	cell = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(pommax3_adc_combo), cell,
			TRUE);
	gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(pommax3_adc_combo),
			cell, "text", 0, NULL);

	vendor_id_w = GTK_WIDGET(gtk_builder_get_object(builder, "vendor_id"));
	device_id_w = GTK_WIDGET(gtk_builder_get_object(builder, "device_id"));
	subsystem_vendor_id_w = GTK_WIDGET(gtk_builder_get_object(builder,
				"subsystem_vendor_id"));
	subsystem_id_w = GTK_WIDGET(gtk_builder_get_object(builder,
				"subsystem_id"));
	pci_bar_w[0] = GTK_WIDGET(gtk_builder_get_object(builder, "pci_bar0"));
	pci_bar_w[1] = GTK_WIDGET(gtk_builder_get_object(builder, "pci_bar1"));
	pci_bar_w[2] = GTK_WIDGET(gtk_builder_get_object(builder, "pci_bar2"));
	pci_bar_w[3] = GTK_WIDGET(gtk_builder_get_object(builder, "pci_bar3"));
	pci_bar_w[4] = GTK_WIDGET(gtk_builder_get_object(builder, "pci_bar4"));
	pci_bar_w[5] = GTK_WIDGET(gtk_builder_get_object(builder, "pci_bar5"));

	cards = GTK_WIDGET(gtk_builder_get_object(builder, "cards"));
	g_signal_connect(G_OBJECT(cards), "changed",
			G_CALLBACK(card_changed), NULL);

	w = GTK_WIDGET(gtk_builder_get_object(builder, "file_open"));
	g_signal_connect(G_OBJECT(w), "activate",
			G_CALLBACK(load_config_button), NULL);
	w = GTK_WIDGET(gtk_builder_get_object(builder, "tool_open"));
	g_signal_connect(G_OBJECT(w), "clicked",
			G_CALLBACK(load_config_button), NULL);

	w = GTK_WIDGET(gtk_builder_get_object(builder, "file_save_as"));
	g_signal_connect(G_OBJECT(w), "activate",
			G_CALLBACK(save_as_config_button), NULL);
	w = GTK_WIDGET(gtk_builder_get_object(builder, "tool_save_as"));
	g_signal_connect(G_OBJECT(w), "clicked",
			G_CALLBACK(save_as_config_button), NULL);

	w = GTK_WIDGET(gtk_builder_get_object(builder, "file_save"));
	g_signal_connect(G_OBJECT(w), "activate",
			G_CALLBACK(save_config_button), NULL);
	w = GTK_WIDGET(gtk_builder_get_object(builder, "tool_save"));
	g_signal_connect(G_OBJECT(w), "clicked",
			G_CALLBACK(save_config_button), NULL);

	w = GTK_WIDGET(gtk_builder_get_object(builder, "configure_card"));
	g_signal_connect(G_OBJECT(w), "clicked",
			G_CALLBACK(configure_card), NULL);

	w = GTK_WIDGET(gtk_builder_get_object(builder, "file_quit"));
	g_signal_connect(G_OBJECT(w), "activate",
			G_CALLBACK(gtk_main_quit), NULL);

	comm_w = GTK_WIDGET(gtk_builder_get_object(builder, "comm"));
	w = GTK_WIDGET(gtk_builder_get_object(builder, "tool_download"));
	g_signal_connect(G_OBJECT(w), "clicked",
			G_CALLBACK(download_all), NULL);
	w = GTK_WIDGET(gtk_builder_get_object(builder, "tool_upload"));
	g_signal_connect(G_OBJECT(w), "clicked",
			G_CALLBACK(upload_all), NULL);
	w = GTK_WIDGET(gtk_builder_get_object(builder, "tool_flash"));
	g_signal_connect(G_OBJECT(w), "clicked",
			G_CALLBACK(flash_all), NULL);

	xfer_w = GTK_WIDGET(gtk_builder_get_object(builder, "xfer_dialog"));
	xfer_spinner_w = GTK_WIDGET(gtk_builder_get_object(builder,
				"xfer_spinner"));

	w = GTK_WIDGET(gtk_builder_get_object(builder, "xfer_abort"));
	g_signal_connect(G_OBJECT(w), "clicked",
			G_CALLBACK(xfer_abort), NULL);
	g_signal_connect(G_OBJECT(xfer_w), "delete-event",
			G_CALLBACK(xfer_abort), NULL);

	xfer_ok = GTK_WIDGET(gtk_builder_get_object(builder, "xfer_ok"));
	g_signal_connect(G_OBJECT(xfer_ok), "clicked",
			G_CALLBACK(xfer_ok_clicked), NULL);

	preferences_wnd = GTK_WIDGET(gtk_builder_get_object(builder,
				"preferences"));
	g_signal_connect(G_OBJECT(preferences_wnd), "delete-event",
			G_CALLBACK(hide_preferences), NULL);
	w = GTK_WIDGET(gtk_builder_get_object(builder, "tool_preferences"));
	g_signal_connect(G_OBJECT(w), "clicked",
			G_CALLBACK(show_preferences), NULL);

	w = GTK_WIDGET(gtk_builder_get_object(builder, "preferences_ok"));
	g_signal_connect(G_OBJECT(w), "clicked",
			G_CALLBACK(update_preferences), NULL);
	w = GTK_WIDGET(gtk_builder_get_object(builder, "preferences_save"));
	g_signal_connect(G_OBJECT(w), "clicked",
			G_CALLBACK(save_preferences), NULL);
	w = GTK_WIDGET(gtk_builder_get_object(builder, "preferences_cancel"));
	g_signal_connect(G_OBJECT(w), "clicked",
			G_CALLBACK(hide_preferences), NULL);

	conn_tcp_w = gtk_builder_get_object(builder, "conn_tcp");
	conn_file_w = gtk_builder_get_object(builder, "conn_file");
	conn_tcp_host_w = gtk_builder_get_object(builder, "conn_tcp_host");
	conn_tcp_port_w = gtk_builder_get_object(builder, "conn_tcp_port");
	conn_file_path_w = gtk_builder_get_object(builder, "conn_file_path");
	conn_file_baudrate_w = gtk_builder_get_object(builder,
			"conn_file_baudrate");

	w = GTK_WIDGET(gtk_builder_get_object(builder, "gtk_about"));
	g_signal_connect(G_OBJECT(w), "activate",
			G_CALLBACK(show_about), NULL);

	w = GTK_WIDGET(gtk_builder_get_object(builder, "pommax3_update"));
	g_signal_connect(G_OBJECT(w), "clicked",
			G_CALLBACK(get_data_all), NULL);

	xfer_log_view = GTK_WIDGET(gtk_builder_get_object(builder,
				"xfer_log_view"));
	xfer_log = GTK_TEXT_BUFFER(gtk_builder_get_object(builder,
				"xfer_log"));

	gtk_widget_show(main_wnd);

	init_cards_list();
}

int main(int argc, char **args)
{
	const gchar *glade_file;

	setlocale(LC_MESSAGES, "");
	setlocale(LC_CTYPE, "");
	bindtextdomain("ar2c-par-utils", LOCALE_DIR);

	pacc = pci_alloc();
	if (pacc) {
		pci_init(pacc);
		pci_set_name_list_path(pacc, strdup(AR2CPARUTILSDIR "pci.ids"),
				1);
	}

	/*
	 * gtk uses UTF-8 to avoid explicit locale encoding -> UTF-8
	 * conversions we just change locale encoding to UTF-8
	 */
	bind_textdomain_codeset("ar2c-par-utils", "UTF-8");
	textdomain("ar2c-par-utils");

	fprintf(stderr, "ar2c-gconf " PACKAGE_VERSION ".\n"
			"Copyright (C) 2012..2015  Krzysztof Mazur.\n");

	gconf_load_config();

	ar2c_config_init(&ar2c_config);
	ar2c_config_init(&ar2c_data);

	gtk_init(&argc, &args);
	parse_command_line(argc, args);
	setlocale(LC_NUMERIC, "C");

	glade_file = AR2CPARUTILSDIR "gconf.glade";
	init_main_window(glade_file);
	gtk_main();
	return 0;
}
