/*
 * ar2c-par-utils - kz-parser.c
 * Copyright (C) 2007, 2008  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Fundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <inttypes.h>
#include <string.h>
#include <getopt.h>
#include <stdarg.h>

#ifdef HAVE_ENDIAN_H
#include <endian.h>
#endif

#ifdef HAVE_BYTESWAP_H
#include <byteswap.h>
#endif

#if BYTE_ORDER == BIG_ENDIAN

#define cpu_to_be16(x)		(x)
#define cpu_to_be32(x)		(x)
#define cpu_to_le16(x)		(bswap_16(x))
#define cpu_to_le32(x)		(bswap_32(x))

#define be16_to_cpu(x)		(x)
#define be32_to_cpu(x)		(x)
#define le16_to_cpu(x)		(bswap_16(x))
#define le32_to_cpu(x)		(bswap_32(x))

#else

#if BYTE_ORDER == LITTLE_ENDIAN

#define cpu_to_be16(x)		(bswap_16(x))
#define cpu_to_be32(x)		(bswap_32(x))
#define cpu_to_le16(x)		(x)
#define cpu_to_le32(x)		(x)

#define be16_to_cpu(x)		(bswap_16(x))
#define be32_to_cpu(x)		(bswap_32(x))
#define le16_to_cpu(x)		(x)
#define le32_to_cpu(x)		(x)

#else

#error "Only little and big endian systems are supported."

#endif
#endif

typedef uint8_t u8;

static char *kz_filename;
static int verbose;
static const char *kz_format;
static int have_filters;

struct buf {
	const u8 *data;
	unsigned int len;
};

static void die(const char *err, ...)
{
	va_list params;

	va_start(params, err);
	vfprintf(stderr, err, params);
	va_end(params);
	exit(EXIT_FAILURE);
}

static void *xmalloc(size_t size)
{
	void *p = malloc(size);

	if (!p) {
		die("cannot allocate memory, %lu bytes\n",
				(unsigned long) size);
	}
	return p;
}

static void *xrealloc(void *p, size_t size)
{
	p = realloc(p, size);

	if (!p) {
		die("cannot reallocate memory to %lu bytes\n",
				(unsigned long) size);
	}
	return p;
}

static size_t buf_get(struct buf *b, void *p, size_t len)
{
	if (len > b->len)
		len = b->len;
	memcpy(p, b->data, len);
	b->data += len;
	b->len -= len;
	return len;
}

static size_t buf_get_or_die(struct buf *b, void *p, size_t len)
{
	size_t ret;

	ret = buf_get(b, p, len);
	if (ret < len) {
		die("unexpected end of file, wanted %lu bytes, read %lu\n",
				(unsigned long) len,
				(unsigned long) ret);
	}
	return ret;
}

static uint8_t read_byte(struct buf *b)
{
	uint8_t x;

	buf_get_or_die(b, &x, sizeof(x));
	return x;
}


static const char *read_string(struct buf *b)
{
	uint8_t len;
	static char str[257];

	buf_get_or_die(b, &len, sizeof(len));
	buf_get_or_die(b, str, len);
	str[len] = '\0';
	return str;
}

static int read_short(struct buf *b)
{
	uint16_t x;

	buf_get_or_die(b, &x, sizeof(x));
	return (int16_t) le16_to_cpu(x);
}

static int read_bool(struct buf *b)
{
	uint32_t x;

	buf_get_or_die(b, &x, sizeof(x));
	return x != 0;
}

static float read_float(struct buf *b)
{
	union {
		uint32_t i;
		float f;
	} x;

	buf_get_or_die(b, &x.i, sizeof(x.i));
	x.i = le32_to_cpu(x.i);
	return x.f;
}

static int32_t read_int(struct buf *b)
{
	int32_t x;

	buf_get_or_die(b, &x, sizeof(x));
	return le32_to_cpu(x);
}

#define kz_all(fmt...)	printf("KZ:0:" fmt)
#define kz_high(fmt...)	printf("KZ:1:" fmt)
#define kz_med(fmt...)	printf("KZ:2:" fmt)
#define kz_low(fmt...)	printf("KZ:3:" fmt)

static const char *read_short_p(struct buf *b)
{
	static char buf[32];
	unsigned short x = read_short(b);

	if (x == 0xffff)
		snprintf(buf, 32, "-1");
	else
		snprintf(buf, 32, "0%04hxh", x);
	return buf;
}

static const char *read_short_p2(struct buf *b)
{
	static char buf[32];
	unsigned short x = read_short(b);

	if (x == 0xffff)
		snprintf(buf, 32, "-1");
	else
		snprintf(buf, 32, "0%02xh", x);
	return buf;
}

static const char *read_int_p2(struct buf *b)
{
	static char buf[32];
	unsigned int x = read_int(b);

	if (x == 0xffff)
		snprintf(buf, 32, "-1");
	else
		snprintf(buf, 32, "0%02hxh", x);
	return buf;
}

static const char *read_int_p2b(struct buf *b)
{
	static char buf[32];
	unsigned int x = read_int(b);

	if (x == 0xff)
		snprintf(buf, 32, "-1");
	else
		snprintf(buf, 32, "0%02hxh", x);
	return buf;
}

static void parse_dp(int i, struct buf *b)
{
	kz_high("card_bus_base %d:%s\n", i, read_short_p(b));
	kz_high("card_bus_sbin_offset %d:%s\n", i, read_short_p2(b));
	kz_high("card_bus_dest_offset %d:%s\n", i, read_int_p2(b));
	kz_high("card_bus_bin_lth %d:%s\n", i, read_int_p2(b));
	kz_high("card_bus_sanal_offset %d:%s\n", i, read_short_p2(b));
	kz_high("card_bus_anal_offset %d:%s\n", i, read_int_p2(b));
	kz_high("card_bus_anal_lth %d:%s\n", i, read_int_p2(b));
	kz_high("card_bus_sHB_offset %d:%s\n", i, read_short_p2(b));
	kz_high("card_bus_HB_offset %d:%s\n", i, read_int_p2(b));
	kz_high("card_bus_HB_lth %d:%s\n", i, read_int_p2(b));
	kz_high("card_bus_stat_adr %d:%s\n", i, read_short_p2(b));
	kz_high("card_bus_stat_ok %d:0%02xh\n", i, read_int(b));
	kz_high("card_bus_stat_bad %d:0%02xh\n", i, read_int(b));
	kz_high("card_bus_chk_value %d:0%02xh\n", i, read_int(b));
}

static void parse_wc8(int i, struct buf *b)
{
	kz_high("WC8_block_base %d:%s\n", i, read_short_p(b));
	kz_high("WC8_dest_offset %d:%s\n", i, read_int_p2b(b));
	kz_high("WC8_block_lth %d:%dd\n", i, read_int(b));
}

static const char *convert_tabway(uint8_t idx)
{
	static const char *tabways[256] = {
		[0] = "noop",
		[1] = "obs_alarmu",
		[2] = "obs_hamulca",
		[3] = "hamulec_bezp",
		[4] = "korekcja_k1",
	};
	static char buf[16];

	if (tabways[idx] == NULL) {
		snprintf(buf, sizeof(buf), "%02x (%d)", idx, idx);
		return buf;
	}
	return tabways[idx];
}

void mk_analog(char *s, size_t size, const char *name, const char *jedn,
		const char *skala)
{
	if (strlen(jedn) || strlen(skala))
		snprintf(s, size, "%s[1dz -%s%s]", name, skala, jedn);
	else
		snprintf(s, size, "%s", name);
}

static void tabway(int id, unsigned int fun)
{
	/*
	 * tabways in kz files are broken. I don't know why but
	 * some tabway functions are not defined in kz files.
	 * If such function exist in asm the kz files have function
	 * id equal to 0 (noop). So when we read 0 function we
	 * don't really know if user disabled some signal or just
	 * was unable to use undefined function.
	 *
	 * We can translate 0 function as noop or just skip this tabway.
	 * When we always generate noop all undefined in kz functions
	 * are corrupted. When we skip such tabways the the user
	 * cannot change function to noop. So we're using simple
	 * heuristics here. The undefined functions are generally
	 * used for signals 15 and 22. The comment in asm file
	 * template says that 20 to 23 tabways should not be
	 * changed. So we simply skip 15 and 20 to 23.
	 */
	if (!fun) {
		if (id == 15 || (id >= 20 && id < 24))
			return;
	}
	kz_low("tabway %d:%s\n", id, convert_tabway(fun));
}

static const char *convert_par1(int x)
{
	static char s[32];

	switch (x) {
	case 4:
		return "proc50";
	case 5:
		return "proc25";
	case 6:
		return "proc12";
	case 7:
		return "proc6";
	case 8:
		return "proc3";
	case 9:
		return "proc1";
	case 10:
		return "proc0";
	case 11:
		return "proc_nic";
	}
	snprintf(s, sizeof(s), "%dd", x);
	return s;
}

static void parse_filter(int i, struct buf *b)
{
	kz_high("filter_nr %d:%dd\n", i, read_byte(b));
	kz_high("fllter_par1 %d:%s\n", i, convert_par1(read_short(b)));
	kz_high("filter_par2 %d:%dd\n", i, read_short(b));
}

static int detect_filters(struct buf *b)
{
	return b->len >= 304;
}

static void parse_kz(struct buf *b)
{
	int i, j;

	kz_med("nazwa szybu:%s\n", read_string(b));
	kz_med("nazwa kopalni:%s\n", read_string(b));
	for (i = 0; i < 32; i++) {
		char nazwy[17], *nt;
		static int d_counter;

		for (nt = nazwy, j = 0; j < 8; j++) {
			const char *name = read_string(b);

			if (name == NULL)
				continue;
			kz_low("dwustan %d:%-2s\n", 8 * i + j, name);
			if (*name == '\xf1') {
				*nt = '\0';
				kz_low("nazwy sygnalow %d:\"%s\"\n",
						d_counter++, nazwy);
				nt = nazwy;
				if (!name[1])
					abort();
				nt += sprintf(nt, "%c", name[1]);
			} else {
				nt += sprintf(nt, "%-2.2s", name);
				if (i * 8 + j == 20) {
					*nt = '\0';
					kz_low("nazwy sygnalow %d:\"%s\"\n",
							d_counter++, nazwy);
					nt = nazwy;
				}
			}
			if (i < 8)
				tabway(8 * i + j, read_byte(b));
		}
		*nt = '\0';
		kz_low("nazwy sygnalow %d:\"%s\"\n", d_counter++, nazwy);
		if (i < 16) {
			char *name, *jedn, *skala;
			static char msg[48], msg2[48];

			name = strdup(read_string(b));
			jedn = strdup(read_string(b));
			skala = strdup(read_string(b));
			kz_low("analog %d:%s\n", i, name);
			kz_low("analog jedn %d:%s\n", i, jedn);
			kz_low("analog skala %d:%s\n", i, skala);
			kz_low("analog offset %d:0%02hxh\n", i, read_int(b));
			kz_med("analog fWpol %d:%f\n", i, read_float(b));

			if (i & 1) {
				int od;
				char *t;
				static int opis_id;
				int len = strlen(msg);

				mk_analog(msg2, 42, name, jedn, skala);
				od = 42 - len - strlen(msg2);
				t = msg + len;
				while (od-- > 0) {
					*t++ = ' ';
					len++;
				}
				strncpy(t, msg2, 42 - len);
				msg[42] = '\0';
				kz_low("opis skali %d:\"%s\"\n", opis_id++,
						msg);
			} else {
				mk_analog(msg, 42, name, jedn, skala);
			}
		}
	}
	kz_low("dp:%d\n", read_bool(b));
	kz_low("pommax:%d\n", read_bool(b));
	kz_low("wa2:%d\n", read_bool(b));
	kz_med("tachozm:%d\n", read_bool(b));
	kz_all("szybowskaz:%d\n", read_bool(b));
	kz_all("imp:%d\n", read_int(b));
	kz_med("wc8:%d\n", read_int(b));
	kz_med("ilpom:%d\n", read_int(b));
	kz_med("ilkankom:%d\n", read_int(b));
	for (i = 0; i < 32; i++)
		kz_low("analog map %d:%03hhxh\n", i, read_byte(b));
	kz_all("pkk1:%d\n", read_int(b));
	kz_all("pkk2:%d\n", read_int(b));

	if (!kz_format)
		have_filters = detect_filters(b);

	if (have_filters) {
		for (i = 0; i < 16; i++)
			parse_filter(i, b);
	}

	for (i = 0; i < 4; i++) {
		parse_dp(i, b);
		parse_wc8(i, b);
	}
}

static void show_usage(char *name)
{
	fprintf(stderr, "kz-parser " PACKAGE_VERSION ".\n"
"Copyright (C) 2007, 2008  Krzysztof Mazur <krzysiek@podlesie.net>.\n\n");

	fprintf(stderr,	"usage: %s [options]\n"
			"Options:\n"
			"  -h, --help		show this help\n"
			"  -f, --format=FORMAT	kz file format\n"
			"  -v, --verbose\t	be verbose\n"
			"  -V, --version\t	show version\n",
			(name == NULL) ? "kz-parser" : name);
	exit(EXIT_SUCCESS);
}

static void show_version(void)
{
	fprintf(stderr, "kz-parser " PACKAGE_VERSION ".\n"
"Copyright (C) 2007..2008  Krzysztof Mazur <krzysiek@podlesie.net>.\n\n");

	fprintf(stderr,	"This program comes with ABSOLUTELY NO WARRANTY; "
			"see COPYING for details.\n"
			"This is free software, and you are welcome to "
			"redistribute it\n"
			"under certain conditions; see COPYING for details.\n");
	exit(EXIT_SUCCESS);
}

static void parse_command_line(int argc, char **args)
{
	int c;
	int option_index = 0;

	static struct option long_options[] = {
		{"format", 1, 0, 'f'},
		{"help", 0, 0, 'h'},
		{"verbose", 0, 0, 'v'},
		{"version", 0, 0, 'V'},
		{0, 0, 0, 0},
	};

	while (1) {
		c = getopt_long(argc, args, "f:hvV",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'f':
			kz_format = strdup(optarg);
			break;
		case 'h':
			show_usage(args[0]);
			break;
		case 'v':
			verbose++;
			break;
		case 'V':
			show_version();
			break;
		default:
			die("Unknown option.\n");
		}
	}
	if (optind < argc) {
		kz_filename = args[optind++];
		if (optind < argc)
			fprintf(stderr, "warning: unknown parameters detected."
					"\n");
	}
}

static const void *read_file(int fd, size_t *flen)
{
	char *p;
	ssize_t ret;
	size_t size = 65536;
	size_t len = 0;

	p = xmalloc(size);
	do {
		if (len == size) {
			size *= 2;
			if (size < len)
				die("too large file");
			p = xrealloc(p, size);
		}
		ret = read(fd, p + len, size - len);
		if (ret < 0) {
			if (errno == EINTR)
				continue;
			die("read error: %s", strerror(errno));
		}
		len += ret;
	} while (ret > 0);
	*flen = len;
	return p;
}

int main(int argc, char **args)
{
	int fd = STDIN_FILENO;
	struct buf b;
	const void *buf;
	size_t len;

	parse_command_line(argc, args);
	if (kz_filename != NULL) {
		fd = open(kz_filename, O_RDONLY);
		if (fd == -1) {
			perror("open");
			exit(EXIT_FAILURE);
		}
	}
	buf = read_file(fd, &len);
	if (kz_format) {
		if (!strcmp(kz_format, "r2011") || !strcmp(kz_format, "r2012"))
			have_filters = 1;
	}
	b.data = buf;
	b.len = len;
	parse_kz(&b);

	return EXIT_SUCCESS;
}
