/*
 * Logger file crypt
 *
 * Copyright (C) 2008..2013  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <stdint.h>

static int format;

typedef uint8_t u8;

#define BIN_RECORD_LEN		70
#define BIN_RECORD_ID		0
#define BIN_RECORD_NAME		14

/*
 * It's a 69 character length Pascal string. Rest of key is probably
 * unused.
 *
 * This key is extracted from lognr.dat - datalogger number extractor
 * from Logger shaft files.
 */
static unsigned char logger_key[256] = {
	0x69, 0x75, 0x72, 0x75,  0x63, 0x68, 0x6f, 0x6d,	/* 00 */
	0x69, 0x65, 0x6e, 0x69,  0x65, 0x20, 0x70, 0x6f,	/* 08 */
	0x7a, 0x61, 0x20, 0x73,  0x69, 0x65, 0x63, 0x69,	/* 10 */
	0x86, 0x20, 0x2d, 0x20,  0x75, 0xa7, 0x79, 0x74,	/* 18 */
	0x6b, 0x6f, 0x77, 0x6e,  0x69, 0x6b, 0x20, 0x75,	/* 20 */
	0x70, 0x72, 0x61, 0x77,  0x6e, 0x69, 0x6f, 0x6e,	/* 28 */
	0x79, 0x20, 0x74, 0x79,  0x6c, 0x6b, 0x6f, 0x20,	/* 30 */
	0x64, 0x6f, 0x20, 0x70,  0x72, 0x7a, 0x65, 0x67,	/* 38 */
	0x6c, 0x86, 0x64, 0x61,  0x6e, 0x69, 0x61, 0x62,	/* 40 */
	0x72, 0x61, 0x6b, 0x20,  0x70, 0x6c, 0x69, 0x6b,	/* 48 */
	0x75, 0x20, 0x6b, 0x6f,  0x6e, 0x66, 0x69, 0x67,	/* 50 */
	0x75, 0x72, 0x61, 0x63,  0x79, 0x6a, 0x6e, 0x65,	/* 58 */
	0x67, 0x6f, 0x71, 0x65,  0x78, 0x76, 0x23, 0x72,	/* 60 */
	0x2a, 0x65, 0x00, 0x00,  0x60, 0x77, 0x8b, 0xaf,	/* 68 */
	0x1f, 0x37, 0x05, 0x08,  0x00, 0x00, 0x00, 0x00,	/* 70 */
	0x60, 0x30, 0xf6, 0xa7,  0x00, 0x6f, 0x05, 0x08,	/* 78 */
	0x00, 0x00, 0x00, 0x00,  0x20, 0x00, 0x00, 0x00,	/* 80 */
	0x60, 0x30, 0xf6, 0xa7,  0x20, 0x00, 0x00, 0x00,	/* 88 */
	0x0c, 0x00, 0x00, 0x00,  0x0c, 0x00, 0x00, 0x00,	/* 90 */
	0x28, 0x77, 0x8b, 0xaf,  0x9b, 0xa3, 0x04, 0x08,	/* 98 */
	0xd0, 0x79, 0x8b, 0xaf,  0x0c, 0x00, 0x00, 0x00,	/* a0 */
	0x87, 0x8f, 0x04, 0x08,  0x00, 0x6f, 0x05, 0x08,	/* a8 */
	0xac, 0x77, 0x8b, 0xaf,  0x86, 0xe4, 0x04, 0x08,	/* b0 */
	0x90, 0x77, 0x8b, 0xaf,  0x19, 0xbd, 0x04, 0x08,	/* b8 */
	0x03, 0x00, 0x00, 0x00,  0x06, 0x00, 0x00, 0x00,	/* c0 */
	0x20, 0xaa, 0x05, 0x08,  0x1c, 0x00, 0x00, 0x00,	/* c8 */
	0x20, 0xaa, 0x05, 0x08,  0x06, 0x00, 0x00, 0x00,	/* d0 */
	0x1c, 0x00, 0x00, 0x00,  0x40, 0xaa, 0x05, 0x08,	/* d8 */
	0x08, 0x7b, 0x8b, 0xaf,  0xb4, 0x77, 0x8b, 0xaf,	/* e0 */
	0xe7, 0xbc, 0x04, 0x08,  0x06, 0x00, 0x00, 0x00,	/* e8 */
	0x06, 0x00, 0x00, 0x00,  0x40, 0xaa, 0x05, 0x08,	/* f0 */
	0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00,	/* f8 */
};

/*
 * Decode single character from two characters. y is decoded value or
 * logger key index. x is used to specify which case occured.
 */
static unsigned char cipher_decode_char(unsigned char *key, unsigned char x,
		unsigned char y)
{
	return (x > 127) ? y : key[y];
}

/*
 * Decode len bytes. Source buffer must be 2 * len bytes length. Output
 * is len bytes.
 *
 * buf is used for input and output buffer.
 */
static void cipher_decode(void *buf, int len)
{
	unsigned char *b = buf;
	int i;

	for (i = 0; i < len; i++)
		b[i] = cipher_decode_char(logger_key, b[2 * i],
				b[2 * i + 1]);
}

/*
 * xread() is the same a read(), but it automatically restarts read()
 * operations with a recoverable error (EAGAIN and EINTR). xread()
 * DOES NOT GUARANTEE that "len" bytes is read even if the data is available.
 */
ssize_t xread(int fd, void *buf, size_t len)
{
	ssize_t nr;
	while (1) {
		nr = read(fd, buf, len);
		if ((nr < 0) && (errno == EAGAIN || errno == EINTR))
			continue;
		return nr;
	}
}

/*
 * xwrite() is the same a write(), but it automatically restarts write()
 * operations with a recoverable error (EAGAIN and EINTR). xwrite() DOES NOT
 * GUARANTEE that "len" bytes is written even if the operation is successful.
 */
ssize_t xwrite(int fd, const void *buf, size_t len)
{
	ssize_t nr;
	while (1) {
		nr = write(fd, buf, len);
		if ((nr < 0) && (errno == EAGAIN || errno == EINTR))
			continue;
		return nr;
	}
}

ssize_t read_in_full(int fd, void *buf, size_t count)
{
	char *p = buf;
	ssize_t total = 0;

	while (count > 0) {
		ssize_t loaded = xread(fd, p, count);
		if (loaded < 0)
			return -1;
		if (loaded == 0)
			return total;
		count -= loaded;
		p += loaded;
		total += loaded;
	}

	return total;
}

ssize_t write_in_full(int fd, const void *buf, size_t count)
{
	const char *p = buf;
	ssize_t total = 0;

	while (count > 0) {
		ssize_t written = xwrite(fd, p, count);
		if (written < 0)
			return -1;
		if (!written) {
			errno = ENOSPC;
			return -1;
		}
		count -= written;
		p += written;
		total += written;
	}

	return total;
}

static void print_name(unsigned int id, unsigned char *r)
{
	unsigned int i;
	unsigned int id_len = r[BIN_RECORD_ID] - 0x60;
	unsigned int len = r[BIN_RECORD_NAME] - 0x60;

	if (id % 100 >= 50)
		return;

	if (id_len > 4)
		return;
	if (len > 16)
		return;

	printf("logger:3:caption ");
	for (i = 1; i <= id_len; i++)
		printf("%c", r[BIN_RECORD_ID + i]);
	printf(":");
	for (i = 1; i <= len; i++)
		printf("%c", r[BIN_RECORD_NAME + i]);
	printf("\n");
}

static void decode_raw(int fd, int ofd)
{
	unsigned char buf[4096];
	int ret;

	do {
		ret = read_in_full(fd, buf, sizeof(buf));
		if (ret < 0) {
			perror("read");
			exit(EXIT_FAILURE);
		}

		cipher_decode(buf, ret / 2);
		ret = write_in_full(ofd, buf, ret / 2);
		if (ret < 0) {
			perror("write");
			exit(EXIT_FAILURE);
		}
	} while (ret > 0);
}

static void decode_binary(int fd)
{
	unsigned char records[100 * BIN_RECORD_LEN];
	unsigned int id = 0;
	unsigned int i;
	int ret;

	do {
		ret = read_in_full(fd, records, sizeof(records));
		if (ret < 0) {
			perror("read");
			exit(EXIT_FAILURE);
		}

		cipher_decode(records, ret / 2);
		for (i = 0; ret - i >= BIN_RECORD_LEN; i += BIN_RECORD_LEN)
			print_name(id++, records + i);
	} while (ret > 0);
}

int main(int argc, char **args)
{
	int fd = STDIN_FILENO;
	int ofd = STDOUT_FILENO;

	if (argc > 1) {
		fd = open(args[1], O_RDONLY);
		if (fd == -1) {
			perror("open");
			exit(EXIT_FAILURE);
		}
	}
//	decode_raw(fd, ofd);
	decode_binary(fd);
	return 0;
}
