/*
 * ar2c-lp-utils - par-reset.c
 *
 * Copyright (C) 2008  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <comm.h>
#include <time.h>
#include <ar2c.h>
#include <signal.h>
#include <tread.h>

#define TT_STDIN	0
#define TT_FILE		1
#define TT_TCP		2
#define TT_FIFO		3
#define TT_SERIAL	4

static int targ_type = TT_STDIN;
static char *targ_opt = NULL;
static char *targ_opt2 = NULL;
static const char *port = "2000";
static int ifd, ofd;
static int verbose = 0;
static FILE *out;
static speed_t bspeed = B9600;
static int speed = 0, rspeed = 0;
int cmd_timeout = 0;
static int synchronize = 1;

static void show_usage(char *name)
{
	fprintf(stderr, "ar2cputdata " PACKAGE_VERSION ".\n"
"Copyright (C) 2008  Krzysztof Mazur <krzysiek@podlesie.net>.\n\n");

	fprintf(stderr,	"usage: %s [options]\n"
			"Options:\n"
			"      --bbitrate	base bitrate(default 9600bps)\n"
			"  -a, --no-sync\t	don't use 0xff 0xd2 sync\n"
			"  -b, --bitrate\t	set bitrate\n"
			"  -d, --device		use selected device\n"
			"  -f, --fifo		use dual fifo(for wr)\n"
			"  -h, --help		show this help\n"
			"  -p, --port		remote port(2000)\n"
			"  -q, --quiet		be quiet\n"
			"      --rbitrate	return bitrate\n"
			"  -s, --host		remote host(localhost)\n"
			"  -t, --timeout\t	set timeout\n"
			"  -v, --verbose\t	be verbose\n"
			"  -V, --version\t	show version\n",
			(name == NULL) ? "ar2c-par-reset" : name);
	exit(EXIT_SUCCESS);
}

static void show_version(void)
{
	fprintf(stderr, "ar2cputdata " PACKAGE_VERSION ".\n"
"Copyright (C) 2008  Krzysztof Mazur <krzysiek@podlesie.net>.\n\n");

	fprintf(stderr,	"This program comes with ABSOLUTELY NO WARRANTY; "
			"see COPYING for details.\n"
			"This is free software, and you are welcome to "
			"redistribute it\n"
			"under certain conditions; see COPYING for details.\n");
	exit(EXIT_SUCCESS);
}

static void parse_command_line(int argc, char **args)
{
	int c;
	int option_index = 0;

	static struct option long_options[] = {
		{"bbitrate", 1, 0, 1},
		{"bitrate", 1, 0, 'b'},
		{"device", 1, 0, 'd'},
		{"fifo", 1, 0, 'f'},
		{"help", 0, 0, 'h'},
		{"host", 1, 0, 's'},
		{"no-sync", 0, 0, 'a'},
		{"port", 1, 0, 'p'},
		{"quiet", 0, 0, 'q'},
		{"rbitrate", 1, 0, 2},
		{"timeout", 1, 0, 't'},
		{"verbose", 0, 0, 'v'},
		{"version", 0, 0, 'V'},
		{0, 0, 0, 0},
	};

	while (1) {
		c = getopt_long(argc, args, "ab:d:f:g:p:hs:t:vqV",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 1:
			bspeed = itospeed(atoi(optarg));
			if (bspeed < 0) {
				fprintf(stderr, "Unsupported initial "
						"speed setting.\n");
				exit(EXIT_FAILURE);
			}
			if (rspeed == 0) {
				rspeed = speedtoar2cspeed(bspeed);
				if (rspeed < 0)
					rspeed = 0;
			}
			targ_type = TT_SERIAL;
			break;
		case 2:
			rspeed = itoar2cspeed(atoi(optarg));
			if (rspeed < 0) {
				fprintf(stderr, "Invalid return speed "
							"setting.");
				exit(EXIT_FAILURE);
			}
			targ_type = TT_SERIAL;
			break;
		case 'a':
			synchronize = 0;
			break;
		case 'b':
			speed = itoar2cspeed(atoi(optarg));
			if (speed < 0) {
				fprintf(stderr, "Invalid speed setting.");
				exit(EXIT_FAILURE);
			}
			targ_type = TT_SERIAL;
			break;
		case 'd':
			targ_opt = optarg;
			if ((targ_type == TT_TCP) || (targ_type == TT_STDIN))
				targ_type = TT_FILE;
			break;
		case 'f':
			targ_opt2 = optarg;
			targ_type = TT_FIFO;
			break;
		case 'h':
			show_usage(args[0]);
			break;
		case 'p':
			port = strdup(optarg);
			targ_type = TT_TCP;
			break;
		case 'v':
			verbose++;
			break;
		case 'q':
			verbose--;
			break;
		case 's':
			targ_opt = optarg;
			targ_type = TT_TCP;
			break;
		case 't':
			cmd_timeout = atoi(optarg);
			break;
		case 'V':
			show_version();
			break;
		default:
			fprintf(stderr, "Unknown option.\n");
			exit(EXIT_FAILURE);
		}
	}
	if (optind < argc)
		fprintf(stderr, "warning: unknown parameters detected.\n");
}

#define log_warning(fmt...)	fprintf(stderr, fmt)

static ssize_t write_all(int fd, void *buf, size_t len)
{
	char *t = buf;
	ssize_t ret, total = 0;

	while (total < len) {
		ret = write(fd, t + total, len - total);
		if (ret <= 0) {
			if (!total)
				return ret;
			break;
		}
		total += ret;
	}
	return total;
}

static void spd_return(void)
{
	if (speed) {
		if (rspeed == 0) {
			rspeed = speedtoar2cspeed(bspeed);
			if (rspeed < 0) {
				fprintf(stderr, "Unable to return to initial "
						"serial speed.\n");
				exit(EXIT_FAILURE);
			}
		}
		if ((speed != rspeed) && ar2c_set_speed(ifd, rspeed)) {
			fprintf(stderr, "Setting serial speed error\n");
			exit(EXIT_FAILURE);
		}
	}
}

static int __do_par_reset(void)
{
	static char __sync_seq[] = { 0xff, 0x0f, 0x08, 0x00, 0xff, 0xd2 };
	int ret;
	char buf[2];

	do {
		ret = write_all(ofd, __sync_seq, sizeof(__sync_seq));
		if (ret < sizeof(__sync_seq))
			return -errno;

		ret = t_read(ifd, buf, 2, 300);
		if (ret < 0)
			return -errno;
	} while (ret < 2);
	return 0;
}

static int do_par_reset(void)
{
	unsigned int i;

	/*
	 * New AR-2c dataloggers require "reset parameter" command issued
	 * at least 2 or 3 times.
	 */	
	for (i = 0; i < 4; i++)
		__do_par_reset();
	return 0;
}

int main(int argc, char **args)
{
	signal(SIGALRM, SIG_IGN);
	parse_command_line(argc, args);
	switch (targ_type) {
	case TT_STDIN:
		ifd = 0;
		ofd = 1;
		out = stderr;
		cmd_timeout = cmd_timeout ? : 2000;
		break;
	case TT_FILE:
		if (targ_opt == NULL) {
			fprintf(stderr, "fatal: null targ_opt.\n");
			exit(EXIT_FAILURE);
		}
		if ((ifd = open(targ_opt, O_RDWR)) < 0) {
			fprintf(stderr, "error: Can't open %s:%s\n",
					targ_opt, strerror(errno));
			exit(EXIT_FAILURE);
		}
		ofd = ifd;
		out = stdout;
		cmd_timeout = cmd_timeout ? : 2000;
		break;
	case TT_TCP:
		if (targ_opt == NULL) {
			fprintf(stderr, "fatal: null targ_opt.\n");
			exit(EXIT_FAILURE);
		}
		if ((ifd = pcomm_open(targ_opt, port)) < 0) {
			fprintf(stderr, "error: Can't connect to %s\n",
					targ_opt);
			exit(EXIT_FAILURE);
		}
		ofd = ifd;
		out = stdout;
		cmd_timeout = cmd_timeout ? : 10000;
		break;
	case TT_FIFO:
		if (targ_opt == NULL) {
			fprintf(stderr, "fatal: null targ_opt.\n");
			exit(EXIT_FAILURE);
		}
		if (targ_opt2 == NULL) {
			fprintf(stderr, "fatal: null targ_opt2.\n");
			exit(EXIT_FAILURE);
		}
		/* order is important */
		if ((ofd = open(targ_opt2, O_WRONLY)) == -1) {
			fprintf(stderr, "fatal: Can't open wr_fifo.\n");
			exit(EXIT_FAILURE);
		}
		if ((ifd = open(targ_opt, O_RDONLY)) == -1) {
			fprintf(stderr, "fatal: Can't open rd_fifo.\n");
			exit(EXIT_FAILURE);
		}
		out = stdout;
		cmd_timeout = cmd_timeout ? : 500;
		break;
	case TT_SERIAL:
		if (targ_opt == NULL) {
			fprintf(stderr, "fatal: null targ_opt.\n");
			exit(EXIT_FAILURE);
		}
		if ((ifd = open(targ_opt, O_RDWR)) < 0) {
			fprintf(stderr, "error: Can't open %s:%s\n",
					targ_opt, strerror(errno));
			exit(EXIT_FAILURE);
		}
		ofd = ifd;
		serial_set_speed(ifd, bspeed);
		if (ar2cspeedtospeed(speed) != bspeed)
			if (ar2c_set_speed(ifd, speed)) {
				fprintf(stderr, "Setting serial speed "
						"error\n");
				exit(EXIT_FAILURE);
			}
		out = stdout;
		cmd_timeout = cmd_timeout ? : 500;
		break;
	default:
		fprintf(stderr, "fatal: Internal error.\n");
		exit(EXIT_FAILURE);
	}

	atexit(spd_return);
	if (synchronize)
		do_resync(ifd, ofd);
	do_par_reset();
	return EXIT_SUCCESS;
}
