/*
 * pommax3
 *
 * Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arbusctl.h>
#include <ar2cpar.h>
#include <pci_regs.h>
#include <mclock.h>

#define POMMAX3_ADC0_PTR		0x80
#define POMMAX3_ADC1_PTR		0xc0

#define POMMAX3_FLAGS_GAIN		1
#define POMMAX3_FLAGS_OFFSET		2

#ifndef ARRAY_SIZE
#define ARRAY_SIZE(x) (sizeof((x)) / sizeof((x)[0]))
#endif

static unsigned long config;
static unsigned long legacy_base;
static unsigned long adc_base = POMMAX3_ADC0_PTR;
static int adc_dead[2];

#define POMMAX3_BAR1_ADDRESS	0xd8000

static void pommax3_assign_resources(struct arbus_dev *dev)
{
	u16 command;
	u16 x;
	u32 bar1;

	fprintf(stderr, "Assigning POMMAX3 resources...\n");

	x = ar2c_readw(0x11e00);
	fprintf(stderr, "x = %04x\n", x);

	/* We need only BAR1 */
	command = arbus_read_word(dev, PCI_COMMAND);
	fprintf(stderr, "command: %04x\n", command);

	fprintf(stderr, "disabling memory decoding\n");
	command &= ~PCI_COMMAND_MEMORY;

	arbus_write_word(dev, PCI_COMMAND, command);
	command = arbus_read_word(dev, PCI_COMMAND);
	fprintf(stderr, "command: %04x\n", command);

	x = ar2c_readw(0x11e00);
	fprintf(stderr, "x = %04x\n", x);

	bar1 = arbus_read_long(dev, PCI_BASE_ADDRESS_1);
	fprintf(stderr, "bar1 = %05x\n", bar1);

	arbus_write_long(dev, PCI_BASE_ADDRESS_1, POMMAX3_BAR1_ADDRESS);
	bar1 = arbus_read_long(dev, PCI_BASE_ADDRESS_1);
	fprintf(stderr, "newbar1 = %05x\n", bar1);

	command = arbus_read_word(dev, PCI_COMMAND);
	fprintf(stderr, "command: %04x\n", command);

	fprintf(stderr, "enabling memory decoding\n");
	command |= PCI_COMMAND_MEMORY;
	arbus_write_word(dev, PCI_COMMAND, command);

	command = arbus_read_word(dev, PCI_COMMAND);
	fprintf(stderr, "command: %04x\n", command);

	x = ar2c_readw(0x11e00);
	fprintf(stderr, "x = %04x\n", x);
}

#define POMMAX3_ADC_CFLAGS	0x8
#define POMMAX3_ADC_CFLAGS_USERSEQ	4
#define POMMAX3_ADC_RX		0x10
#define POMMAX3_ADC_TX		0x30
#define POMMAX3_ADC_CTRL	0x20

#define POMMAX3_ADC_RESET	0x01

#define POMMAX3_PROTO_CONFIG	0x02
#define POMMAX3_MAX_CMD_SIZE	3

#define POMMAX3_ADC_EXEC	0x01
#define POMMAX3_ADC_READ	0x02
#define POMMAX3_ADC_WRITE	0x03

#define POMMAX3_ADC_FREQ	0x1
#define POMMAX3_ADC_FLAGS	0x2
#define POMMAX3_ADC_OFFSET	0x3
#define POMMAX3_ADC_SERIAL	0x4
#define POMMAX3_ADC_GAIN_0	0x8
#define POMMAX3_ADC_OFIX_0	0x10
#define POMMAX3_ADC_MAJOR	0x80
#define POMMAX3_ADC_MINOR	0x81
#define POMMAX3_ADC_SAVE	0x82

#define POMMAX3_ADC_SAVE_MAGIC	0xadde

#define readb(addr)	ar2c_readb(addr)
#define writeb(x, addr)	ar2c_writeb(x, addr)
#define readw(addr)	ar2c_readw(addr)
#define writew(x, addr)	ar2c_writew(x, addr)

static inline u16 bswap16(u16 x)
{
	return (x >> 8) | (x << 8);
}

#ifndef writew_be
#define writew_be(x, addr)	writew(bswap16(x), addr)
#endif

/*
 * pommax3_wait - wait for command completion
 *
 * The time needed to send command to ADC depends on sampling frequency.
 */
static int pommax3_wait(u8 flags)
{
	unsigned long long t;
	unsigned int i;
	u8 x;

	t = time_ns();
	i = 0;
	for (;;) {
		x = readb(config + adc_base + POMMAX3_ADC_CFLAGS);

		if ((x ^ flags) == POMMAX3_ADC_CFLAGS_USERSEQ)
			return 0;
		if (time_ns() >= t + 200000000ULL)
			return -ETIMEDOUT;
		i++;
		if (i >= 4)
			usleep(1000);
		else if (i >= 10)
			usleep(5000);
	}
}

static int __exec_command(const u16 *cmd, unsigned int size,
		u16 *result, unsigned int rsize)
{
	u16 cmd_be[8];
	unsigned int i;
	u16 csum;
	u8 flags;
	int ret;
	unsigned int retry = 5;

	if (size > POMMAX3_MAX_CMD_SIZE)
		return -EINVAL;

	do {
		retry--;
		if (!retry)
			return -ETIMEDOUT;

		csum = 0;
		for (i = 0; i < size; i++) {
			csum += cmd[i];
			cmd_be[i] = bswap16(cmd[i]);
		}
		csum = ~csum & 0x7fff;
		cmd_be[i] = bswap16(csum);
		ar2c_iowrite8_copy(config + adc_base + POMMAX3_ADC_TX,
				(const u8 *) cmd_be, 2 * size + 2);
		flags = readb(config + adc_base + POMMAX3_ADC_CFLAGS);
		writeb(POMMAX3_ADC_EXEC, config + adc_base + POMMAX3_ADC_CTRL);

		ret = pommax3_wait(flags);
	} while (ret);

	if (result != NULL)
		for (i = 1; i < rsize; i++)
			result[i] = readw(config + adc_base + POMMAX3_ADC_RX
					+ 2 * i);
	return 0;
}

static unsigned int cmd_seq[2];

static int exec_simple_command(unsigned int adc, unsigned int opcode,
		void *result, unsigned int rsize)
{
	u16 cmd[2];

	cmd[0] = ((ARRAY_SIZE(cmd) - 1) << 8) | POMMAX3_PROTO_CONFIG;
	cmd[1] = opcode | cmd_seq[adc];
	cmd_seq[adc] ^= 0x4000;

	return __exec_command(cmd, ARRAY_SIZE(cmd), result, rsize);
}

static int pommax3_adc_readw(unsigned int adc, unsigned int reg,
		u16 *data)
{
	u16 result[2];
	int ret;

	adc_base = (adc == 1) ? POMMAX3_ADC1_PTR : POMMAX3_ADC0_PTR;
	ret = exec_simple_command(adc, (POMMAX3_ADC_READ << 8) | reg,
			result, ARRAY_SIZE(result));
	if (ret < 0)
		return ret;
	*data = result[1];
	return 0;
}

static int pommax3_adc_writew(unsigned int adc, unsigned int value,
		unsigned int reg)
{
	u16 cmd[3];
	int ret;

	adc_base = (adc == 1) ? POMMAX3_ADC1_PTR : POMMAX3_ADC0_PTR;
	cmd[0] = ((ARRAY_SIZE(cmd) - 1) << 8) | POMMAX3_PROTO_CONFIG;
	cmd[1] = (POMMAX3_ADC_WRITE << 8) | cmd_seq[adc] | reg;
	cmd[2] = value;
	cmd_seq[adc] ^= 0x4000;

	ret = __exec_command(cmd, ARRAY_SIZE(cmd), NULL, 0);
	return ret;
}

static void reset_adc(struct arbus_dev *dev, int adc)
{
	int ret;

	adc_base = (adc == 1) ? POMMAX3_ADC1_PTR : POMMAX3_ADC0_PTR;
	ret = exec_simple_command(adc, POMMAX3_ADC_RESET, NULL, 0);
	if (ret) {
		dev_error(dev, "ADC reset error: %s, Marking ADC as dead\n",
				strerror(-ret));
		adc_dead[adc] = 1;
	}
}

static void show_data(int adc)
{
	int ret;
	u16 v;
	s16 x;
	unsigned int i;
	s16 gain;
	s16 ofix;

	for (i = 0; i < 4; i++) {
		ret = pommax3_adc_readw(adc, POMMAX3_ADC_GAIN_0 + i, &v);
		if (ret < 0) {
			fprintf(stderr, "%d: read error\n", i);
			continue;
		}
		gain = v << 1;

		ret = pommax3_adc_readw(adc, POMMAX3_ADC_OFIX_0 + i, &v);
		if (ret < 0) {
			fprintf(stderr, "%d: read error\n", i);
			continue;
		}

		ofix = v << 1;
		v = readw(legacy_base + 16 * (adc * 16 + i));

		x = v << 4;
		fprintf(stderr, "%d: %6.4f   %6.3f V   %6.3f mA (%04x)  "
				"gain=%8.5f ofix:%d\n",
				i, x / 32768.0,
				x / 32768.0 * 10,
				x / 32768.0 * 20, v,
				gain / 4096.0, ofix);
	}
}

static void show_config(int adc)
{
	u16 major;
	u16 minor;
	u16 freq;
	u16 flags;
	u16 offset;
	u16 serial;
	u16 v;
	s16 gain[8];
	s16 ofix[8];
	unsigned int i;
	int ret;

	ret = pommax3_adc_readw(adc, POMMAX3_ADC_MAJOR, &major);
	if (ret < 0)
		goto err;
	ret = pommax3_adc_readw(adc, POMMAX3_ADC_MINOR, &minor);
	if (ret < 0)
		goto err;
	ret = pommax3_adc_readw(adc, POMMAX3_ADC_FREQ, &freq);
	if (ret < 0)
		goto err;
	ret = pommax3_adc_readw(adc, POMMAX3_ADC_FLAGS, &flags);
	if (ret < 0)
		goto err;
	ret = pommax3_adc_readw(adc, POMMAX3_ADC_OFFSET, &offset);
	if (ret < 0)
		goto err;
	ret = pommax3_adc_readw(adc, POMMAX3_ADC_SERIAL, &serial);
	if (ret < 0)
		goto err;
	for (i = 0; i < 8; i++) {
		ret = pommax3_adc_readw(adc, POMMAX3_ADC_GAIN_0 + i, &v);
		gain[i] = v << 1;
		if (ret < 0)
			goto err;

		ret = pommax3_adc_readw(adc, POMMAX3_ADC_OFIX_0 + i, &v);
		ofix[i] = v << 1;
		if (ret < 0)
			goto err;
	}

	printf("POMMAX3 ADC version %d.%d  freq: %d  serial: %d\n",
			major, minor, freq, serial);
	printf("POMMAX3 flags: %x  offset: %d\n", flags, offset);
	for (i = 0; i < 8; i++)
		printf("channel %d: gain %.3f (%04x), offset %d (%04x)\n", i,
				gain[i] / 4096.0, gain[i],
				ofix[i], ofix[i]);
	return;
err:
	fprintf(stderr, "error: cannot read configuration from ADC %d\n",
			adc);
}

char *std_gets(void)
{
	char *tmp;

	if (feof(stdin))
		return NULL;
	tmp = malloc(1024);
	if (tmp == NULL) {
		fprintf(stderr, "fatal: Can't allocate memory.\n");
		exit(1);
	}
	if (fgets(tmp, 1024, stdin) == NULL) {
		free(tmp);
		return NULL;
	}
	return tmp;
}

char *rl_gets(void)
{
#ifdef HAVE_LIBREADLINE
	char *tmp;

	tmp = readline(PROMPT);
	if (tmp && *tmp)
		add_history(tmp);
	return tmp;
#else
	fprintf(stderr, "> ");
	return std_gets();
#endif
}

static void show_help(void)
{
	fprintf(stderr, "Commands:\n");
	fprintf(stderr, "\texit\n");
	fprintf(stderr, "\thelp\n");
	fprintf(stderr, "\tsave\n");
	fprintf(stderr, "\tfreq\n");
	fprintf(stderr, "\tgain\n");
}

int gen_parse_1ul(const char *args, unsigned long *addr)
{
	char cmd[64], addr_s[16];
	int ret;

	ret = sscanf(args, "%63s %15s", cmd, addr_s);
	if (ret < 2)
		return 1;
	*addr = strtoll(addr_s, NULL, 0);
	return 0;
}

static int parse_error(const char *fmt)
{
	fprintf(stderr, "parse error: %s\n", fmt);
	return 1;
}

static int error(const char *fmt)
{
	fprintf(stderr, "error: %s\n", fmt);
	return 1;
}

int cmd_show_data(const char *args)
{
	unsigned long adc;
	int ret;

	ret = gen_parse_1ul(args, &adc);
	if (ret)
		return parse_error(args);
	if (adc > 1) {
		fprintf(stderr, "invalid ADC\n");
		return 1;
	}

	show_data(adc);
	return 0;
}

int cmd_show_config(const char *args)
{
	unsigned long adc;
	int ret;

	ret = gen_parse_1ul(args, &adc);
	if (ret)
		return parse_error(args);
	if (adc > 1) {
		fprintf(stderr, "invalid ADC\n");
		return 1;
	}

	show_config(adc);
	return 0;
}

int cmd_set_gain(const char *args)
{
	char cmd[64];
	int adc;
	int ch;
	float gain;
	int ret;
	int g;

	ret = sscanf(args, "%63s %d %d %f", cmd, &adc, &ch, &gain);
	if (ret < 4)
		return parse_error(args);
	if (adc < 0 || adc > 1)
		return error("invalid ADC number, must be from 0 to 1\n");
	if (ch < 0 || ch > 7)
		return error("invalid channel number, must be from 0 to 7\n");
	if (gain < -8 || gain >= 8)
		return error("invalid gain, must be in [-8; 8) range\n");

	g = (int)(gain * 4096.0);
	if (g > 32767)
		g = 32767;
	else if (g < -32768)
		g = -32768;

	ret = pommax3_adc_writew(adc, (u16) g >> 1, POMMAX3_ADC_GAIN_0 + ch);
	if (ret)
		fprintf(stderr, "gain: cannot set gain: %s\n", strerror(-ret));
	return ret;
}

int cmd_set_ofix(const char *args)
{
	char cmd[64];
	int adc;
	int ch;
	int ofix;
	int ret;

	ret = sscanf(args, "%63s %d %d %d", cmd, &adc, &ch, &ofix);
	if (ret < 4)
		return parse_error(args);
	if (adc < 0 || adc > 1)
		return error("invalid ADC number, must be from 0 to 1\n");
	if (ch < 0 || ch > 7)
		return error("invalid channel number, must be from 0 to 7\n");
	if (ofix < -32768 || ofix >= 32768)
		return error("invalid ofix, must be in [-32768; 32768) range\n");

	ret = pommax3_adc_writew(adc, (u16) ofix >> 1, POMMAX3_ADC_OFIX_0 + ch);
	if (ret)
		fprintf(stderr, "ofix: cannot set ofix: %s\n", strerror(-ret));
	return ret;
}

int cmd_set_flags(const char *args, int enable)
{
	char cmd[64];
	char flag[64];
	int adc;
	int ret;
	u16 flags;
	unsigned int mask = 0;

	ret = sscanf(args, "%63s %d %64s", cmd, &adc, flag);
	if (ret < 3)
		return parse_error(args);

	if (adc < 0 || adc > 1)
		return error("invalid ADC number, must be from 0 to 1\n");

	if (!strcmp(flag, "gain")) {
		mask = POMMAX3_FLAGS_GAIN;
	} else {
		fprintf(stderr, "unknown flag %s\n", flag);
		return 1;
	}

	ret = pommax3_adc_readw(adc, POMMAX3_ADC_FLAGS, &flags);
	if (ret) {
		fprintf(stderr, "error: cannot read flags from ADC %d\n",
				adc);
		return 1;
	}
	if (enable)
		flags |= mask;
	else
		flags &= ~mask;
	ret = pommax3_adc_writew(adc, flags, POMMAX3_ADC_FLAGS);
	if (ret) {
		fprintf(stderr, "error: cannot write flags to ADC %d\n",
				adc);
		return 1;
	}
	return 0;
}

void pommax3_setup(struct arbus_dev *dev)
{
	u8 rev;
	u16 x;

	legacy_base = dev->base_addr[2];
	config = dev->base_addr[1];
	if (!config) {
		fprintf(stderr, "warning: BAR1 resource not assigned\n");

		pommax3_assign_resources(dev);
		config = POMMAX3_BAR1_ADDRESS;
	}

	fprintf(stderr, "Configuration registers at %05lx\n", config);

	cmd_seq[0] = 0;
	cmd_seq[1] = 0;

	rev = arbus_read_word(dev, PCI_REVISION_ID);
	fprintf(stderr, "pommax3: revision %d\n", rev);
	if (rev >= 2) {
		ar2c_writew(0xaaaa, config + POMMAX3_ADC1_PTR + POMMAX3_ADC_TX);
		x = ar2c_readw(config + POMMAX3_ADC1_PTR + POMMAX3_ADC_TX);
		if (!x) {
			fprintf(stderr, "pommax3: ADC1 is not present\n");
			adc_dead[1] = 1;
		} else if (x == 0xaaaa) {
			fprintf(stderr, "pommax3: ADC1 is present\n");
		} else {
			fprintf(stderr, "pommax3: ADC1 TX readback error: %02x\n",
					x);
		}
	}

	reset_adc(dev, 0);
	if (!adc_dead[1])
		reset_adc(dev, 1);
}

void configure_pommax3(struct arbus_dev *dev)
{
	char *tmp = NULL;
	char cmd[64];
	int ret;

	pommax3_setup(dev);
	while (1) {
		if (tmp)
			free(tmp);
		tmp = rl_gets();
		if (tmp == NULL) {
			fprintf(stderr, "\n");
			break;
		}
		if ((*tmp == '\0') || (*tmp == '#'))
			continue;
		ret = sscanf(tmp, "%63s\n", cmd);
		if ((ret < 1) || (*cmd == '\0'))
			continue;

		if (!strcmp(cmd, "exit") || !strcmp(cmd, "quit") ||
				!strcmp(cmd, "q")) {
			break;
		} else if (!strcmp(cmd, "help")) {
			show_help();
		} else if (!strcmp(cmd, "gain")) {
			cmd_set_gain(tmp);
		} else if (!strcmp(cmd, "ofix")) {
			cmd_set_ofix(tmp);
		} else if (!strcmp(cmd, "enable")) {
			cmd_set_flags(tmp, 1);
		} else if (!strcmp(cmd, "disable")) {
			cmd_set_flags(tmp, 0);
		} else if (!strcmp(cmd, "show")) {
			cmd_show_config(tmp);
		} else if (!strcmp(cmd, "s")) {
			cmd_show_data(tmp);
		}
	}
}

static void getconfig_adc(struct arbus_dev *dev, struct ar2c_config *c,
		unsigned int adc)
{
	char name[128];
	char source[128];
	char devn[128];
	u16 major;
	u16 minor;
	u16 freq;
	u16 flags;
	u16 offset;
	u16 serial;
	u16 v;
	s16 gain[8];
	s16 ofix[8];
	unsigned int i;
	int ret;

	ret = pommax3_adc_readw(adc, POMMAX3_ADC_MAJOR, &major);
	if (ret < 0)
		goto err;
	ret = pommax3_adc_readw(adc, POMMAX3_ADC_MINOR, &minor);
	if (ret < 0)
		goto err;
	ret = pommax3_adc_readw(adc, POMMAX3_ADC_FREQ, &freq);
	if (ret < 0)
		goto err;
	ret = pommax3_adc_readw(adc, POMMAX3_ADC_FLAGS, &flags);
	if (ret < 0)
		goto err;
	ret = pommax3_adc_readw(adc, POMMAX3_ADC_OFFSET, &offset);
	if (ret < 0)
		goto err;
	ret = pommax3_adc_readw(adc, POMMAX3_ADC_SERIAL, &serial);
	if (ret < 0)
		goto err;
	for (i = 0; i < 8; i++) {
		ret = pommax3_adc_readw(adc, POMMAX3_ADC_GAIN_0 + i, &v);
		gain[i] = v << 1;
		if (ret < 0)
			goto err;

		ret = pommax3_adc_readw(adc, POMMAX3_ADC_OFIX_0 + i, &v);
		ofix[i] = v << 1;
		if (ret < 0)
			goto err;
	}

	snprintf(source, 128, "pommax3");
	snprintf(devn, 128, "PCI %02x.%d",
			PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn));

	snprintf(name, 128, "%s adc%d version", devn, adc);
	ar2c_config_set_all_fmt(c, name, source, 1, "%d.%d", major, minor);

	snprintf(name, 128, "%s adc%d flags", devn, adc);
	ar2c_config_set_all_fmt(c, name, source, 1, "%d", flags);

	snprintf(name, 128, "%s adc%d offset", devn, adc);
	ar2c_config_set_all_fmt(c, name, source, 1, "%d", offset);

	snprintf(name, 128, "%s adc%d serial", devn, adc);
	ar2c_config_set_all_fmt(c, name, source, 1, "%d", serial);

	snprintf(name, 128, "%s adc%d freq", devn, adc);
	ar2c_config_set_all_fmt(c, name, source, 1, "%d", freq);

	for (i = 0; i < 8; i++) {
		snprintf(name, 128, "%s adc%d gain%d", devn, adc, i);
		ar2c_config_set_all_fmt(c, name, source, 1, "%d", gain[i]);

		snprintf(name, 128, "%s adc%d ofix%d", devn, adc, i);
		ar2c_config_set_all_fmt(c, name, source, 1, "%d", ofix[i]);

		if (flags & POMMAX3_FLAGS_OFFSET) {
			snprintf(name, 128, "%s adc%d offset%d", devn, adc, i);
			ar2c_config_set_all_fmt(c, name, source, 1, "%d",
					-ofix[i]);
		}
	}
	return;
err:
	dev_error(dev, "cannot read config from ADC %d\n", adc);
}

void pommax3_getconfig(struct arbus_dev *dev, struct ar2c_config *c)
{
	pommax3_setup(dev);

	if (!adc_dead[0])
		getconfig_adc(dev, c, 0);
	if (!adc_dead[1])
		getconfig_adc(dev, c, 1);
}

static void putconfig_adc(struct arbus_dev *dev, struct ar2c_config *c,
		unsigned int adc)
{
	char name[128];
	char devn[128];
	unsigned int i;
	const char *x;
	int ret;
	u16 v;

	snprintf(devn, 128, "PCI %02x.%d",
			PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn));

	snprintf(name, 128, "%s adc%d freq", devn, adc);
	x = __ar2c_config_get(c, name);
	if (x) {
		v = strtol(x, NULL, 0);
		ret = pommax3_adc_writew(adc, v, POMMAX3_ADC_FREQ);
		if (ret) {
			dev_error(dev, "POMMAX3_ADC_FREQ write failed: %s\n",
					strerror(-ret));
		}
		fprintf(stderr, "freq to %d\n", v);
	}

	snprintf(name, 128, "%s adc%d offset", devn, adc);
	x = __ar2c_config_get(c, name);
	if (x) {
		v = strtol(x, NULL, 0);
		ret = pommax3_adc_writew(adc, v, POMMAX3_ADC_OFFSET);
		if (ret) {
			dev_error(dev, "POMMAX3_ADC_OFFSET write failed: %s\n",
					strerror(-ret));
		}
	}

	for (i = 0; i < 8; i++) {
		snprintf(name, 128, "%s adc%d gain%d", devn, adc, i);
		x = __ar2c_config_get(c, name);
		if (x) {
			v = strtol(x, NULL, 0);
			ret = pommax3_adc_writew(adc, v >> 1,
					POMMAX3_ADC_GAIN_0 + i);
			if (ret) {
				dev_error(dev, "POMMAX3_ADC_GAIN%d write failed: %s\n",
						i, strerror(-ret));
			}
			fprintf(stderr, "gain[%d] to %d\n", i, v);
		}

		snprintf(name, 128, "%s adc%d ofix%d", devn, adc, i);
		x = __ar2c_config_get(c, name);
		if (x) {
			v = strtol(x, NULL, 0);
			ret = pommax3_adc_writew(adc, v >> 1,
					POMMAX3_ADC_OFIX_0 + i);
			if (ret) {
				dev_error(dev, "POMMAX3_ADC_OFIX%d write failed: %s\n",
						i, strerror(-ret));
			}
			fprintf(stderr, "ofix[%d] to %d\n", i, v);
		}
	}

	snprintf(name, 128, "%s adc%d flags", devn, adc);
	x = __ar2c_config_get(c, name);
	if (x) {
		v = strtol(x, NULL, 0);
		ret = pommax3_adc_writew(adc, v, POMMAX3_ADC_FLAGS);
		if (ret) {
			dev_error(dev, "POMMAX3_ADC_FLAGS write failed: %s\n",
					strerror(-ret));
		}

		fprintf(stderr, "flags to %d\n", v);
	}
}

void pommax3_putconfig(struct arbus_dev *dev, struct ar2c_config *c)
{
	pommax3_setup(dev);

	if (!adc_dead[0])
		putconfig_adc(dev, c, 0);
	if (!adc_dead[1])
		putconfig_adc(dev, c, 1);
}

static void pommax3_adc_save(struct arbus_dev *dev, unsigned int adc)
{
	int ret;
	
	ret = pommax3_adc_writew(adc, POMMAX3_ADC_SAVE_MAGIC,
			POMMAX3_ADC_SAVE);
	if (ret) {
		dev_error(dev, "adc%d: POMMAX3_ADC_SAVE failed: %s\n",
				adc, strerror(-ret));
	}
}

void pommax3_save(struct arbus_dev *dev)
{
	pommax3_setup(dev);

	if (!adc_dead[0])
		pommax3_adc_save(dev, 0);
	if (!adc_dead[1])
		pommax3_adc_save(dev, 1);
}

#define swap(x, y) \
	do { \
		int t = x; \
		x = y; \
		y = t; \
	} while (0)

static int median3(int a, int b, int c)
{
	if (a > b)
		swap(a, b);
	if (b > c)
		swap(b, c);
	if (a > b)
		swap(a, b);
	return b;
}

static s16 pommax_legacy_read(const u8 *data, unsigned int ch,
		unsigned int lag)
{
	unsigned int base = 16 * ch + 2 * lag;

	return (s16)(((data[base + 1] << 8) | data[base]) << 4);
}

void pommax3_get_data(struct arbus_dev *dev, struct ar2c_config *c)
{
	u8 data[256];
	char name[128];
	char devn[128];
	unsigned int adc;
	unsigned int i;
	int err;

	legacy_base = dev->base_addr[2];
	if (!legacy_base)
		return;

	snprintf(devn, 128, "PCI %02x.%d",
			PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn));

	for (adc = 0; adc < 2; adc++) {
		err = ar2c_dump_memory(legacy_base + adc * 0x100, data);
		if (err)
			continue;

		for (i = 0; i < 8; i++) {
			s16 x, y, z;

			x = pommax_legacy_read(data, i, 0);
			y = pommax_legacy_read(data, i, 1);
			z = pommax_legacy_read(data, i, 2);

			snprintf(name, 128, "%s adc%d value%d", devn, adc, i);
			ar2c_config_set_all_fmt(c, name, "pommax3", 1, "%hd",
					median3(x, y, z));
		}
	}
}
